package eve.database;
import java.util.Vector;
import java.io.IOException;

import eve.sys.Handle;
import eve.sys.HandleStoppedException;
import eve.sys.IRemoteProxyMaker;
import eve.sys.Task;
import eve.sys.Time;
import eve.util.IntArray;
import eve.util.mString;
import eve.sys.Vm;
/**
* This class provides utilities for synchronizing items in a mobile database with
* another. The methods in here only support synchronization with a single other database,
* you will have to implement more complex methods to support any other type of synchronization.
**/
//##################################################################
public class Synchronizer implements IRemoteSynchronizer{
//##################################################################

Database database;
public Time syncTime = new Time();
public int batchSize = 100;
public int options = 0;

public int getSetOptions(int toSet, int toClear)
{
	options |= toSet;
	options &= ~toClear;
	return options;
}
//===================================================================
public Synchronizer(Database database)
//===================================================================
{
	this.database = database;	
}
/**
 * Setup this Synchronizer to act as the "Remote" portion of a synchronization
 * process, by accepting remote method calls over the specified channel.<p>
 * After calling this method this Synchronizer will receive method calls in
 * a separate Thread from the peer Synchronizer on the other side of the
 * connection. This Thread should monitor the returned Handle and once it
 * has stopped you can assume the synchronization is complete.  
 * @param inputChannel this can be an InputStream, a BlockInputStream or
 * a Socket.
 * @param outputChannel this can be an OutputStream, a BlockOutputStream or
 * a Socket.
 * @return a Handle that can be used to monitor or close the connection.
 * The stopped bit will be set once the channels are closed.
 * @throws IOException on any IOException.
 */
public Handle actAsRemote(Object inputChannel, Object outputChannel) throws IOException
{
	IRemoteProxyMaker pm = Vm.getRemoteProxyMaker(IRemoteSynchronizer.class, 0, null);
	return pm.makeServerObject(this, inputChannel, outputChannel);
}
/**
 * Get an IRemoteSynchronizer that is a proxy for a remote IRemoteSynchronizer
 * that is at the other end of the channel and which will receive remote method
 * calls over the channels.
 * @param inputChannel this can be an InputStream, a BlockInputStream or
 * a Socket.
 * @param outputChannel this can be an OutputStream, a BlockOutputStream or
 * a Socket.
 * @return an IRemoteSynchronizer that you can use with synchronize(Handle h,
 * IRemoteSynchronizer remote).
 * @throws IOException on any IO error.
 */
public IRemoteSynchronizer getRemoteProxy(Object inputChannel,Object outputChannel) throws IOException
{
	IRemoteProxyMaker pm = Vm.getRemoteProxyMaker(IRemoteSynchronizer.class, 0, null);
	pm.setConnection(inputChannel, outputChannel);
	return (IRemoteSynchronizer)pm.createProxy();
}
//-------------------------------------------------------------------
private int checkSort(String sortName)
//-------------------------------------------------------------------
{
	return checkSort(database,sortName);
}
//-------------------------------------------------------------------
protected static int checkSort(Database database,String sortName) throws IllegalStateException
//-------------------------------------------------------------------
{
	int got = database.findSort(sortName);
	if (got == 0) throw new IllegalStateException("Database does not have the appropriate sort criteria.");
	return got;
}
//===================================================================
public static EntriesView getUnsynchronized(Handle h,Database database) throws IOException, IllegalStateException
//===================================================================
{
	return getUnsynchronized(h, database, 0);
}
//===================================================================
public static EntriesView getUnsynchronized(Handle h,Database database,int options) throws IOException, IllegalStateException
//===================================================================
{
	if ((options & OPTION_SYNCHRONIZE_ALL_ENTRIES) != 0){
		return database.getFoundEntries(h, 0).getFullView(null);
	}else{
		FoundEntries all = database.getFoundEntries(h,checkSort(database,database.SyncSortName));
		if (all == null) return null;
		return all.getEmptyView().search(h,new EntrySelector(all,null,1,true){
			protected int doCompare(DatabaseEntry data,DatabaseEntry de){
				if (!de.hasField(Database.FLAGS_FIELD)) return 1;
				int flags = de.getField(Database.FLAGS_FIELD,0);
				if ((flags & Database.FLAG_SYNCHRONIZED) != 0) return -1;
				return 0;
			}
		});
	}
}
/**
* This gets all the unsynchronized entries, but not the deleted entries. You
* can get deleted items from the database directly.
 * @exception IOException 
 * @exception IllegalStateException If the database does not have the "_SortBySync" sort
**/
//===================================================================
public EntriesView getUnsynchronized(Handle h) throws IOException, IllegalStateException
//===================================================================
{
	return getUnsynchronized(h,database,options);
}
//===================================================================
public static void markAsSynchronized(DatabaseEntry entry)
//===================================================================
{
	int value = entry.getField(Database.FLAGS_FIELD,0);
	value |= Database.FLAG_SYNCHRONIZED;
	entry.setField(Database.FLAGS_FIELD,value);
}
/**
* This marks the entry at the specified index as being synchronized.
**/
//===================================================================
public static void markAsSynchronized(EntriesView entries,int index) throws IOException
//===================================================================
{
	DatabaseEntry de = entries.get(index);
	if (de != null){
		markAsSynchronized(de);
		entries.store(index,de);
	}
}
//===================================================================
public static FoundEntries getByOID(Handle h,Database database) throws IOException, IllegalStateException
//===================================================================
{
	return database.getFoundEntries(h,checkSort(database,Database.OidSortName));
}
//===================================================================
public FoundEntries getByOID(Handle h) throws IOException, IllegalStateException
//===================================================================
{
	return getByOID(h,database);
}
/**
* This locates the entry with the specified OID. The FoundEntries object must
* have been found using the findByOID() method (i.e. using the sortByOid
* criteria).
* @param entries The FoundEntries of all the entries.
* @param oid The oid looking for.
* @return The index of the entry, -1 if it was not found.
*/
//===================================================================
public static int findByOID(FoundEntries entries,long oid) throws IOException
//===================================================================
{
	//return entries.findFirst(entries.getSearchData(Database.OID_FIELD,new ewe.sys.Long().set(oid)),entries.getFieldComparer());
	try{
		return entries.findFirst(null,new Long(oid),false);
	}catch(HandleStoppedException e){
		return -1;
	}
}
/**
 * Get the OID for an entry.
 * @param entry The entry to look for.
 * @return the OID for the entry.
 * @exception IllegalStateException if the entry does not have an OID assigned.
 */
//===================================================================
public static long getOID(DatabaseEntry entry) throws IllegalStateException
//===================================================================
{
	if (!entry.hasField(Database.OID_FIELD)) throw new IllegalStateException("No OID assigned.");
	return entry.getField(Database.OID_FIELD,(long)0);
}
/**
 * Returns the state of the synchronized flag.
 * @param entry The entry to check.
 * @return true if the entry has the syncrhonized flag set, false if not.
 * @exception IllegalStateException If synchronization data is not set in the entry.
 */
//===================================================================
public boolean isSynchronized(DatabaseEntry entry) throws IllegalStateException
//===================================================================
{
	if (!entry.hasField(Database.FLAGS_FIELD)) throw new IllegalStateException("No Flags assigned.");
	return ((entry.getField(Database.FLAGS_FIELD,(int)0) & Database.FLAG_SYNCHRONIZED) != 0);
}
//===================================================================
public static int getMatchingEntry(DatabaseEntry remoteEntry,FoundEntries myEntriesSortedByOID) throws IOException, IllegalStateException
//===================================================================
{
	long oid = getOID(remoteEntry);
	return findByOID(myEntriesSortedByOID,oid);
}
/**
* Given an encoded remote entry, add or replace it in this database.
**/
public static void addOrReplace(byte[] encodedEntry, FoundEntries myEntriesSortedByOID) throws IOException, IllegalStateException
{
	DatabaseEntry de = myEntriesSortedByOID.getNew();
	de.decode(encodedEntry,0,encodedEntry.length);
	addOrReplace(de,myEntriesSortedByOID);
}
/**
* Given a remote entry, add or replace it in this database.
**/
//===================================================================
public static void addOrReplace(DatabaseEntry entry,FoundEntries myEntriesSortedByOID) throws IOException, IllegalStateException
//===================================================================
{
	int toReplace = getMatchingEntry(entry,myEntriesSortedByOID);
	if (toReplace == -1){
		//Vm.debug("Adding: "+entry.getField(Database.OID_FIELD,(long)0)+" = "+entry.getData());
		DatabaseEntry de = myEntriesSortedByOID.getNew();
		de.duplicateFrom(entry);
		markAsSynchronized(de);
		de.store();
	}else{
		DatabaseEntry de = myEntriesSortedByOID.get(toReplace);
		/*
		Vm.debug("Replacing oid: "+de.getField(Database.OID_FIELD,(long)0)+" with: "+
			entry.getField(Database.OID_FIELD,(long)0));
		Vm.debug("Replacing: "+de.getData()+"with: "+entry.getData());
		*/
		de.duplicateFrom(entry);
		markAsSynchronized(de);
		myEntriesSortedByOID.store(toReplace,de);		
	}
}
/**
 * This is called after an OID marked as deleted on this database has been
 * erased from the remote database. It erases it from this database too.
 * @param oid The oid to erase.
 * @exception IOException 
 */
//===================================================================
public void eraseDeleted(long oid) throws IOException
//===================================================================
{
	database.eraseDeletedEntry(oid);
}
/**
* Given the OID of an entry in the remote database, erase it completely from this database.
**/
//===================================================================
public void erase(long oid,FoundEntries myEntriesSortedByOID) throws IOException
//===================================================================
{
	int found = findByOID(myEntriesSortedByOID,oid);
	if (found == -1) return;
	myEntriesSortedByOID.erase(found);
}

//===================================================================
public int countRemoteUnsynchronizedEntries() throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");
}
//===================================================================
public int countRemoteDeletedEntries() throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");
}
//===================================================================
public long getRemoteDeletedEntry(int index) throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");
}
//===================================================================
public void eraseDeletedOnRemote(int index) throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");
}
//===================================================================
public void eraseEntryOnRemote(long OID) throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");
}
//===================================================================
public byte [] getRemoteUnsynchronizedEntry(int index) throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");
}
//===================================================================
public void markRemoteAsSynchronized(int index) throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");	
}
//===================================================================
public void sendEntryToRemote(DatabaseEntry de) throws IOException
//===================================================================
{
	throw new IllegalStateException("You must override this method.");	
}
//-------------------------------------------------------------------
protected String getDisplayString(DatabaseEntry entry)
//-------------------------------------------------------------------
{
	return mString.toString(entry.getData(null));
}
//-------------------------------------------------------------------
protected String getResolveString(DatabaseEntry remote,DatabaseEntry local)
//-------------------------------------------------------------------
{
	return "The item:\n"+getDisplayString(remote)+
	"\n was modified on the remote database \nand on the local database.\n\nWhat action should be taken?";
}

protected int lastConflictCommand = 0;

public static final int REMOTE_TO_LOCAL = 1;
public static final int LOCAL_TO_REMOTE = 2;
public static final int RESOLVE_LATER = 3;
public static final int ALWAYS = 0x40000000;
public static final int CANCELLED = 0;
/**
 * This must resolve a conflict when an item has been modified on both databases.
 * @param remote The remote entry.
 * @param local The local entry.
 * @return RESOLVE_LATER or REMOTE_TO_LOCAL or LOCAL_TO_REMOTE or CANCELLED. This can
	 also be ORed with ALWAYS so that the value will apply to other conflicts.
 */
//===================================================================
public int resolveConflict(DatabaseEntry remote,DatabaseEntry local)
//===================================================================
{
	throw new IllegalMonitorStateException("You must override resolveConflict()");
}
/*
{
	
	MessageBox mb = new MessageBox("Resolve Conflict",getResolveString(remote,local),MessageBox.MBOKCANCEL);
	mb.windowFlagsToClear |= Window.FLAG_HAS_CLOSE_BUTTON;
	ResolveConflict rc = new ResolveConflict();
	rc.fromCommand(lastConflictCommand);
	mb.getAfterPanel().addLast(rc.getEditor(0)).setTag(mb.INSETS,new ewe.fx.Insets(4,2,4,2));
	if (mb.execute() == mb.IDCANCEL) return CANCELLED;
	return lastConflictCommand = rc.toCommand();
}
*/
/**
* This is called at the end of the syncrhonization. Override if necessary.
**/
//===================================================================
public void close()
//===================================================================
{
}
//===================================================================
public Handle synchronize()
//===================================================================
{
	Task t = new Task(){
		protected void doRun(){
			try{
				try{
					synchronize(this);
				}finally{
					close();
				}
				set(Succeeded);
			}catch(Exception e){
				//e.printStackTrace();
				fail(e);
			}
		}
	};
	return t;
}

/**
 * Start the synchronization and display an optional progress bar.
 * @param title if this is non-null a progress bar display will be shown.
 * @return The Handle that monitors the progress of the synchronization.
 */
/*
//===================================================================
public Handle synchronize()
//===================================================================
{
	final Handle h = synchronize();
	if (title != null)
		new Thread(){
			public void run(){
				ProgressBarForm pbf = new ProgressBarForm();
				pbf.showSubTask = true;
				pbf.showMainTask = false;
				pbf.showStop = true;
				pbf.execute(h,title);
			}
		}.start();
	return h;
}
*/
/**
* If this is true then the local database will take entries from remote database
* but not vice versa. It is false by default.
**/
public boolean remoteToLocalOnly = false;
/**
* If this is true, then deleting an item on one database will not cause the item
* to be deleted in the other. It is false by default.
**/
public boolean dontSynchronizeDeletedItems = false;
/**
* This goes through the motion of synchronization but does not actually make any
* changes to either database.
**/
public boolean dontSyncJustDebug = false;

/**
 * @deprecated use synchronize(Handle progress, IRemoteSynchronizer remote) instead.
 * @param progress
 * @return
 * @throws IOException
 */
//===================================================================
public boolean synchronize(Handle progress) throws IOException
//===================================================================
{
	Handle h = progress == null ? new Handle() : progress;
//.................................................	
	h.doing = "Checking databases...";
	h.progress = 0;
	h.set(h.Running);
	h.changed();
//.................................................	
	int num = countRemoteUnsynchronizedEntries();
	int deleted = dontSynchronizeDeletedItems ? 0 : countRemoteDeletedEntries();
	int always = 0;
	Vector dontSync = new Vector();
	int total = num+deleted;
	//Vm.out.println("Total local to sync: "+total);
	int remoteSize = total;
	if (!remoteToLocalOnly){
 		total += getUnsynchronized(h/*.getSubHandle()*/).size();
		total += dontSynchronizeDeletedItems ? 0 : database.getDeletedSince(null).length;
	}
	//Vm.out.println("Total local & remote to sync: "+total);
	FoundEntries fe = num == 0 && deleted == 0 ? null : getByOID(h/*.getSubHandle()*/);
	if (num != 0){
		for (int i = 0; i<num && !h.shouldStop; i++){
			try{
				h.doing = "Syncing: "+(i+1)+" of: "+total;
				h.progress = i/(float)total;
				h.changed();
				byte [] un = getRemoteUnsynchronizedEntry(i);
				if (un == null) break;
				DatabaseEntry de = fe.getNew();
				de.decode(un,0,un.length);
				if (dontSyncJustDebug)
					Vm.debug("Remote entry: "+de.getData(null));
				int found = getMatchingEntry(de,fe);
				if (found != -1){
					DatabaseEntry my = fe.get(found);
					if (!isSynchronized(my)){
						int resolve = always == 0 ?  resolveConflict(de,my) : always;
						int r = resolve & ~ALWAYS;
						if (r == CANCELLED) return false;
						if ((resolve & ALWAYS) != 0) always = r;
						if (r == LOCAL_TO_REMOTE) {
							if (!dontSyncJustDebug)
								markRemoteAsSynchronized(i);
							else
								Vm.debug("Marking remote entry as synchronized.");
							continue;
						}else if (r == RESOLVE_LATER){
							dontSync.add(new Long(getOID(de)));
							//Vm.debug("Resolve later: "+database.getData(de,null));
							continue;
						}
					}
				}
				if (!dontSyncJustDebug){
					addOrReplace(de,fe);
					markRemoteAsSynchronized(i);
				}else{
					Vm.debug("Accepting remote entry.");
				}
			}catch(IllegalStateException e){
			}
		}
	}
	if (deleted != 0){
		for (int i = 0; i<deleted && !h.shouldStop; i++){
			h.doing = "Syncing: "+(num+i+1)+" of: "+total;
			h.progress = (num+i)/(float)total;
			h.changed();
			long del = getRemoteDeletedEntry(i);
			if (del == 0) break;
			if (!dontSyncJustDebug){
				erase(del,fe);
				eraseDeletedOnRemote(i);
			}else{
				Vm.debug("Erasing deleted entry on local and remote.");
			}
		}
	}
	//Vm.debug("--------- Local to remote ---------- : "+total);
//..................................................................................	
	if (!remoteToLocalOnly){
//..................................................................................	
		EntriesView mine = getUnsynchronized(h/*.getSubHandle()*/);
		for (int i = 0; i<mine.size() && !h.shouldStop; i++){
			h.doing = "Syncing: "+(remoteSize+i+1)+" of: "+total;
			h.progress = (remoteSize+i)/(float)total;
			h.changed();
			DatabaseEntry de = mine.get(i);
			if (dontSyncJustDebug)
				Vm.debug("Local entry: "+de.getData(null));
			if (dontSync.size() != 0)
				if (dontSync.indexOf(new Long(getOID(de))) != -1) continue;
			//Vm.debug("Local->Remote: "+database.getData(de,null));
			if (!dontSyncJustDebug){
				sendEntryToRemote(de);
				markAsSynchronized(mine,i);
			}else{
				Vm.debug("Sending entry and synchronizing.");
			}
		}
		if (!dontSynchronizeDeletedItems){
			long [] del = database.getDeletedSince(null);
			int ds = remoteSize+mine.size();
			for (int i = 0; i<del.length && !h.shouldStop; i++){
				h.doing = "Syncing: "+(ds+i+1)+" of: "+total;
				h.progress = (ds+i)/(float)total;
				h.changed();
				if (!dontSyncJustDebug){
					eraseEntryOnRemote(del[i]);
					eraseDeleted(del[i]);
				}else{
					Vm.debug("Erasing on local and remote.");
				}
			}
		}
	}
//..................................................................................	
	h.doing = "Complete.";
	h.progress = 1.0f;
	h.changed();
	return true;
}
private void progress(Handle h, int cur, int total)
{
	h.doing = "Syncing: "+(cur+1)+" of: "+total;
	h.setProgress(cur/(float)total);
}
//===================================================================
public boolean synchronize(Handle progress,IRemoteSynchronizer remote) throws IOException
//===================================================================
{
	Handle h = progress == null ? new Handle() : progress;
//.................................................	
	h.doing = "Checking databases...";
	h.progress = 0;
	h.set(h.Running);
	h.changed();
//.................................................	
	int remoteUnsynchronized = remote.countUnsynchronizedEntries();//countRemoteUnsynchronizedEntries();
	int remoteDeleted = dontSynchronizeDeletedItems ? 0 : remote.countDeletedEntries();
	int always = 0;
	Vector dontSync = new Vector();
	int total = remoteUnsynchronized+remoteDeleted;
	if (dontSyncJustDebug) Vm.debug("Remote to sync: "+total);
	int remoteSize = total;
	if (!remoteToLocalOnly){
 		total += getUnsynchronized(h/*.getSubHandle()*/).size();
		total += dontSynchronizeDeletedItems ? 0 : database.getDeletedSince(null).length;
	}
	int doneSoFar = 0;
	if (dontSyncJustDebug) Vm.debug("Total to sync: "+total);
	IntArray ia = new IntArray();
	FoundEntries fe = remoteUnsynchronized == 0 && remoteDeleted == 0 ? null : getByOID(h);
	if (!h.shouldStop){
		//
		// Do remote unsynchronized.
		//
		int left = remoteUnsynchronized;
		while(left != 0 && !h.shouldStop){
			int toDo = batchSize;
			if (toDo > left) toDo = left;
			int startIndex = remoteUnsynchronized-left;
			Vector all = remote.getUnsynchronizedEntries(startIndex, toDo);
			int num = all.size();
			left -= num;
			for (int i = 0; i<num && !h.shouldStop; i++){
				try{
					progress(h,doneSoFar,total);
					byte[] un = (byte[])all.get(i);
					DatabaseEntry de = fe.getNew();
					de.decode(un,0,un.length);
					if (dontSyncJustDebug)
						Vm.debug("Remote entry: "+de.getData(null));
					int found = getMatchingEntry(de,fe);
					if (found != -1){
						DatabaseEntry my = fe.get(found);
						if (!isSynchronized(my)){
							int resolve = always == 0 ?  resolveConflict(de,my) : always;
							int r = resolve & ~ALWAYS;
							if (r == CANCELLED) return false;
							if ((resolve & ALWAYS) != 0) always = r;
							if (r == LOCAL_TO_REMOTE) {
								if (!dontSyncJustDebug)
									// Mark remote as synchronized, the next phase
									// will overwrite it with my unsynchronized version.
									ia.add(startIndex+i);
								else
									Vm.debug("Marking remote entry as synchronized.");
								continue;
							}else if (r == RESOLVE_LATER){
								dontSync.add(new Long(getOID(de)));
								//Vm.debug("Resolve later: "+database.getData(de,null));
								continue;
							}
						}
					}
					if (!dontSyncJustDebug){
						addOrReplace(de,fe);
						ia.add(startIndex+i);
					}else{
						Vm.debug("Accepting remote entry.");
					}
				}catch(IllegalStateException e){
				}
				doneSoFar++;
			}
			remote.markAsSynchronized(ia.data,0,ia.length);
			ia.clear();
		}
	}
	if (!h.shouldStop){
		//
		// Do remote deleted.
		//
		int left = remoteDeleted;
		while(left != 0 && !h.shouldStop){
			int toDo = batchSize;
			if (toDo > left) toDo = left;
			int startIndex = remoteDeleted-left;
			long[] all = remote.getDeletedEntries(startIndex, toDo);
			int num = all.length;
			left -= num;
			for (int i = 0; i<num && !h.shouldStop; i++){
				progress(h,doneSoFar,total);
				long del = all[i];
				if (del == 0) break;
				if (!dontSyncJustDebug){
					erase(del,fe);
					ia.add(startIndex+i);
				}else{
					Vm.debug("Erasing deleted entry on local and remote.");
				}
				doneSoFar++;
			}
			remote.eraseDeleted(ia.data,0,ia.length);
			ia.clear();
		}
	}
	//Vm.debug("--------- Local to remote ---------- : "+total);
//..................................................................................	
	if (!remoteToLocalOnly && !h.shouldStop){
//..................................................................................	
		EntriesView mine = getUnsynchronized(h);
		Vector toSend = new Vector();
		ia.clear();
		for (int i = 0; i<mine.size() && !h.shouldStop; i++){
			progress(h,doneSoFar,total);
			DatabaseEntry de = mine.get(i);
			if (dontSyncJustDebug)
				Vm.debug("Local entry: "+de.getData(null));
			if (dontSync.size() != 0)
				if (dontSync.indexOf(new Long(getOID(de))) != -1) {
					doneSoFar++;
					continue;
				}
			//Vm.debug("Local->Remote: "+database.getData(de,null));
			if (!dontSyncJustDebug){
				toSend.add(de.encode());
				if (toSend.size() >= batchSize){
					remote.addEntries(toSend);
					toSend.clear();
				}
				ia.add(i);
			}else{
				Vm.debug("Sending entry and synchronizing.");
			}
			doneSoFar++;
		}
		// Send left over new entries.
		if (!h.shouldStop && toSend.size() != 0)
			remote.addEntries(toSend);
		toSend.clear();
		// Now synchronize mine.
		if (!dontSyncJustDebug)
			for (int i = 0; i<ia.length && !h.shouldStop; i++)
				markAsSynchronized(mine,ia.data[i]);
		ia.clear();
		// Deleted items.
		if (!h.shouldStop && !dontSynchronizeDeletedItems){
			long [] del = database.getDeletedSince(null);
			int ds = remoteSize+mine.size();
			int inArr = 0;
			long [] toEraseOnRemote = new long[batchSize];
			for (int i = 0; i<del.length && !h.shouldStop; i++){
				progress(h,doneSoFar,total);
				if (!dontSyncJustDebug){
					toEraseOnRemote[inArr++]  = del[i];
					if (inArr >= toEraseOnRemote.length){
						remote.eraseEntries(toEraseOnRemote,0,inArr);
						inArr = 0;
					}
				}else{
					Vm.debug("Erasing on local and remote.");
				}
			}
			if (!h.shouldStop && inArr != 0)
				remote.eraseEntries(toEraseOnRemote,0,inArr);
			if (!dontSyncJustDebug)
				for (int i = 0; i<del.length && !h.shouldStop; i++){
					database.eraseDeletedEntries();
				}
		}
	}
	if (h.shouldStop) return false;
//..................................................................................
	h.doing = "Done.";
	h.setProgress(1.0f);
	return true;
}
/**
* This finds all items modified since the specfied date.
* @param Time The time after which you want to check for modifications.
* @return The entries found.
* @exception IOException 
* @exception IllegalStateException if the database does not have the "Sort by Modified" sort
* criteria set.
*/
/*
//===================================================================
public FoundEntries getModifiedSince(final Time when) throws IOException, IllegalStateException
//===================================================================
{
	int id = database.findSort(database.ModifiedSortName);
	if (id == 0) throw new IllegalStateException("The database is not set to sort by Modified Date");
	FoundEntries all = database.getEntries(id);
	if (when == null) return all;
	return all.getSubSet(all.findAll(null,new Comparer(){
		Time myTime = new Time();
		public int compare(Object one,Object two){
			DatabaseEntry de = (DatabaseEntry)two;
			if (!de.hasField(Database.MODIFIED_FIELD)) return 1;
			Time t = de.getField(Database.MODIFIED_FIELD,myTime);
			return when.compareTo(t);
		}
	},null));
}
*/
/**
* This finds all items modified since the database was synchronized with another database.
**/
/*
//===================================================================
public FoundEntries getModifiedSinceLastSynchronized(int remoteDatabaseID) throws IOException, IllegalStateException
//===================================================================
{
	return getModifiedSince(database.getSynchronizedTime(remoteDatabaseID));
}
*/
/**
* This marks the database as having being fully synchronized with another at a point in time.
**/
/*
//===================================================================
public void finalizeSynchronization(int remoteDatabaseID,Time when) throws IOException
//===================================================================
{
	database.setSynchronizedTime(remoteDatabaseID,when == null ? new Time() : when);
}
*/
public void addEntries(Vector entries) throws IOException 
{
	FoundEntries sbo = remoteGetByOid();
	if (sbo == null) return;
	int num = entries.size();
	for (int i = 0; i<num; i++){
		Object obj = entries.get(i);
		if (!(obj instanceof byte[])) continue;
		addOrReplace((byte[])obj, sbo);
	}
}
public int countDeletedEntries() throws IOException 
{
	myDeleted = database.getDeletedEntries();
	return myDeleted.length;
}

private long[] myUnsynced, myDeleted;
private FoundEntries sortedByOid;

private FoundEntries remoteGetByOid() throws IOException
{
	if (sortedByOid == null) sortedByOid = getByOID(null);
	return sortedByOid;
}
public int countUnsynchronizedEntries() throws IOException 
{
	EntriesView ev = getUnsynchronized(null);
	int s = ev.size();
	myUnsynced = new long[s];
	DatabaseEntry de = ev.getNew();
	for (int i = 0; i<s; i++){
		ev.get(i,de);
		myUnsynced[i] = getOID(de);
	}
	return s;
}
public void eraseDeleted(int[] indexesToDelete, int offset, int length)
		throws IOException {
	for (int i = 0; i<length; i++){
		long oid = myDeleted[indexesToDelete[offset+i]];
		database.eraseDeletedEntry(oid);
	}
}
public void eraseEntries(long[] oids, int offset, int length)
		throws IOException {
	FoundEntries sbo = remoteGetByOid();
	if (sbo == null) return;
	for (int i = 0; i<length; i++){
		long oid = oids[offset+i];
		int where = findByOID(sbo, oid);
		if (where != -1) sbo.erase(where);
	}
	
}
public long[] getDeletedEntries(int startIndex, int count) throws IOException {
	int need = count;
	int left = myDeleted.length-startIndex; 
	if (count > left) count = left;
	if (count <= 0) return new long[0];
	long[] ret = new long[count];
	System.arraycopy(myDeleted, startIndex, ret, 0, count);
	return ret;
}
public Vector getUnsynchronizedEntries(int startIndex, int count)
		throws IOException {
	FoundEntries sbo = remoteGetByOid();
	if (sbo == null) return new Vector();
	int need = count;
	int left = myUnsynced.length-startIndex; 
	if (count > left) count = left;
	if (count <= 0) return new Vector();
	Vector v = new Vector();
	for (int i = 0; i<count; i++){
		int where = findByOID(sbo, myUnsynced[startIndex+i]);
		if (where == -1) {
			v.add(null);
			continue;
		}
		DatabaseEntry de = sbo.get(where);
		v.add(de.encode());
	}
	return v;
}
public void markAsSynchronized(int[] indexes, int offset, int length)
		throws IOException {
	FoundEntries sbo = remoteGetByOid();
	if (sbo == null) return;
	for (int i = 0; i<length; i++){
		int where = findByOID(sbo, myUnsynced[indexes[offset+i]]);
		if (where == -1) continue;
		DatabaseEntry de = sbo.get(where);
		markAsSynchronized(de);
		sbo.store(where, de);
	}
}

/*
public static void main(String args[])throws IOException
{
	ewe.sys.Vm.startEwe(args);
	Synchronizer sync = new Synchronizer(DataStore.openDatabase(args[0],"rw"));
	EntriesView view = sync.getUnsynchronized();
	for (int i = 0; i<fe.size(); i++){
		Object got = fe.getData(i);
		Vm.debug(got.toString());
		sync.markAsSynchronized(view,i);
	}
	Vm.debug("Done.");
	ewe.sys.Vm.exit(0);
}
*/
//##################################################################
}
//##################################################################


