/*********************************************************************************
 *  Ewe Virtual Machine - Version 1.14, January 24, 2002                         *
 *  Copyright (C) 1999-2002 Michael L Brereton <michael_brereton@ewesoft.com>    *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>                         *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is free software; you can redistribute      *
 *  it and/or modify it under the terms of the Amended GNU Lesser General        *
 *  Public License distributed with this software.                               *
 *                                                                               *
 *  Under this license, linking this library or part thereof with other files to *
 *  produce an executable does not in itself require the executable to be        *
 *  covered by the GNU Lesser General Public License.                            *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, please download it at: www.ewesoft.com/LGPL.txt                      *
 *                                                                               *
 *********************************************************************************/
package eve.fx;

import java.awt.GraphicsEnvironment;

import eve.fx.gui.WindowSurface;
import eve.sys.Convert;
import eve.sys.Vm;

/**
 * Font is the character font used when drawing text on a surface.
 * <p>
 * Fonts have a name, such as "Helvetica", a style and a point size (usually
 * around 10). It's important to note that many devices have an extremely
 * limited number of fonts. For example, most PalmPilot devices have only
 * two fonts: plain and bold. If the font specified can't be found during
 * drawing, the closest matching font will be used.
 * <p>
 * Here is an example showing text being drawn in a given font:
 *
 * <pre>
 * ...
 * Font font = new Font("Helvetica", Font.BOLD, 10);
 * }
 *
 * public void onPaint(Graphics g)
 * {
 * g.setFont(font);
 * g.drawText("Hello", 10, 10);
 * ...
 * </pre>
 */

public class Font
{
String name;
int style;
int size;

/** A plain font style. */
public static final int PLAIN  = 0;
/** A bold font style. */
public static final int BOLD   = 0x1;
/** An italic font style. */
public static final int ITALIC = 0x2;
/** An underlined font style. */
public static final int UNDERLINE = 0x4;
/** An outlined font style. */
public static final int OUTLINE = 0x8;
/** A strikethrough font style. */
public static final int STRIKETHROUGH = 0x10;
/** A superscript font style. */
public static final int SUPERSCRIPT = 0x80;
/** A subscript font style. */
public static final int SUBSCRIPT = 0x100;

/**
  * Creates a font of the given name, style and size. Font styles are defined
  * in this class.
  * @see #PLAIN
  * @see #BOLD
  * @see Graphics
  */
public Font(String name, int style, int size)
	{
	if (name.toLowerCase().endsWith(" new")) name = name.substring(0,name.length()-4);
	this.name = name;
	this.style = style;
	this.size = size;
	}

protected static double fontScale = 0;
protected java.awt.Font _getAWTFont()
	{
	if (fontScale <= 0){
		String got = Vm.getProperty("java.font.scale",null);
		if (got == null) got = "1.0";
		fontScale = Convert.toDouble(got);
		if (fontScale < 0.5) fontScale = 0.5;
		if (fontScale >3) fontScale = 3;
	}
	int style = 0;
	if (this.style == PLAIN)
		style = java.awt.Font.PLAIN;
	if ((this.style & BOLD) != 0) style |= java.awt.Font.BOLD;
	if ((this.style & ITALIC) != 0) style |= java.awt.Font.ITALIC;
	//if ((this.style & UNDERLINE) != 0) style |= java.awt.Font.UNDERLINE;
	return new java.awt.Font(name, style,(int)(this.size*fontScale));//this.size - 3);
	}

/** Returns the name of the font. */
public String getName()
	{
	return name;
	}

/** Returns the size of the font. */
public int getSize()
	{
	return size;
	}

/**
 * Returns the style of the font. Font styles are defined in this class.
 * @see #PLAIN
 * @see #BOLD
 */
public int getStyle()
	{
	return style;
	}
/**
 * Return a new Font with the same name and size, but a different style.
 * @param newStyle The new style for the Font.
 * @return A new Font (or the same font if the style is the same).
 */
//===================================================================
public Font changeStyle(int newStyle)
//===================================================================
{
	if (newStyle == style) return this;
	return new Font(name,newStyle,size);
}
/**
 * Change the name and/or size of a Font.
 * @param newName The newName for the Font. If it is null the old name is used.
 * @param newSize The newSize for the Font. If it is <= 0, the old size is used.
 * @return A new Font.
 */
//===================================================================
public Font changeNameAndSize(String newName,int newSize)
//===================================================================
{
	if (newName == null) newName = name;
	if (newSize <= 0) newSize = size;
	return new Font(newName,style,newSize);
}

public static String [] listFonts(ISurface s)
{
	return GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
}

public String toString()
{
	return name+":"+size+", "+style;
}
public int hashCode()
{
	return name.hashCode()*size*style;
}
/**
 * Return if this Font is the same as another. This is quicker than equals()
 * since no class cast is involved.
 * @param f another Font.
 * @return true if its name, size and style are the same as this one, false if not.
 */
public boolean isSame(Font f)
{
	if (f == null) return false;
	if (f == this) return true;
	return name.equals(f.name) && size == f.size && style == f.style;
}
public boolean equals(Object obj)
{
	if (obj instanceof Font) return isSame((Font)obj);
	return super.equals(obj);
}
/** For use with getSystemFont() **/
public final static int SYSTEM_FONT_DEFAULT = 1;
public final static int SYSTEM_FONT_GUI = 2;
public final static int SYSTEM_FONT_MONOSPACED = 3;
public final static int SYSTEM_FONT_SERIF = 4;
public final static int SYSTEM_FONT_SANS_SERIF = 5;

private static native Font loadSystemFont(int which);

public static Font getSystemFont(int which)
{
	if (Vm.javaEveLibraryLoaded()) try{
		Font got = loadSystemFont(which);
		return got;
	}catch(Throwable t){
		return null;
	}else
		return null;
}

}
