/* java.math.BigDecimal -- Arbitrary precision decimals.
   Copyright (C) 1999, 2000, 2001 Free Software Foundation, Inc.

This file is part of GNU Classpath.

GNU Classpath is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.
 
GNU Classpath is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Classpath; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version. */

package eve.math;
import java.lang.reflect.Field;

import eve.nativeaccess.NativeAccess;
import eve.sys.Convert;
import eve.sys.LocaleFormat;
import eve.sys.Vm;
import eve.util.CharArray;
import eve.util.Utils;
import eve.util.ByteEncodable;
import eve.util.ByteArray;
//import java.math.BigInteger;
/**
 * This version of BigDecimal is similar to java.math.BigDecimal and is also Immutable, but is ByteEncodable.<p>
 * Wherever possible, you should use this version of BigDecimal since it is compatible
 * with Decimal - a mutable version of BigDecimal.
 * @author Michael L Brereton
 *
 */
public class BigDecimal extends Number implements Comparable, ByteEncodable
{
  private BigInteger intVal;
  private int scale;
  private static final long serialVersionUID = 6108874887143696463L;

  private final static BigDecimal ZERO = 
    new BigDecimal (BigInteger.valueOf (0), 0);

  private final static BigDecimal ONE = 
    new BigDecimal (BigInteger.valueOf (1), 0);

  public final static int ROUND_UP = 0;
  public final static int ROUND_DOWN = 1;
  public final static int ROUND_CEILING = 2;
  public final static int ROUND_FLOOR = 3;
  public final static int ROUND_HALF_UP = 4;
  public final static int ROUND_HALF_DOWN = 5;
  public final static int ROUND_HALF_EVEN = 6;
  public final static int ROUND_UNNECESSARY = 7;

	private BigDecimal()
	{
	}
	
  public BigDecimal (BigInteger num) 
  {
    this (num, 0);
  }

  public BigDecimal (BigInteger num, int scale) throws NumberFormatException 
  {
    if (scale < 0) 
      throw new NumberFormatException ("scale of " + scale + " is < 0");
    this.intVal = num;
    this.scale = scale;
  }

  public BigDecimal (double num) throws NumberFormatException 
  {
    if (Double.isInfinite (num) || Double.isNaN (num))
      throw new NumberFormatException ("invalid argument: " + num);
    // Note we can't convert NUM to a String and then use the
    // String-based constructor.  The BigDecimal documentation makes
    // it clear that the two constructors work differently.

    final int mantissaBits = 52;
    final int exponentBits = 11;
    final long mantMask = (1L << mantissaBits) - 1;
    final long expMask = (1L << exponentBits) - 1;

    long bits = Convert.toLongBitwise(num);
    long mantissa = bits & mantMask;
    long exponent = (bits >>> mantissaBits) & expMask;
    boolean denormal = exponent == 0;
    // Correct the exponent for the bias.
    exponent -= denormal ? 1022 : 1023;
    // Now correct the exponent to account for the bits to the right
    // of the decimal.
    exponent -= mantissaBits;
    // Ordinary numbers have an implied leading `1' bit.
    if (! denormal)
      mantissa |= (1L << mantissaBits);

    // Shave off factors of 10.
    while (exponent < 0 && (mantissa & 1) == 0)
      {
	++exponent;
	mantissa >>= 1;
      }

    intVal = BigInteger.valueOf (bits < 0 ? - mantissa : mantissa);
    if (exponent < 0)
      {
	// We have MANTISSA * 2 ^ (EXPONENT).
	// Since (1/2)^N == 5^N * 10^-N we can easily convert this
	// into a power of 10.
	scale = (int) (- exponent);
	BigInteger mult = BigInteger.valueOf (5).pow (scale);
	intVal = intVal.multiply (mult);
      }
    else
      {
	intVal = intVal.shiftLeft ((int) exponent);
	scale = 0;
      }
  }

  
  BigDecimal (String num, LocaleFormat locale, boolean isCurrency)
  throws NumberFormatException
  {
	  if (num == null) throw new NullPointerException();
	  if ((locale.options & LocaleFormat.OPTION_IS_CURRENCY) != 0) isCurrency = true;
	  FormattedDecimal fd = FormattedDecimal.getCached();
	  try{
		  fd.formattedNumber.append(num);
		  DecimalSymbols ds = DecimalSymbols.getFor(locale.locale, isCurrency);
		  fd.parse(ds).reformatAfterParse();
		  fromNum(fd.formattedNumber.toString());
	  //}catch(Exception e){e.printStackTrace(); fromNum("0");
	  }finally{
		  fd.cache();
	  }
  }
  public BigDecimal (String num) throws NumberFormatException
  {
	  fromNum(num);
  }
  private void fromNum(String num) throws NumberFormatException
  {
	//System.out.println("From: "+num);
    int len = num.length();
    int start = 0, point = 0;
    int dot = -1;
    boolean negative = false;
    if (num.charAt(0) == '+')
      {
	++start;
	++point;
      }
    else if (num.charAt(0) == '-')
      {
	++start;
	++point;
	negative = true;
      }

    while (point < len)
      {
	char c = num.charAt (point);
	if (c == '.')
	  {
	    if (dot >= 0)
	      throw new NumberFormatException ("multiple `.'s in number");
	    dot = point;
	  }
	else if (c == 'e' || c == 'E')
	  break;
	else if (BigInteger.digit (c, 10) < 0)
	  throw new NumberFormatException ("unrecognized character: " + c);
	++point;
      }

    String val;
    if (dot >= 0)
      {
	val = num.substring (start, dot) + num.substring (dot + 1, point);
	scale = point - 1 - dot;
      }
    else
      {
	val = num.substring (start, point);
	scale = 0;
      }
    if (val.length () == 0)
      throw new NumberFormatException ("no digits seen");

    if (negative)
      val = "-" + val;
    intVal = new BigInteger (val);

    // Now parse exponent.
    if (point < len)
      {
        point++;
        if (num.charAt(point) == '+')
          point++;

        if (point >= len )
          throw new NumberFormatException ("no exponent following e or E");
	
        try 
	  {
	    int exp = Convert.toInt(num.substring (point));//Integer.parseInt (num.substring (point));
	    exp -= scale;
	    if (signum () == 0)
	      scale = 0;
	    else if (exp > 0)
	      {
		intVal = intVal.multiply (BigInteger.valueOf (10).pow (exp));
		scale = 0;
	      }
	    else
	      scale = - exp;
	  }
        catch (NumberFormatException ex) 
	  {
	    throw new NumberFormatException ("malformed exponent");
	  }
      }
  }

  public static BigDecimal valueOf (long val) 
  {
    return valueOf (val, 0);
  }

  public static BigDecimal valueOf (long val, int scale) 
    throws NumberFormatException 
  {
	
    if ((scale == 0) && ((int)val == val))
      switch ((int) val)
	{
	case 0:
	  return ZERO;
	case 1:
	  return ONE;
	}
  
    return new BigDecimal (BigInteger.valueOf (val), scale);
  }

  private static Field eveBigDecimal;
  private static boolean hasNative = true;
  
  /**
   * Convert a java.math.BigDecimal into an eve.math.BigDecimal as effeciently as possible. Do not
   * attempt to modify the returned BigDecimal or this may modify the original java.math.BigDecimal
 * @param jbd the java.math.BigDecimal value.
 * @return a eve.math.BigDecimal.
 */
public synchronized static BigDecimal toEveBigDecimal(java.math.BigDecimal jbd)
  {
  	try{
  		if (eveBigDecimal == null && hasNative) 
  			eveBigDecimal = NativeAccess.getNativeField(java.math.BigDecimal.class, "bd");
  		if (eveBigDecimal == null) throw new Exception();
  		return (BigDecimal)eveBigDecimal.get(jbd);
  	}catch(Throwable t){
  		hasNative = false;
  		return new BigDecimal(jbd.toString());
  	}
  }
/**
 * Convert a BigDecimal into a java.math.BigDecimal as effeciently as possible.
* @param ebd the BigDecimal value. 
* @return a java.math.BigDecimal.
*/
public synchronized static java.math.BigDecimal toJavaBigDecimal(BigDecimal ebd)
{
	if (ebd == null) throw new NullPointerException();
	try{
		if (eveBigDecimal == null && hasNative) 
			eveBigDecimal = NativeAccess.getNativeField(java.math.BigDecimal.class, "bd");
  		if (eveBigDecimal == null) throw new Exception();
		java.math.BigDecimal jbd = (java.math.BigDecimal)NativeAccess.createNewObject(java.math.BigDecimal.class, false);
		eveBigDecimal.set(jbd,ebd);
		return jbd;
	}catch(Throwable t){
		hasNative = false;
		return new java.math.BigDecimal(ebd.toString());
	}
}
  
  public BigDecimal add (BigDecimal val) 
  {
    // For addition, need to line up decimals.  Note that the movePointRight
    // method cannot be used for this as it might return a BigDecimal with
    // scale == 0 instead of the scale we need.
    BigInteger op1 = intVal;
    BigInteger op2 = val.intVal;
    if (scale < val.scale)
      op1 = op1.multiply (BigInteger.valueOf (10).pow (val.scale - scale));
    else if (scale > val.scale)
      op2 = op2.multiply (BigInteger.valueOf (10).pow (scale - val.scale));

    return new BigDecimal (op1.add (op2), java.lang.Math.max (scale, val.scale));
  }

  public BigDecimal subtract (BigDecimal val) 
  {
    return this.add(val.negate());
  }

  public BigDecimal multiply (BigDecimal val) 
  {
    return new BigDecimal (intVal.multiply (val.intVal), scale + val.scale);
  }

  public BigDecimal divide (BigDecimal val, int roundingMode) 
    throws ArithmeticException, IllegalArgumentException 
  {
    return divide (val, scale, roundingMode);
  }

  public BigDecimal divide(BigDecimal val, int newScale, int roundingMode)
    throws ArithmeticException, IllegalArgumentException 
  {
    if (roundingMode < 0 || roundingMode > 7)
      throw 
	new IllegalArgumentException("illegal rounding mode: " + roundingMode);

    if (newScale < 0)
      throw new ArithmeticException ("scale is negative: " + newScale);

    if (intVal.signum () == 0)	// handle special case of 0.0/0.0
      return newScale == 0 ? ZERO : new BigDecimal (ZERO.intVal, newScale);
    
    // Ensure that pow gets a non-negative value.
    int valScale = val.scale;
    BigInteger valIntVal = val.intVal;
    int power = newScale + 1 - (scale - val.scale);
    if (power < 0)
      {
	// Effectively increase the scale of val to avoid an
	// ArithmeticException for a negative power.
        valIntVal = valIntVal.multiply (BigInteger.valueOf (10).pow (-power));
	power = 0;
      }

    BigInteger dividend = intVal.multiply (BigInteger.valueOf (10).pow (power));
    
    BigInteger parts[] = dividend.divideAndRemainder (valIntVal);
//      System.out.println("int: " + parts[0]);
//      System.out.println("rem: " + parts[1]);

    int roundDigit = parts[0].mod (BigInteger.valueOf (10)).intValue ();
    BigInteger unrounded = parts[0].divide (BigInteger.valueOf (10));

    if (roundDigit == 0 && parts[1].signum () == 0) // no rounding necessary
      return new BigDecimal (unrounded, newScale);

    int sign = unrounded.signum ();

    switch (roundingMode)
      {
      case ROUND_UNNECESSARY:
	throw new ArithmeticException ("newScale is not large enough");
      case ROUND_CEILING:
	roundingMode = (sign == 1) ? ROUND_UP : ROUND_DOWN;
	break;
      case ROUND_FLOOR:
	roundingMode = (sign == 1) ? ROUND_DOWN : ROUND_UP;
	break;
      case ROUND_HALF_UP:
	roundingMode = (roundDigit >= 5) ? ROUND_UP : ROUND_DOWN;
	break;
      case ROUND_HALF_DOWN:
	roundingMode = (roundDigit > 5) ? ROUND_UP : ROUND_DOWN;
	break;
      case ROUND_HALF_EVEN:
	if (roundDigit < 5)
	  roundingMode = ROUND_DOWN;
	else
	  {
	    int rightmost = 
	      unrounded.mod (BigInteger.valueOf (10)).intValue ();
	    if (rightmost % 2 == 1) // odd, then ROUND_HALF_UP
	      roundingMode = ROUND_UP;
	    else // even, then ROUND_HALF_DOWN
	      roundingMode = (roundDigit > 5) ? ROUND_UP : ROUND_DOWN;
	  }
	break;
      }

    if (roundingMode == ROUND_UP)
      return new BigDecimal (unrounded.add (BigInteger.valueOf (1)), newScale);

    // roundingMode == ROUND_DOWN
    return new BigDecimal (unrounded, newScale);
  }
    
  public int compareTo (BigDecimal val) 
  {
    if (scale == val.scale)
      return intVal.compareTo (val.intVal);

    BigInteger thisParts[] = 
      intVal.divideAndRemainder (BigInteger.valueOf (10).pow (scale));
    BigInteger valParts[] =
      val.intVal.divideAndRemainder (BigInteger.valueOf (10).pow (val.scale));
    
    int compare;
    if ((compare = thisParts[0].compareTo (valParts[0])) != 0)
      return compare;

    // quotients are the same, so compare remainders

    // remove trailing zeros
    if (thisParts[1].equals (BigInteger.valueOf (0)) == false)
      while (thisParts[1].mod (BigInteger.valueOf (10)).equals
	     (BigInteger.valueOf (0)))
      thisParts[1] = thisParts[1].divide (BigInteger.valueOf (10));
    // again...
    if (valParts[1].equals(BigInteger.valueOf (0)) == false)
      while (valParts[1].mod (BigInteger.valueOf (10)).equals
	     (BigInteger.valueOf (0)))
	valParts[1] = valParts[1].divide (BigInteger.valueOf (10));

    // and compare them
    return thisParts[1].compareTo (valParts[1]);
  }

  public int compareTo (Object val) 
  {
    return(compareTo((BigDecimal)val));
  }

  public boolean equals (Object o) 
  {
    return (o instanceof BigDecimal 
	    && scale == ((BigDecimal) o).scale
	    && compareTo ((BigDecimal) o) == 0);
  }

  public int hashCode() 
  {
    return intValue() ^ scale;
  }

  public BigDecimal max (BigDecimal val)
  {
    switch (compareTo (val)) 
      {
      case 1:
	return this;
      default:
	return val;
      }
  }

  public BigDecimal min (BigDecimal val) 
  {
    switch (compareTo (val)) 
      {
      case -1:
	return this;
      default:
	return val;
      }
  }

  public BigDecimal movePointLeft (int n)
  {
    return (n < 0) ? movePointRight (-n) : new BigDecimal (intVal, scale + n);
  }

  public BigDecimal movePointRight (int n)
  {
    if (n < 0)
      return movePointLeft (-n);

    if (scale >= n)
      return new BigDecimal (intVal, scale - n);

    return new BigDecimal (intVal.multiply 
			   (BigInteger.valueOf (10).pow (n - scale)), 0);
  }

  public int signum () 
  {
    return intVal.signum ();
  }

  public int scale () 
  {
    return scale;
  }
  
  public BigInteger unscaledValue()
  {
    return intVal;
  }

  public BigDecimal abs () 
  {
    return new BigDecimal (intVal.abs (), scale);
  }

  public BigDecimal negate () 
  {
    return new BigDecimal (intVal.negate (), scale);
  }

  	
  public CharArray toString(CharArray ca)
  {
  	if (ca == null) ca = new CharArray();
  	return ca.append(toString());
  }
  
    public String toString () 
    {
		if (intVal == null) return "0";
    String bigStr = intVal.toString();
    if (scale == 0) 
      return bigStr;

    boolean negative = (bigStr.charAt(0) == '-');

    int point = bigStr.length() - scale - (negative ? 1 : 0);

    StringBuffer sb = new StringBuffer(bigStr.length() + 2 +
				       (point <= 0 ? (-point + 1) : 0));
    if (point <= 0)
      {
        if (negative)
          sb.append('-');
        sb.append('0').append('.');
        while (point < 0)
          {
            sb.append('0');
            point++;
          }
        sb.append(bigStr.substring(negative ? 1 : 0));
      }
    else
      {
	sb.append(bigStr);
	String str = sb.toString();
	char [] all = Vm.getStringChars(str);
	sb.setLength(0);
	int p = point+(negative?1:0);
	if (p != 0) sb.append(new String(all,0,p));
	sb.append('.');
	sb.append(new String(all,p,all.length-p));
	//sb.insert(point + (negative ? 1 : 0), '.');
      }
    return sb.toString();
  }
    String format(LocaleFormat locale, Decimal value)
    {
      boolean isCurrency = value.isCurrency; 
  	  //if (locale.isCurrency) isCurrency = true;
  	  DecimalSymbols ds = DecimalSymbols.getFor(locale.locale, isCurrency);
  	  if (intVal == null) return FormattedDecimal.format(ds,"0",null,false,locale,value.localeFormatOptionsToSet,value.localeFormatOptionsToClear);
	  String bigStr = intVal.toString();
	  boolean negative = (bigStr.charAt(0) == '-');
	  if (negative) bigStr = bigStr.substring(1);
	  int len = bigStr.length();
	  if (scale == 0) return FormattedDecimal.format(ds,bigStr, null, negative, locale,value.localeFormatOptionsToSet,value.localeFormatOptionsToClear);
	  String frac = "", integer = "";
	  if (scale >= len){
		  integer = "0";
		  frac = bigStr;
		  int toAdd = scale-len;
		  for (int i = 0; i<toAdd; i++) frac = '0'+frac;
	  }else{
		  integer = bigStr.substring(0,len-scale);
		  frac = bigStr.substring(len-scale);
	  }
	  return FormattedDecimal.format(ds,integer, frac, negative, locale,value.localeFormatOptionsToSet,value.localeFormatOptionsToClear);
    }

  public BigInteger toBigInteger () 
  {
    return scale == 0 ? intVal :
      intVal.divide (BigInteger.valueOf (10).pow (scale));
  }

  public int intValue () 
  {
    return toBigInteger ().intValue ();
  }

  public long longValue ()
  {
    return toBigInteger().longValue();
  }

  public float floatValue() 
  {
   // return Float.valueOf(toString()).floatValue();
		return Convert.toFloat(toString());
  }

  public double doubleValue() 
  {
    return Convert.toDouble(toString());//Double.valueOf(toString()).doubleValue();
  }

  public BigDecimal setScale (int scale) throws ArithmeticException
  {
    return setScale (scale, ROUND_UNNECESSARY);
  }

  public BigDecimal setScale (int scale, int roundingMode)
    throws ArithmeticException, IllegalArgumentException
  {
    return divide (ONE, scale, roundingMode);
  }
	/**
	 * Write the BigDecimal as a series of bytes, or find out how many bytes are needed 
	 * for storing the BigDecimal.
	 * @param data The destination byte array or null to find out how many bytes are needed.
	 * @param offset The offset in the byte array to write to.
	 * @return The number of bytes written or needed to be written.
	 */
	//===================================================================
	public int write(byte[] data,int offset)
	//===================================================================
	{
		int count = 4;
		if (data == null)
			return count+(intVal == null ? 0 : intVal.write(null,0));
		Utils.writeInt(scale,data,offset,4); offset += 4;
		if (intVal != null) count += intVal.write(data,offset);
		return count;
	}
	//===================================================================
	public int encodeBytes(ByteArray dest)
	//===================================================================
	{
		int need = write(null,0);
		if (dest == null) return need;
		dest.makeSpace(dest.length,need);
		return write(dest.data,dest.length-need);
	}
	/**
	 * Find out how many bytes constitute an encoded BigDecimal.
	 * @param data the data to read.
	 * @param offset the start of the data.
	 * @return the number of bytes representing the encoded BigDecimal.
	 */
		/*
	//===================================================================
	public static int sizeOfEncoded(byte[] data,int offset)
	//===================================================================
	{
		return 5 + (data[offset+4] != 0 ? BigInteger.sizeOfEncoded(data,offset+5) : 0);
	}
	*/
	//-------------------------------------------------------------------
	private void readFrom(byte[] data,int offset,int length)
	//-------------------------------------------------------------------
	{
		scale = Utils.readInt(data,offset,4); offset += 4;
		intVal = length > 4 ? new BigInteger(data,offset,length-4) : null;
	}
	/**
	 * Create a new BigDecimal from a byte array.
	 * @param data the data to read.
	 * @param offset the start of the data.
	 * @param length the number of bytes of encoded data.
	 */
	//===================================================================
	public BigDecimal(byte[] data,int offset,int length)
	//===================================================================
	{
		readFrom(data,offset,length);
	}
	//===================================================================
	public BigDecimal(byte[] data)
	//===================================================================
	{
		this(data,0,data.length);
	}
	//===================================================================
	public byte[] toByteArray()
	//===================================================================
	{
		int need = write(null,0);
		byte[] ret = new byte[need];
		write(ret,0);
		return ret;
	}
	//-------------------------------------------------------------------
	static BigDecimal c1 = new BigDecimal(), c2 = new BigDecimal();
	//-------------------------------------------------------------------
	/**
	 * Compare two encoded versions of a BigDecimal as fast as possible. 
	 */
	//===================================================================
	public static synchronized int compareEncoded(byte[] one, int oneOffset, int oneLen, byte[] two, int twoOffset,int twoLen)
	//===================================================================
	{
		BigDecimal o = c1, t = c2;
		if (oneLen <= 0) o = null;
		else o.readFrom(one,oneOffset,oneLen);
		if (twoLen <= 0) t = null;
		else t.readFrom(two,twoOffset,twoLen);
		if (o == t) return 0;
		if (o == null) return -1;
		if (t == null) return 1;
		return o.compareTo(t);
	}
	
}

