/*
 * Created on Jul 31, 2005
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui;

import java.util.Iterator;
import java.util.Stack;
import java.util.Vector;

import eve.fx.BackgroundImage;
import eve.fx.BufferedGraphics;
import eve.fx.Color;
import eve.fx.Dimension;
import eve.fx.Drawable;
import eve.fx.Font;
import eve.fx.FontMetrics;
import eve.fx.Graphics;
import eve.fx.IImage;
import eve.fx.ISurface;
import eve.fx.Image;
import eve.fx.ImageRefresher;
import eve.fx.Metrics;
import eve.fx.PixelBuffer;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.Sound;
import eve.fx.gui.Cursor;
import eve.fx.gui.IKeys;
import eve.fx.gui.SurfaceKeyEvent;
import eve.fx.gui.SurfacePointerEvent;
import eve.fx.gui.SurfaceTextEvent;
import eve.fx.gui.WindowSurface;
import eve.sys.Cache;
import eve.sys.DataInterface;
import eve.sys.DeferredEvent;
import eve.sys.Device;
import eve.sys.Event;
import eve.sys.EventListener;
import eve.sys.FieldTransfer;
import eve.sys.Gate;
import eve.sys.Handle;
import eve.sys.ImageData;
import eve.sys.ReadWriteLock;
import eve.sys.TimerProc;
import eve.sys.Vm;
import eve.sys.mThread;
import eve.ui.event.ControlEvent;
import eve.ui.event.DataChangeEvent;
import eve.ui.event.KeyEvent;
import eve.ui.event.MenuEvent;
import eve.ui.event.PenEvent;
import eve.util.CharArray;
import eve.util.TagList;
import eve.util.mString;
/**
 * @author Michael L Brereton
 *
 */
//####################################################
public class Control extends ControlBase implements UIConstants, EventListener,  ImageRefresher, Drawable, DataInterface, TimerProc{

	/** The control's x location */
	protected int x;
	/** The control's y location */
	protected int y;
	/** The control's width */
	protected int width;
	/** The control's height */
	protected int height;
	/** The parent of the control. */
	protected Container parent;
	/** The control's next sibling. */
	protected Control next;
	/** The control's previous sibling. */
	protected Control prev;
	/** The children of the container. */
	protected Control children;
	/** The tail of the children list. */
	protected Control tail;
	/** Modifiers of the control. */
	protected int modifiers = 0;
//	...........................................
//	 Variables after this may be moved.
//	...........................................
	/**
	* This is the "HotKey" for the control - use setHotKey() to set it.
	**/
	public int hotKey = 0;
	/**
	* The border style of the Control. Should be one of the BDR_XXX or EDGE_XXX constants.
	**/
	public int borderStyle = 0;
	/**
	* The Color of the border of the Control (if a single line, non-3D border is used).
	**/
	public Color borderColor = Color.DarkGray;
	
	public int borderWidth = 0;
	
	public int penStatus = 0;
	/**
	 * If this is true, then the repaintNow()/paintDataNow() methods will invoke
	 * a readLock() on the Control before calling doPaint() or doPaintData().
	 */
	public boolean lockPaint;
	//public float alpha = 1.0f;
	public int rows = 1;
	public int columns = 1;
	public String name = unnamed;
	public String text = emptyString;
	/**
	* This is an optional string that is used as a user informative prompt for the control.
	**/
	public String prompt = emptyString;
	/**
	* This is optional - if an mLabel is used as the on-screen prompt for this control, then set this value
	* to be that control. This ensures that if this control gets the focus and its container
	* must scroll to make it visible, it will also attempt to make the prompt visible as well.
	**/
	public Control promptControl = null;
	/**
	* The minimum number of pixels that the Pen/Mouse has to move during a drag operation to
	* be considered a drag to a new location.
	**/
	public int dragResolution = 3;
	/**
	* The number of pixels that the Pen/Mouse has to move before it is considered the start
	* of a drag operation.
	**/
	public int startDragResolution = 3;

	private ReadWriteLock lock;
	
public synchronized ReadWriteLock getLock()
{
	if (lock == null) lock = new ReadWriteLock();
	return lock;
}

public void readLock()
{
	getLock().read();
}
public void writeLock()
{
	getLock().write();
}
public void unlock()
{
	getLock().unlock();
}
public String toString() {return (name == unnamed ? mString.rightOf(getClass().getName(),'.') : name)+(text != null && text.length() != 0 ? (" = "+text) : "");}
/*###############################################################
 * 
 * Prompt, Text and Focus Control
 * 
 ###############################################################*/
/**
 * This is called when the Control gets the focus. The default version
 * of this does nothing.
 * @param how one of the ByXXX values.
 */
public void gotFocus(int how)
{
	
}
/**
 * This is called when the Control loses the focus. The default version
 * of this does nothing.
 * @param how one of the ByXXX values.
 */
public void lostFocus(int how)
{
	
}
/**
 * Set the text of the Control, automatically repainting it.
 */
public void setText(String what) 
{
	text = what; 
	if (text == null) text = ""; 
	repaintNow();
}
/**
 * Get the text of the Control.
 * @return
 */
public String getText() 
{
	return text;
}
/**
If this Control represents the input of a password that must be kept
hidden, then this method should return a password character (usually a '*').
If not then this should return (char)0.
**/
//===================================================================
public char getPasswordCharacter()
//===================================================================
{
	return 0;
}

/**
 * Get the text to display for the Control (which may be different
 * to what is considered the text of the control).
By default getDisplayText() returns getText();
*/
public String getDisplayText() 
{
	return getText();
}
/**
 * Get the prompt for the Control if any is set.
 */
public String getPrompt()
{
	for (Control c = this;c != null;c = c.parent)
		if (c.prompt != null)
			if (c.prompt.length() != 0)
				return c.prompt;
	return emptyString;
}
/**
* This tells the control to consider its current data to be unchanged.
**/
//===================================================================
public void updateData()
//===================================================================
{
	Iterator it = getSubControls((ControlIterator)Cache.get(ControlIterator.class));
	while(it.hasNext()){
		Object obj = it.next();
		if (obj instanceof Control)
			((Control)obj).updateData();
	}
	Cache.put(it);
}

/**
 * This returns either the value of prompt, if it is not null and not an empty string, or
 * the text value of promptControl, if that is not null and not an empty string. If no non-empty
 * string prompt can be found, it will return null.
 */
//===================================================================
public String getPromptText()
//===================================================================
{
	if (prompt != null && prompt.length() != 0) return prompt;
	if (promptControl == null) return null;
	String p = promptControl.getText();
	if (p != null && p.length() != 0) return p;
	return null;
}

/**
 * On a normal Control this will set the promptControl variable to "prompt", but
 * on a Container, the first child non-container will have "prompt" assigned to it.
 * @param prompt the Control acting as the prompt (usually an mLabel).
 */
public void setPromptControl(Control prompt)
{
	takePromptControl(prompt);
}

protected boolean takePromptControl(Control prompt)
{
	this.promptControl = prompt;
	return true;
}
/**
 * Set the focus on this Control.
 * @param how - one of the ControlConstant.ByXXX values.
 */
//===================================================================
public void takeFocus(int how)
//===================================================================
{
	Gui.takeFocus(this,how);
}
/**
* The object's tool tip which should be a String, IImage, Control or ToolTip
**/
public Object toolTip;
/**
* Set the tooltip of the object.
* @param tip the tip which should be a String, IImage, Control or ToolTip
* @return this Control.
*/
//===================================================================
public Control setToolTip(Object tip)
//===================================================================
{
	toolTip = tip;
	PenEvent.wantPenMoved(this,PenEvent.WANT_PEN_MOVED_ONOFF,true);
	return this;
}
/**
* This should return an acceptable ToolTip object. That is either a String, IImage, Control or ToolTip
* @param x The x location of the mouse in this control.
* @param y The y location of the mouse in this control.
* @return A suitable ToolTip or null.
*/
//===================================================================
public Object getToolTip(int x,int y)
//===================================================================
{
	return toolTip;	
}
//==================================================================
public final Dimension recalculatePreferredSize(Dimension dest) {modifyAll(0,CalculatedSizes,true); return getPreferredSize(dest);}
//==================================================================

/*###############################################################
 * 
 * Fonts and Colors
 * 
 ###############################################################*/

public Color backGround;
public Color foreGround;
public Font font;
public IImage image;

//===================================================================
public IImage getImage() {return image;}
//===================================================================

//===================================================================
public Font getFont()
//===================================================================
{
	//if (font == null) new Exception().printStackTrace();
	for (Control c = this; c != null; c = c.parent)
		if (c.font != null) return c.font;
	return Application.guiFont;
}
/**
 * Set the Font for the control.
 * @param f the Font to set.
 * @return this Control.
 */
//===================================================================
public Control setFont(Font f)
//===================================================================
{
	font = f;
	return this;
}

//==================================================================
public Color getBackground() 
//==================================================================
{
	for (Control c = this; c != null; c = c.parent)
		if (c.backGround != null) {
			return c.backGround;
		}
	if (Application.mainApp != null)
		if (Application.mainApp.backGround != null)
			return Application.mainApp.backGround;
	return Color.LightGray;
}
//==================================================================
public Color getForeground() 
//==================================================================
{
	for (Control c = this; c != null; c = c.parent)
		if (c.foreGround != null) return c.foreGround;
	return Color.Black;
}
/** Returns the font metrics for a given font. */
public FontMetrics getFontMetrics(Font font)
{
	return Application.mainApp.getFontMetrics(font);
}
/** Returns the font metrics for the Font for this Control. */
public FontMetrics getFontMetrics() 
{
	return getFontMetrics(getFont());
}

/*###############################################################
 * 
 * Modifier handling.
 * 
 ###############################################################*/
/**
This does NOT enable a disabled control - but tells it to put itself in an "active"
state. For example, a button in an active state looks as if the mouse/pen is being 
pressed down on it.<p>
* If you want to re-enable a disabled control
* do modify(0,Control.Disabled)
**/
//===================================================================
public void activate()
//===================================================================
{
}
/**
* This does NOT disable a control - but tells it to put itself in a "non-active" state.
* For example, a button that has is showing itself as being pressed, will show itself
* as not pressed when this is called.<p>
* If you want to disable a control
* do modify(Control.Disabled,0)
**/
//===================================================================
public void deactivate()
//===================================================================
{
}

/**
 * Switch on or off a modifier depending on the status parameter.
 * @param flag The flags to set on or off.
 * @param status if true the flags will be set on, if false they will be set off.
 * @return the return value from the modify() call.
 */
//==================================================================
public final int set(int flag,boolean status)
//==================================================================
{
	if (status) return modify(flag,0);
	else return modify(0,flag);
}
/**
 * Set/Clear modifier flags for the Control. Examples of flags are Disabled, NotEditable, Invisible... Check
 * ControlConstants for the complete list. The return value can be used to restore the 
 * modifiers to the original state like this:
	<pre>
	int old = myControl.modify(flagsSet,flagsClear);
	//do some processing...
	myControl.restore(old,flagsSet|flagsClear);
	</pre>
 * @param flagsToSet The flags to set.
 * @param flagsToClear The flags to clear.
 * @return A value that can be use to restore the modifier to its original state using
 * restore().
 */
//==================================================================
public int modify(int flagsToSet,int flagsToClear)
//==================================================================
{
	int ret = modifiers & (flagsToSet|flagsToClear);
	modifiers |= flagsToSet;
	modifiers &= ~flagsToClear;
	if (((modifiers & (Disabled|MouseSensitive)) == (Disabled|MouseSensitive)) 
	&& ((Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_NO_MOUSE_POINTER) == 0))
	penStatus &= ~PenIsOn;
	return ret;
}
/**
* Set/Clear modifier flags for the Control. Examples of flags are Disabled, NotEditable, Invisible... Check
* ControlConstants for the complete list. The return value can be used to restore the 
* modifiers to the original state like this:
	<pre>
	long old = myControl.modifyFlags(flagsSet,flagsClear);
	//do some processing...
	myControl.restoreFlags(old);
	</pre>
* @param flagsToSet The flags to set.
* @param flagsToClear The flags to clear.
* @return A value that can be use to restore the altered modifier bits to its original state using
* restoreFlags().
*/
public final long modifyFlags(int flagsToSet, int flagsToClear)
{
	long mask = flagsToSet|flagsToClear;
	long cf = modify(flagsToSet,flagsToClear);
	return (mask << 32) | (cf & 0xffffffffL);
}
/**
* Restore the modifier flags to their previous value. An example use:
	<pre>
	long old = myControl.modifyFlags(flagsSet,flagsClear);
	//do some processing...
	myControl.restoreFlags(old);
	</pre>
	@param oldValue the value returned by modifyFlags.
*/
public final void restoreFlags(long oldValue)
{
	long mask = (oldValue >> 32) & 0xffffffffL;
	long cf = oldValue & 0xffffffffL;
	long got = modifiers & 0xffffffffL;
	got &= ~mask;
	got |= cf;
	modifiers = (int)got;
}
/**
* Restore the modifier flags to their previous value. An example use:
	<pre>
	int old = myControl.modify(flagsSet,flagsClear);
	//do some processing...
	myControl.restore(old,flagsSet|flagsClear);
	</pre>
* @param oldValue the value returned by modify().
* @param mask the paramters flagsToSet OR'ed with flagsToClear as used in the modify() method.
*/
//==================================================================
public void restore(int oldValue,int mask)
//==================================================================
{
	modifiers = (modifiers &= ~mask)|oldValue;
}
/**
* This checks if the specified flags are set or clear.
* @param flagsSet Flags to check for being set.
* @param flagsClear Flags to check for being clear.
* @return true if all the flags are set/cleared as specified.
*/
//===================================================================
public boolean checkModifiers(int flagsSet,int flagsClear)
//===================================================================
{
	if ((modifiers & flagsSet) != flagsSet) return false;
	if (((~modifiers) & flagsClear) != flagsClear) return false;
	return true;
}
/**
* This ensures that the specified flags are set or clear. Return true
* if a change was made. Returns false if no change was made.
**/
//===================================================================
public boolean change(int flagsToSet,int flagsToClear)
//===================================================================
{
	if (checkModifiers(flagsToSet,flagsToClear)) return false;
	modify(flagsToSet,flagsToClear);
	return true;
}
/**
 * Check if the control has the specified modifier flags set.
 * @param what The modifier flags to check for.
 * @param shouldInherit if this is true then all parents of the control will also be checked.
 * @return true if the flag is set in this control or in any of its ancestors (if should inherit is true).
 */
//==================================================================
public boolean hasModifier(int what,boolean shouldInherit)
//==================================================================
{
	int got = getModifiers(shouldInherit);
	return ((got & what) == what);
}
/**
* This will get the modifiers for this control. If shouldInherit is true,
* then all the ancestor modifiers will be OR'ed together.
**/
//===================================================================
public int getModifiers(boolean shouldInherit)
//===================================================================
{
	int flags = modifiers;
	if (shouldInherit)
		for (Control c = parent; c != null; c = c.parent)
			flags |= c.modifiers;
	return flags;
}
/**
* Get particular modifiers flags from another control and optionally set children flags.
* @param fromWho The control to get from.
* @param inheritFromAll if this is true then also check the ancestors of the control for the flags.
* @param what the flags to get (or'ed together).
* @param setChildren if this is true then modify my children.
*/
//==================================================================
public final void inheritModifiers(Control fromWho,boolean inheritFromAll,int what,boolean setChildren)
//==================================================================
{
	if (fromWho == null) return;
	int toSet = fromWho.getModifiers(inheritFromAll) & what;
	int toClear = (~toSet) & what;
	if (setChildren) modifyAll(toSet,toClear,true);
	else modify(toSet,toClear);
}
/**
 * Modify this control and all child controls (and their children).
 * @param set The flags to set.
 * @param clear The flags to clear.
 */
//===================================================================
public void modifyAll(int set,int clear) {modifyAll(set,clear,true);}
//===================================================================
/**
 * Modify all child controls (and their children).
 * @param set The flags to set.
 * @param clear The flags to clear.
 * @param doThisOne if this is true then this control will also be modified otherwise
 * only the children will be.
 */
//===================================================================
public void modifyAll(int set,int clear,boolean doThisOne)
//===================================================================
{
	if (doThisOne) modify(set,clear);
	Iterator e = getCachedSubControls();
	for (;e.hasNext();){
		Object obj = e.next();
		if (obj instanceof Control)
			((Control)obj).modifyAll(set,clear,true);
	}
	Cache.put(e);
}
/**
 * Returns if the Control can edit its contents.
 */
//===================================================================
public boolean canEdit()
//===================================================================
{
	int flags = getModifiers(true);
	boolean editable = (((flags & (NotEditable|DisplayOnly)) == 0) || ((flags & NotAnEditor) != 0));
	return editable;
}

/*###############################################################
 * 
 * Tags, constraints and sizes.
 * 
 ###############################################################*/
//==================================================================
public int constraints = STRETCH|FILL|CENTER;
public TagList tags = null;
//===================================================================
//===================================================================
public Control setControl(int val) {
	if ((val & (HEXPAND|HCONTRACT|VEXPAND|VCONTRACT)) == 0){
		if ((val & (LEFT|RIGHT)) == (LEFT|RIGHT)) val |= HFILL;
		if ((val & (TOP|BOTTOM)) == (TOP|BOTTOM)) val |= VFILL;
	}
	constraints = (constraints & ~CONTROLMASK)|(val & CONTROLMASK); return this;
}
public Control setCell(int val) {constraints = (constraints & ~CELLMASK)|(val & CELLMASK); return this;}
//===================================================================
public Control setTag(int tag,Object value)
//===================================================================
{
	if (tags == null) tags = new TagList();
	tags.set(tag,value);
	return this;
}
//===================================================================
public Object getTag(int tag,Object defaultValue)
//===================================================================
{
	if (tags == null) return defaultValue;
	return tags.getValue(tag,defaultValue);
}
//===================================================================
public Control clearTag(int tag)
//===================================================================
{
	if (tags != null) tags.clear(tag);
	return this;
}
//===================================================================
public Control setTags(TagList tl)
//===================================================================
{

	if (tags == null) tags = new TagList();
	tags.set(tl);
	return this;
}
//===================================================================
public Control defaultTo(int tag,Object value)
//===================================================================
{
	if (tags == null) tags = new TagList();
	tags.defaultTo(tag,value);
	return this;
}
//===================================================================
public boolean hasTag(int tag)
//===================================================================
{
	if (tags == null) return false;
	return tags.hasTag(tag);

}
/**
 * Set the background image for the Control scaled to fit the Control.
 * @param image the image.
 * @param scaleAlphaBy a value to scale the transparency of the Image by. A value
 * of 1.0 indicates no change to the alpha value. Values less than 1 makes the
 * image more transparent by that value.
 * @return this Control.
 */
public Control setScaledBackgroundImage(ImageData image, double scaleAlphaBy)
{
	return setBackgroundImage(image,BackgroundImage.DISPLAY_TYPE_CACHED_SCALED,scaleAlphaBy);
}
/**
 * Set the background image for the Control in a tiled pattern.
 * @param image the image.
 * @param scaleAlphaBy a value to scale the transparency of the Image by. A value
 * of 1.0 indicates no change to the alpha value. Values less than 1 makes the
 * image more transparent by that value.
 * @return this Control.
 */
public Control setTiledBackgroundImage(ImageData image, double scaleAlphaBy)
{
	return setBackgroundImage(image,BackgroundImage.DISPLAY_TYPE_TILED,scaleAlphaBy);
}
/**
 * Set the background image for the Control.
 * @param image the image.
 * @param backgroundImageType one of the BackgroundImage.DISPLAY_TYPE_XXX values.
 * @param scaleAlphaBy a value to scale the transparency of the Image by. A value
 * of 1.0 indicates no change to the alpha value. Values less than 1 makes the
 * image more transparent by that value.
 * @return this Control.
 */
public Control setBackgroundImage(ImageData image, int backgroundImageType, double scaleAlphaBy)
{
	if (scaleAlphaBy < 1)
		image = new PixelBuffer(image).scaleAlpha(scaleAlphaBy).toPicture();
	setTag(TAG_BACKGROUND_IMAGE,new BackgroundImage(image,backgroundImageType,0));
	return this;
}
//......................................................
//These are values which will be calculated. Do not use
//them to set the preferred or minimum values.
//......................................................
/** Do not set this directly - use setPrefferedSize() instead. */
protected int preferredWidth = 10;
/** Do not set this directly - use setPrefferedSize() instead. */
protected int preferredHeight = 10;
/** Do not set this directly - use setMinimumSize() instead. */
protected int minWidth = 0;
/** Do not set this directly - use setMinimumSize() instead. */
protected int minHeight = 0;
/** Do not set this directly - use setMaximumSize() instead. */
protected int maxWidth = -1;
/** Do not set this directly - use setMaximumSize() instead. */
protected int maxHeight = -1;

//==================================================================
public Control setFixedSize(int width,int height) {return setTag(TAG_FIXEDSIZE,new Dimension(width,height));}
public Control setPreferredSize(int width,int height) {return setTag(TAG_PREFERREDSIZE,new Dimension(width,height));}
public Control setTextSize(int width,int height) {return setTag(TAG_TEXTSIZE,new Dimension(width,height));}
public Control setMinimumSize(int width,int height){return setTag(TAG_MINIMUMSIZE,new Dimension(width,height));}
public Control setMaximumSize(int width,int height){return setTag(TAG_MAXIMUMSIZE,new Dimension(width,height));}
//==================================================================
//-------------------------------------------------------------------
private void doCalculateSizes()
//-------------------------------------------------------------------
{
	modify(CalculatedSizes,0);
	calculateSizes();
	Dimension d = (Dimension)getTag(TAG_TEXTSIZE,null);
	if (d != null) {
		calculateTextSize(d.width,d.height,tempD);
		preferredWidth = java.lang.Math.max(tempD.width,preferredWidth); 
		preferredHeight = java.lang.Math.max(tempD.height,preferredHeight);
	}
}
//-------------------------------------------------------------------
private Dimension checkStandardSizes(Dimension d)
//-------------------------------------------------------------------
{
	int myFlags = modifiers;
	if (((myFlags & CalculatedSizes) == 0) ||((myFlags & AlwaysRecalculateSizes) != 0))
		doCalculateSizes();
//------------------------------------------------------------------
	for (Control c = parent; c != null; c = c.parent)
		myFlags |= c.modifiers;
	if ((myFlags & ShrinkToNothing) != 0) return d.set(0,0);
	Dimension d2 = (Dimension)getTag(TAG_FIXEDSIZE,null);
	if (d2 != null) return d.set(d2.width,d2.height);
	return null;
}
//-------------------------------------------------------------------
private Dimension getASize(Dimension dest,int tag)
//-------------------------------------------------------------------
{
	if (dest == null) dest = new Dimension();
//------------------------------------------------------------------
	if (checkStandardSizes(dest) != null) return dest; 
//------------------------------------------------------------------
	Dimension d = (Dimension)getTag(tag,null);
//------------------------------------------------------------------
	if (tag == TAG_PREFERREDSIZE) dest.set(preferredWidth,preferredHeight);
	else if (tag == TAG_MINIMUMSIZE) dest.set(minWidth,minHeight);
	else if (tag == TAG_MAXIMUMSIZE) dest.set(maxWidth,maxHeight);
	if (d != null) {
		if (d.width >= 0) dest.width = d.width;
		if (d.height >= 0) dest.height = d.height;
		//return dest.set(d.width,d.height);
	}
	return dest;
}
static Dimension tempD = new Dimension();
/**
* This gets all of the control sizes. Currently it returns the
* preferredSize, minimumSize and maximumSize. They fit into the array
* as follows:
* 
* preferredWidth,preferredHeight,minWidth,minHeight,maxWidth,maxHeight
* 
**/
//===================================================================
public int [] getSizes(int [] values)
//===================================================================
{
	int myFlags = getModifiers(false);
	int flags = getModifiers(true);
	if (values == null) values = new int[6];
	if (values.length < 6) values = new int[6];
//..................................................................
	if (((myFlags & CalculatedSizes) == 0) ||((myFlags & AlwaysRecalculateSizes) != 0))
		doCalculateSizes();
//..................................................................
	if ((flags & ShrinkToNothing) != 0){
		values[0] = values[2] = values[4] = 
		values[1] = values[3] = values[5] = 0;
		return values;
	}
//..................................................................
	Dimension d2 = (Dimension)getTag(TAG_FIXEDSIZE,null);
	if (d2 != null) {
		values[0] = values[2] = values[4] = d2.width;
		values[1] = values[3] = values[5] = d2.height;

		return values;
	}
//..................................................................
	Dimension dest = getPreferredSize(tempD);
	values[0] = dest.width; values[1] = dest.height;
	dest = getMinimumSize(dest);
	values[2] = dest.width; values[3] = dest.height;
	dest = getMaximumSize(dest);
	values[4] = dest.width; values[5] = dest.height;
	return values;
}
//==================================================================
public Dimension getPreferredSize(Dimension dest)
//==================================================================
{
	return getASize(dest,TAG_PREFERREDSIZE);
}
//==================================================================
public Dimension getMinimumSize(Dimension dest)
//==================================================================
{
	return getASize(dest,TAG_MINIMUMSIZE);
}
//==================================================================
public Dimension getMaximumSize(Dimension dest)
//==================================================================
{
	return getASize(dest,TAG_MAXIMUMSIZE);
}
/**
* Override this to calculate the preferred, minimum and maximum size of your control. This is only
* called once unless the modifier flag AlwaysRecalculateSizes is true.
* During the calculation you will set the variables preferredWidth, preferredHeight, etc. directly.
**/
//-------------------------------------------------------------------
protected void calculateSizes() 
//-------------------------------------------------------------------
{
	preferredWidth = preferredHeight = 15;
}
//-------------------------------------------------------------------
protected Dimension calculateTextSize(int width,int height,Dimension dest)
//-------------------------------------------------------------------
{
	dest = Dimension.unNull(dest);
	FontMetrics fm = getFontMetrics();
	Metrics.getAverageSize(fm,height < 0 ? 10 : height,width < 0 ? 10 : width,2,2,dest);
	if (width < 0) dest.width = width;
	if (height < 0) dest.height = height;
	return dest;
}
protected final void changeDimension(int newWidth, int newHeight)
{
	this.width = newWidth;
	this.height = newHeight;
	if (width <= 0 || height <= 0) return;
	Object obj = getTag(TAG_BACKGROUND_IMAGE,null);
	if (obj instanceof BackgroundImage){
		BackgroundImage bi = (BackgroundImage)obj;
		if (bi.displayType == bi.DISPLAY_TYPE_CACHED_SCALED){
			bi.scaleImageTo(width,height,getBackground());
		}		
	}
}
/**
* This is called to change the width and height values.
* It is always called when the Control gets resized so you can override
* it to react to a resizing of a Control.
*/
//==================================================================
public void resizeTo(int width,int height)
//==================================================================
{
	changeDimension(width,height);
}
/**
* Do not use this to repaint a control - use repaintNow() instead. This is used
* to cause a control to re-layout its components and re-calculate its preferred size.
**/
//==================================================================
public void redisplay()
//==================================================================
{
	modifyAll(ForceResize,CalculatedSizes,true);
	getPreferredSize(null);
}
/**
* This requests that the size be changed to the following dimensions
* but the control is free to ignore it or to change to a different value.
*/
//==================================================================
public void requestResizeTo(int width,int height)
//==================================================================
{
	int myFlags = getModifiers(false);
	if (width == this.width && height == this.height && ((myFlags & ForceResize) == 0)) return;
	modify(0,ForceResize);
	if (((myFlags & PreferredSizeOnly) != 0) || getTag(TAG_FIXEDSIZE,null) != null) {
		Dimension r = getPreferredSize(getCachedDim());
		resizeTo(r.width,r.height);
		cache(r);
	}else
		resizeTo(width,height);
}
/**
 * Return the on-screen coordinates of this Control in a Rect.
 * @param dest an optional destination Rect. If it is null a new one
 * will be created and returned.
 * @return the on-screen co-ordinates of this Control in a Rect.
 */
public Rect getRect(Rect dest)
//===================================================================
{
	if (dest == null) dest = new Rect();
	return dest.set(x,y,width,height);
}
public Rect getRect()
{
	return getRect(new Rect());
}
/**
 * This calls setRect() with the current rect parameters.
 *
 */
public void resetRect() 
//==================================================================
{
	setRect(x,y,width,height);
}
/**
 * Set the on-screen coordinates of this Control. 
 * @param r the Rect representing the x, y, width and height of the
 * Control.
 */
public void setRect(Rect r) 
{
	setRect(r.x,r.y,r.width,r.height);
}
/**
 * Set the on-screen coordinates of this Control. 
 * @param x the x-coordinate.
 * @param y the y-coordinate.
 * @param width the new width of the Control.
 * @param height the new height of the Control.
 */
public void setRect(int x,int y,int width,int height)
//==================================================================
{
	this.x = x;
	this.y = y;
	requestResizeTo(width,height);
}
/**
 * Return the size of the Control in a Rect rather than a Dimension.
 * @param dest an optional destination Rect. If it is null a new one
 * will be created and returned.
 * @return the size of the Control in a Rect rather than a Dimension.
 */
public Rect getDim(Rect dest)
//===================================================================
{
	if (dest == null) dest = new Rect();
	return dest.set(0,0,width,height);

}
/**
 * Return the size of the Control in a Dimension object.
 * @param dest an optional destination Dimension. If it is null a new one
 * will be created and returned.
 * @return the size of the Control in a Dimension.
 */
public Dimension getSize(Dimension dest)
//===================================================================
{
	if (dest == null) dest = new Dimension();
	return dest.set(width,height);
}
/**
 * Return the location of the Control (x and y coordinates) within
 * its parent in a Point.
 * @param dest an optional destination Point. If it is null a new one
 * will be created and returned.
 * @return the location of the Control in a Point.
 */
public Point getLocation(Point dest) 
{
	if (dest == null) dest = new Point();
	return dest.set(x,y);
}
/**
 * Set the location (x and y coordinates) of this Control within its
 * parent.
 * @param x the new x-coordinate.
 * @param y the new y-coordinate.
 */
public void setLocation(int x,int y)
{
	this.x = x; 
	this.y = y;
}

/*###############################################################
 * 
 * These methods deal with the Control heirarchy.
 * 
 ###############################################################*/
//private static final Iterator noChildren = new ObjectIterator(null);
/**
* This returns an Iterator that goes forwards through the components which are physically
* added to this Control - i.e. may be displayed within the Control.
* <p>Note that a sub-control may not be a "child" of the control -
* i.e. it may not have been added into the Control heirarchy.
* @deprecated - use getChildren(ControlIterator,boolean) instead.
**/
public final ControlIterator getChildren() {return getChildren(null,false);}
/**
* This returns an Iterator that goes backwards through the components which are physically
* added to this Control - i.e. may be displayed within the Control.
* <p>Note that a sub-control may not be a "child" of the control -
* i.e. it may not have been added into the Control heirarchy.
**/
public final ControlIterator getChildrenBackwards() {return getChildren(null,true);}
/**
 * Get an Iterator to go through the list of all the direct children of this
 * Control re-using a ControlIterator if provided. 
 * @param dest an optional destination ControlIterator.
 * @param backwards true to iterate through the children in the reverse order
 * in which they were added. 
 * @return a ControlIterator to go through the list of all the direct children of this Control.
 */
public ControlIterator getChildren(ControlIterator dest, boolean backwards)
{
	return ControlIterator.unNull(dest).setForNone();
}
public final ControlIterator getCachedChildren(boolean backwards)
{
	return getChildren((ControlIterator)Cache.get(ControlIterator.class),backwards); 
}
public final ControlIterator getCachedSubControls()
{
	return getSubControls((ControlIterator)Cache.get(ControlIterator.class)); 
}
//===================================================================
/**
* Get all the children for this Control and their children.
**/
//===================================================================
public Iterator getAllDescendants(boolean backwards)
//===================================================================
{
	Stack s = new Stack();
	Vector v = new Vector();
	Iterator it = getCachedChildren(backwards);
	while(true){
		if (it.hasNext()){
			Control c = (Control)it.next();
			v.add(c);
			s.push(it);
			it = c.getCachedChildren(backwards);
		}else {
			Cache.put(it);
			if (s.size() == 0) break;
			else it = (Iterator)s.pop();
		}
	}
	return v.iterator();
}
/**
* Get all the sub-controls for this Control and their sub-controls.
**/
//===================================================================
public Iterator getAllSubControls()
//===================================================================
{
	Stack s = new Stack();
	Vector v = new Vector();
	Iterator it = getCachedSubControls();
	while(true){
		if (it.hasNext()){
			Control c = (Control)it.next();
			v.add(c);
			s.push(it);
			it = c.getCachedSubControls();
		}else {
			Cache.put(it);
			if (s.size() == 0) break;
			else it = (Iterator)s.pop();
		}
	}
	return v.iterator();
}
/**
* This returns an Iterator that goes forwards through the components
* which are considered a sub-control of this Control, even though they
* may not be a "child" of the control -
* i.e. it may not have been added into the Control heirarchy.
* @deprecated - use getChildren(ControlIterator,boolean) instead.
**/
//===================================================================
public final ControlIterator getSubControls() {return getChildren();}
//===================================================================
/**
* This returns an Iterator that goes forwards through the components
* which are considered a sub-control of this Control, even though they
* may not be a "child" of the control -
* i.e. it may not have been added into the Control heirarchy.
**/
//===================================================================
public ControlIterator getSubControls(ControlIterator dest) 
//===================================================================
{
	return getChildren(dest,false);
}
/**
* This is used to "make" the control before being displayed. This is usually only
* overridden by containers. The default version of this does nothing.
* @param reMake if this is true then you should do a full re-make.
*/
//===================================================================
public void make(boolean reMake)
//===================================================================
{
}
/** Returns the control's parent container. */
public Container getParent()
{
	return parent;
}

/** Returns the next child in the parent's list of controls. */
public Control getNext()
{
	return next;
}

/**
 * Returns true if the given x and y coordinate in the parent's
 * coordinate system is contained within this control.
 */
public boolean contains(int x, int y)
{
	int rx = this.x;
	int ry = this.y;
	if (x < rx || x >= rx + this.width || y < ry || y > ry + this.height)
		return false;
	return true;
}
/**
 * See if this Control contains the specified Control.
 * @param who the possible child or anscestor control.
 * @return true if this Control contains the specified Control.
 */
//==================================================================
public boolean contains(Control who)
//==================================================================
{
	if (who == null) return false;
	for (Container c = who.parent; c != null; c = c.parent)
		if (c == this) return true;
	return false;
}

//===================================================================
public final boolean isChildOf(Container who)
//===================================================================
{
	if (who == null) return false;
	for (Control c = this; c != null; c = c.parent)
		if (c == who) return true;
	return false;
}

/*###############################################################
 * 
 * Navigating frames and windows.
 * 
 ###############################################################*/

/**
* This determines if the Control is being displayed in a modal frame.
**/
//===================================================================
public final boolean isModal()
//===================================================================
{
	for (Control c = this;c != null;c = c.parent)
		if (c instanceof Frame) 
			if (((Frame)c).isModal) return true;
	return false;
}
/**
 * A convenience method for setting up a Control for later hide() and unhide()
 * calls. 
 * @param hiddenAtFirst true to have the Control added to its parent in an
 * initially hidden state, false for it to be added normally.
 * @param parentForHiding the parent Container that will be used for hide()/unhide()
 * calls. If this is null it will default to the Frame of the Control when
 * hide()/unhide() is called.
 */
public void setHiddenParent(boolean hiddenAtFirst, Container parentForHiding)
{
	if (parentForHiding == null) clearTag(TAG_HIDE_PARENT);
	else setTag(TAG_HIDE_PARENT,parentForHiding);
	if (hiddenAtFirst) constraints |= INITIALLY_HIDDEN;
	else constraints &= ~INITIALLY_HIDDEN;
}
/**
 * Get the parent Container used to hide()/unhide() the Control. This
 * is the Container object tagged as TAG_HIDE_PARENT. If this is null
 * or invalid then getFrame() is used to determine the parent. 
 * @return the parent Container used to hide()/unhide() the Control.
 */
public Container getHideParent()
{
	Object got = getTag(TAG_HIDE_PARENT,null);
	if (!(got instanceof Container)) return getFrame();
	return (Container)got;
}
/**
 * Temporarily hide the Control. The parent Control for hiding/unhiding
 * is the value returned by getHideParent().
 * @param updateParent if this is true then the hide parent will have
 * relayout() called on it and will update itself onscreen.
 */
public void hide(boolean updateParent)
{
	Container c = getHideParent();
	if (c == null) return;
	c.hide(this,updateParent);
}
/**
 * Reveal the Control after being hidden by hide(). The parent Control for hiding/unhiding
 * is the value returned by getHideParent().
 * @param updateParent if this is true then the hide parent will have
 * relayout() called on it and will update itself onscreen.
 */
public void unhide(boolean updateParent)
{
	Container c = getHideParent();
	if (c == null) return;
	c.unhide(this,updateParent);
}
public void setHidden(boolean isHidden,boolean updateParent)
{
	Container c = getHideParent();
	if (c == null) return;
	c.setHidden(this,isHidden,updateParent);
}
//==================================================================
public Frame getFrame()
//==================================================================
{
	//if (myFrame != null) return myFrame;
	for (Control c = this;c != null;c = c.parent)
		if (c instanceof Frame && c != this) return (Frame)c;
	return null;
}
//===================================================================
public Window getWindow()
//===================================================================
{
	for (Control c = this; c != null; c = c.parent)
		if (c instanceof Window) return (Window)c;
		
	return null;
}
//==================================================================
public Container getFrameOrContainer()
//==================================================================
{
	for (Control c = this;c != null;c = c.parent)
		if (c instanceof Frame) return (Frame)c;
		else if (c.parent == null) 
			if (c instanceof Container) return (Container)c;
			else return null;
	return null;
}
/**
 * Returns true if this Control is in the topmost Frame.
 */
//==================================================================
public boolean amOnTopFrame()
//==================================================================
{
	if (getWindow() == null) return false;
	Frame tf = Gui.topFrame(this);
	if (tf == null) return true;
	for (Frame f = getFrame(); f != null; f = f.getFrame())
		if (f == tf) return true;
	return false;
}
public Point getPosInFrame()
{
	return getPosInFrame(null);
}
//==================================================================
public Point getPosInFrame(Point dest)
//==================================================================
{
	Point p = Point.unNull(dest);
	dest.x = dest.y = 0;
	for (Control c = this;c != null;c = c.parent){
		if (c instanceof Frame) return p;
		Rect r = c.getRect();
		p.translate(r.x,r.y);
	}
	return p;
}


/*###############################################################
 * 
 * Painting
 * 
 ###############################################################*/
/**
 * Avoid using this method - use getGraphics() instead to get a BufferedGraphics
 * for painting to the Control.
 * Return a Graphics context for this control for drawing directly onto the
	window surface if possible. The Graphics is translated so that (0,0) will represent
	the top left corner of the Control and is cliped so areas outside of the
	Control are not drawn on. 
 * @return a Graphics context for this control or null if not supported
 */
//==================================================================
public final Graphics tryGetGraphics() //throws UnsupportedOperationException
//==================================================================
{
	if (canMakeGraphics == 0) return null;
	Point p = Point.getCached(0,0);
	Rect cr = Rect.getCached();
	try{
		Rect r = Gui.getRectInWindow(this,cr,true,p);
		if (r == null) {
			//Vm.debug("r == null");
			return null;
		}
		Window w = getWindow();
		if (w == null) return null;
		WindowSurface ws = w.getWindowSurface();
		if (ws == null) return null;
		Graphics g = ws.getGraphics();
		if (g == null) {
			//Vm.debug("Gui.getGraphics() == null");
			return null;
		}
		if (!g.isValid()) {
			//Vm.debug("!g.isValid()");
			return null;
		}
		g.setClip(r.x,r.y,r.width,r.height);
		g.translate(p.x,p.y);
		g.setDrawOp(g.DRAW_OVER);
		g.setFont(getFont());
		return g;
	}catch(UnsupportedOperationException e){
		canMakeGraphics = 0;
		return null;
	}finally{
		p.cache();
		cr.cache();
	}
}

/**
 * Get a BufferedGraphics for drawing to this Control given an original Graphics that
 * may be unbuffered (ie drawing directly to a Window). If the original is null it 
 * will automatically call getGraphics() to get a new Graphics context setup for this
 * Control.
 * @param original the original Graphics that may be unbuffered or may be null.
 * @param x the x location within the Graphics to draw to. 
 * This is usually the x location
 * within the control you wish to draw to as the original Graphics will usually be
 * translated so that (0,0) in the Graphics is set to the top-lef corner of the Control. 
 * @param y the y location within the Graphics to draw to. 
 * This is usually the y location
 * within the control you wish to draw to as the original Graphics will usually be
 * translated so that (0,0) in the Graphics is set to the top-lef corner of the Control. 
 * @param width the width within the Control you wish to draw to.
 * @param height the height within the Control you wish to draw to.
 * @param options one of the BufferedGraphics OPTION_XXX values.
 * @return a BufferedGraphics for use with drawing or null if original was null and getGraphics()
 * also returned null. Or if buffering was not possible (due to memory constraints or
 * system resource problems) and the BufferedGraphics.OPTION_DONT_ACCEPT_UNBUFFERED
 * was specified. You must call release() or cancel() on the BufferedGraphics to release it
 * and restore the original or free() the original if it was created by getGraphics(). 
 */
public BufferedGraphics bufferGraphics(Graphics original, int x, int y, int width, int height, int options)
{
	return bufferGraphics(original,x,y,width,height,options,null);
}
public BufferedGraphics getGraphics(int x, int y, int width, int height)
{
	return bufferGraphics(null,x,y,width,height,0);
}
public BufferedGraphics getGraphics()
{
	return getGraphics(0,0,width,height);
}
public BufferedGraphics bufferGraphics(Graphics gr)
{
	return bufferGraphics(gr,0,0,width,height,0);
}

private static int canMakeGraphics = -1;

public BufferedGraphics bufferGraphics(Graphics original, int x, int y, int width, int height, int options,BufferedGraphics reUse)
{
	//options |= BufferedGraphics.OPTION_EXACT_SIZE;
	if (width <= 0 || height <= 0) {
		return null;
	}
	ISurface surface = null;
	int cx = 0, cy = 0, cw = 0, ch = 0;
	int px = 0, py = 0;
	Graphics useGraphics = original;
	if (useGraphics == null && canMakeGraphics != 0)
		useGraphics = tryGetGraphics();
	BufferedGraphics bg = reUse;
	if (bg == null) bg = BufferedGraphics.getCached();
	boolean amReturning = false;
	try{
		if (useGraphics == null) {
			//
			// If I'm here then no original Graphics was provided
			// AND the Window did not give me a Graphics.
			//
			Window w = getWindow();
			surface = w == null ? null : w.getSurface();
			if (surface == null) return null;
			Rect r = Rect.getCached();
			Point p = Point.getCached(0,0);
			try{
				if (Gui.getRectInWindow(this,r,true,p) == null)
					return null;
				cx = r.x; cy = r.y;
				cw = r.width; ch = r.height;
				px = p.x; py = p.y;
			}finally{
				r.cache();
				p.cache();
			}
			bg.setFor(surface,px,py,x,y,width,height,options);
			bg.clipSurface(cx,cy,cw,ch);
		}else{
			bg.setFor(useGraphics,x,y,width,height,options);
			bg.freeOriginal = (useGraphics != original);
		}
		bg.cacheWhenDone = bg != reUse; 
		bg.dontFree = true;
		
		if ((modifiers & Transparent) == 0)
			options |= BufferedGraphics.OPTION_DONT_COPY_FROM_ORIGINAL;
		if ((options & BufferedGraphics.OPTION_DONT_ACCEPT_UNBUFFERED) != 0)
			if (!bg.drawingToImage()){
				//System.out.println("Cancelling!");
				bg.cancel(); //Will be cached.
				bg = null;
			}
		if (bg != null){
			//
			// See if I am inside a Frame that has an alpha value < 1.0.
			//
			Frame mine = null;
			if (original == null)
				for (Container c = parent; c != null && mine == null; c = c.parent){
					if (c instanceof Frame && ((Frame)c).alpha < 1.0f)
						mine = (Frame)c;
				}
			if (mine != null){
				//
				// My frame is semi transparent, so will have to do things
				// differently.
				//
				fillBackgroundNotInFrame(bg.getGraphics(),mine);
				//
				//System.out.println("Layering: "+getClass());
				bg.createAlphaLayer(mine.alpha);
				//
			}else if ((options & BufferedGraphics.OPTION_DONT_COPY_FROM_ORIGINAL) == 0)
				if (original == null || !bg.canCopyFromOriginal()){
					fillBackground(bg.getGraphics());
				}
		}
		if (bg == null){
			//new Exception("I Could not get buffered Graphics: "+getClass().getName()+", "+new Rect(x,y,width,height)).printStackTrace();
		}
		amReturning = true;
		return bg;
	}finally{
		if (!amReturning && bg != null && bg != reUse){
			Cache.put(bg);
		}
	}
}
/**
 * This is only used in special circumstances such as repaintNow()
 * when you are or may be buffering directly to the Window surface. It
 * will cancel the buffer if the Control is no longer a child of the
 * window.  
 * @param g the BufferedGraphics.
 */
public void releaseBufferedGraphics(BufferedGraphics g)
{
	synchronized(Window.windowReleaseLock){
		if (g.bufferedToSurface()){
			Window w = getWindow();
			if (w == null){
				//System.out.println("Not a child anymore: "+getClass());
				g.cancel();
				return;
			}
			g.release();
			return;
		}
	}
	g.release();
}
/**
 * Get a BufferedGraphics for drawing to this Control given an original Graphics that
 * may be unbuffered (ie drawing directly to a Window). If the original is null it 
 * will automatically call getGraphics() to get a new Graphics context setup for this
 * Control.
 * @param original the original Graphics that may be unbuffered or may be null.
 * @param area the area within the Graphics to draw to. 
 * This is usually the area
 * within the control you wish to draw to as the original Graphics will usually be
 * translated so that (0,0) in the Graphics is set to the top-lef corner of the Control. 
 * @param options one of the BufferedGraphics OPTION_XXX values.
 * @return a BufferedGraphics for use with drawing or null if original was null and getGraphics()
 * also returned null. Or if buffering was not possible (due to memory constraints or
 * system resource problems) and the BufferedGraphics.OPTION_DONT_ACCEPT_UNBUFFERED
 * was specified. You must call release() or cancel() on the BufferedGraphics to release it
 * and restore the original or free() the original if it was created by getGraphics(). 
 */
public BufferedGraphics bufferGraphics(Graphics original, Rect area, int options)
{
	if (area == null) return bufferGraphics(original,0,0,width,height,options,null);
	else return bufferGraphics(original,area.x,area.y,area.width,area.height,options,null);
}
public BufferedGraphics bufferGraphics(Graphics original, Rect area, int options,BufferedGraphics reUse)
{
	if (area == null) return bufferGraphics(original,0,0,width,height,options,reUse);
	else return bufferGraphics(original,area.x,area.y,area.width,area.height,options,reUse);
}
/**
 * This is used for paint buffering during a repaintNow() operation. If this
 * returns null then double buffering will not be used on the Control.
 * @param r the area to be buffered. If this is null you should return an Image
 * for the whole area.
 * @return an Image for paint buffering.
 * @throws ewe.sys.SystemResourceException if an Buffer could not be created.
 */
//==================================================================
/*
public Image getControlBuffer(Rect r) throws SystemResourceException
{	
	if (r == null){
		if (width <= 0 || height <= 0) return null;
		return new Image(width,height);
	}
	if (r.width <= 0 || r.height <= 0) return new Image(1,1); 
	else return new Image(r.width,r.height);
}
public ImageBuffer getControlBuffer(int width, int height) throws SystemResourceException
{
	ImageBuffer ib = (ImageBuffer)getCached(ImageBuffer.class);
	ib.get(width,height,false);
	return ib;
}
public ImageBuffer getControlBuffer(Rect r) throws SystemResourceException
{
	int w = r == null ? width : r.width;
	int h = r == null ? height : r.height;
	return getControlBuffer(w,h);
}
public void freeControlBuffer(ImageBuffer buffer)
{
	cache(buffer);
}
*/
/**
 * This is used for paint buffering during a repaintNow() operation. If this
 * returns null then double buffering will not be used on the Control.
 * @return an Image for paint buffering.
 * @throws ewe.sys.SystemResourceException if an Buffer could not be created.
 */
/*
public ImageBuffer getControlBuffer() throws SystemResourceException 
{
	return getControlBuffer(null);
}
*/
	/**
	Repaint the Control now. 
	 */
//	===================================================================
	public void repaintNow() {
//	===================================================================
		repaintNow(null,null);
	}
	/**
	Repaint the Control now if the "yes" parameter is true.
	 */
//	===================================================================
	public void repaintNow(boolean yes) {if (yes) repaintNow(null,null);}
//	==================================================================

	/**
	 * This method calls pauseEventThread() on this Control and then
	 * waits until the provided Handle (which must not be null) to
	 * complete (have the Stopped bit set). It then resumes the
	 * Thread and returns the "returnValue" field of the Handle. 
	 * If the Handle failed and the error field was set in the Handle
	 * that error is thrown as a RuntimeException.
	 * @param h the Handle to wait for.
	 * @return the returnValue field of the Handle.
	 */
	public Object waitEventThread(Handle h)
	{
		if (h == null) throw new NullPointerException("No handle!");
		Window w = getWindow();
		//System.out.println("Wait: "+(w == null ? false : w.inCurrentEventThread()));
		Object exec = WindowEventThread.execThread(w,this);
		try{
			return h.waitForReturnValue();
		}finally{
			WindowEventThread.endExecThread(exec);
			//System.out.println("Done: "+(w == null ? false : w.inCurrentEventThread()));
		}
	}
	/**
	 * Call this within a Thread handling an Event for the Control if
	 * the Event handling may block. This method will spawn another Thread
	 * to handle Paint/Resize events while this thread is blocked. No user
	 * events are allowed while this thread is blocked.<p>
	 * Use resumeEventThread() when the operation is complete to resume this thread.  
	 * @return an Object to use with resumeEventThread().
	 */
	public Object pauseEventThread()
	{
		return WindowEventThread.execThread(getWindow(),this);
	}
	/**
	 * Call this within a Thread handling an Event after a call to 
	 * pauseEventThread(). It allows normal event handling to continue
	 * for the Control.
	 */
	public void resumeEventThread(Object fromPause)
	{
		WindowEventThread.endExecThread(fromPause);
	}
	/**
	 * Call this within a Thread handling an Event for the Control if
	 * the Event handling may block. This method will spawn another Thread
	 * to handle Paint/Resize events while this thread is blocked. User
	 * Events for the toDisable Control are blocked while the new Event
	 * Thread is running.<p>
	 * Use resumeEventThread() when the operation is complete to resume this thread.
	 * @param toDisable an optional Control to disable while the new
	 * Event thread is running.  
	 * @return an Object to use with resumeEventThread().
	 */
	public Object newEventThread(Control toDisable)
	{
		return WindowEventThread.execThread(getWindow(),toDisable);
	}
	
	/**
	 * See if a paint is allowed on the Control at this time.
	 */
	public boolean requestPaint()
	{
		return Gui.requestPaint(this);
	}
	/**
	 * Request a repaint of the control in the Window thread.
	 * @param alwaysInSeparateCall if this is true the repaint is always done in
	 * a separate call of the Window Event Thread even if the current thread is
	 * the active Window Event Thread.
	 * @param area the area to repaint, or null to repaint the entire area.
	 */
	public void repaintInWindowThread(boolean alwaysInSeparateCall, final Rect area)
	{
		Window window = getWindow();
		if (window == null) return;
		if (!alwaysInSeparateCall && window.inCurrentEventThread()) repaintNow(null,area);
		else{
			WindowSurface ws = window.getWindowSurface();
			if (ws != null) {
				Point p = Point.getCached(0,0); 
				Gui.getPosInParent(this,window,p);
				if (area == null) ws.requestRepaint(p.x,p.y,width,height,false);
				else ws.requestRepaint(p.x+area.x,p.y+area.y,area.width,area.height,false);
				p.cache();
			}
			/*
			new WindowThreadTask(){
				protected void doRun(){
					repaintNow(null,area);
				}
			}.run(this,alwaysInSeparateCall);
			*/
		}
	}
	/**
	 * Request a repaint of the entire control in the Window thread.
	 * If the current thread is the active Window Event Thread then the
	 * repaint is done immediately. Otherwise it will be done in a Window Event
	 * Thread call.
	 */
	public void repaintInWindowThread()
	{
		repaintInWindowThread(false,null);
	}
	/**
	 * This is the same as repaintInWindowThread() and requests a
	 * paint of the Control synchronized to the Window Event Thread.
	 */
	public void repaint()
	{
		repaintInWindowThread();
	}
	public void repaint(int x, int y, int width, int height)
	{
		Rect r = Rect.getCached(x,y,width,height);
		try{
			repaintInWindowThread(false,r);
		}finally{
			r.cache();
		}
	}
	static Object guiLock = new Object();
	
	//public float alpha = 1.0f;
	
	private static PixelBuffer doPaint(Control c, Graphics g, Rect where)
	{
		if (where.width <= 0 || where.height <= 0 || !(c instanceof Frame) || ((Frame)c).alpha >= 1.0f){
			c.doPaint(g,where);
			return null;
		}
		PixelBuffer ib = new PixelBuffer(where.width,where.height);
		Graphics g2 = ib.getDrawingBuffer(null,null,((Frame)c).alpha);
		g2.translate(-where.x,-where.y);
		c.doPaint(g2,where);
		return ib;
	}
	/**
	Repaint a section of the Control now. 
	 * @param gr The Graphics to paint to. If this is null a new Graphics will be created.
	 * @param where The rectangle within the Control to paint. If this is null the entire Control will be painted.
	 */
//	===================================================================
	public void repaintNow(Graphics gr,Rect where)
//	===================================================================
	{
		
		boolean isWin = this instanceof Window;
			if (!isWin){
				Window w = getWindow();
				if (w == null) return;
				if (!w.hasBeenPainted()) {
					return;
				}
			}
			Rect old = Rect.getCached();
			Rect dim = Rect.getCached();
			try{
				if (gr == null)
					if (!Gui.requestPaint(this)) {
						//System.out.println("Can't paint: "+this);
						//if (isWin||true) System.out.println("Window can't paint!");
						return;
					}
				int flags = getModifiers(true);
				if ((flags & Invisible) != 0) {
					if (isWin||true) {
						//System.out.println("Window is invisible!");
					}
					return;
				}
				getDim(dim);
				if (where == null) where = dim; 
				else if (where.x < 0 || where.x+where.width > width || where.y < 0 || where.y+where.height > height){
					getDim(dim);
					dim.getIntersection(where,dim);
					where = dim;
				}
				boolean doAll = (Window.bufferFullWindow) && gr == null && doubleBuffer;
				if (doAll){
					Window w = getWindow();
					gr = w.getWindowGraphics();
					doAll = gr != null;
					if (doAll){
						//Rect r = Gui.getAppRect(this,Rect.getCached());
						Point r = Point.getCached(0, 0);
						Rect rr = Rect.getCached();
						Rect wr = Gui.getRectInWindow(this, rr, true, r);
						try{
							if (wr != null){
								Image im = new Image(width,height);
								Graphics g = new Graphics(im);
								if ((modifiers & Transparent) != 0)
									fillBackground(g);
								PixelBuffer pb = doPaint(this,g,where);
								doPaintChildren(pb == null ? g : pb.getCurrentDrawingBuffer(),where.x,where.y,where.width,where.height);
								if (pb != null){
									pb.putDrawingBuffer(pb.PUT_SET);
									pb.draw(g,where.x,where.y,0);
									pb.free();
								}
								g.free();
								gr.setClip(r.x+where.x,r.y+where.y,where.width,where.height);
								//System.out.println(getClass()+": "+r+", "+where+", "+wr);
								gr.reduceClip(wr.x,wr.y,wr.width,wr.height,rr);
								gr.drawImage(im,r.x,r.y);
								im.free();
								gr.free();
								//System.out.println("Did all: "+where);
							}
							return;
						}finally{
							r.cache();
							rr.cache();
						}
					}
				}
				int opts = 0;
				if ((modifiers & Transparent) == 0)
					opts |= BufferedGraphics.OPTION_DONT_COPY_FROM_ORIGINAL;
				if (!doubleBuffer || ((modifiers & DontBuffer) != 0)) {
					opts |= BufferedGraphics.OPTION_DONT_BUFFER;
				}
				BufferedGraphics bg =
					bufferGraphics(gr,where.x,where.y,where.width,where.height,opts);
				if (bg == null) return;
				try{
					Graphics use = bg.getGraphics();
					boolean didLock = lockPaint;
					if (didLock) readLock();
					PixelBuffer al = null;
					try{
						al = doPaint(this,use,where);
					}finally{
						if (didLock) unlock();
					}
					doPaintChildren(al == null ? use : al.getCurrentDrawingBuffer(),where.x,where.y,where.width,where.height);
					if (al != null){
						al.putDrawingBuffer(al.PUT_SET);
						al.draw(use,where.x,where.y,0);
						al.free();
					}
				}catch(Throwable t){
					//t.printStackTrace();
				}finally{
					releaseBufferedGraphics(bg);
				}
			}finally{
				old.cache();
				dim.cache();
			}
	}
	/**
	 * By default this calls paintBackground().
	 * @param g the non-null Graphics to paint to.
	 */
	public void doBackground(Graphics g) 
//	==================================================================
	{
		paintBackground(g);
	}
//	===================================================================
	public void fillBackground(Graphics g)
//	===================================================================
	{
		if (g == null) return;
		if (!g.isValid()) return;
		Point p = new Point();
		Rect r = Gui.getRectInWindow(this,null,true,p);
		if (r == null) return;
		Window w = getWindow();
		g.translate(-p.x,-p.y);
		int m = modify(Invisible,0);
		w.repaintNow(g,r);
		restore(m,Invisible);
		g.translate(p.x,p.y);
	}
//	===================================================================
	public void fillBackgroundNotInFrame(Graphics g,Frame f)
//	===================================================================
	{
		if (g == null) return;
		if (!g.isValid()) return;
		Point p = new Point();
		Rect r = Gui.getRectInWindow(this,null,true,p);
		if (r == null) return;
		Window w = getWindow();
		g.translate(-p.x,-p.y);
		int m = f.modify(Invisible,0);
		w.repaintNow(g,r);
		f.restore(m,Invisible);
		g.translate(p.x,p.y);
	}
	/**
	 * This is the main method to override when doing custom Control painting.
	 * It is called when the Control needs refreshing or when repaintNow() is called.
	 * @param g the graphics to paint to.
	 * @param r the area within the Control to paint. This can be ignored and the entire
	 * area can be painted if so desired. 
	 */
//	==================================================================
	public void doPaint(Graphics g,Rect r)
//	==================================================================
	{

		int flags = getModifiers(true);
		if ((flags & Invisible) != 0) return;
		doBackground(g);
		g.setColor(Color.Black);
		doBorder(g);
	}
	/**
	 * Return a Graphics context for this control for drawing directly onto the
	 * window surface IF supported by the Window.
	 * @return a Graphics context for this control or null if the Control is not
	 * visible or if the Window does not allow direct drawing via a Graphics Object. 
	 */
	public Graphics getWindowGraphics()
//	==================================================================
	{
		Point p = new Point();
		Rect r = Gui.getRectInWindow(this,null,true,p);
		if (r == null) return null;
		Window w = getWindow();
		try{
			Graphics g = w.getWindowSurface().getGraphics(); 
			if (!g.isValid())return null;
			g.setClip(r.x,r.y,r.width,r.height);
			g.translate(p.x,p.y);
			g.setDrawOp(g.DRAW_OVER);
			g.setFont(getFont());
			return g;
		}catch(Exception e){
			return null;
		}
	}
	
	/**
	 * This can be used when painting a Control. If you are using an ImageBuffer
	 * @param im
	 * @param whereOnMe
	 * @param exactSize
	 * @param alreadyDrawn
	 * @return
	 */
	/*
	public Graphics prepareImageBuffer(ImageBuffer im, Rect whereOnMe, boolean exactSize,Graphics alreadyDrawn)
	{
		Graphics g = im.get(whereOnMe.width,whereOnMe.height,exactSize);
		if (prepareImageBuffer(g,whereOnMe,alreadyDrawn)) return g;
		return null;
	}
	public boolean prepareImageBuffer(Graphics g, Rect whereOnMe, Graphics alreadyDrawn)
	{
		if ((modifiers & Transparent) == 0){
			g.translate(-whereOnMe.x,-whereOnMe.y);
			doBackground(g);
			g.translate(whereOnMe.x,whereOnMe.y);
			return true;
		}
		if (alreadyDrawn != null && alreadyDrawn.canCopyFrom()){
			System.out.println("Copying!");
			g.copyRect(alreadyDrawn,whereOnMe.x,whereOnMe.y,whereOnMe.width,whereOnMe.height,0,0);
			return true;
		}
		int xx = whereOnMe.x, yy = whereOnMe.y;
		Window w = null;
		for(Control c = this; c != null; c = c.getParent()){
			if (c instanceof Window) {
				w = (Window)c;
				break;
			}
			whereOnMe.x += c.x;
			whereOnMe.y += c.y;
		}
		try{
			if (w == null) return false;
			int was = modify(Invisible,0);
			g.translate(-whereOnMe.x,-whereOnMe.y);
			w.repaintNow(g,whereOnMe);
			g.translate(whereOnMe.x,whereOnMe.y);
			restore(was,Invisible);
			return true;
		}finally{
			whereOnMe.x = xx;
			whereOnMe.y = yy;
		}
	}
	*/
	/**
	 * Do a standard painting of the Background.
	 * @param g the non-null Graphics to paint to.
	 */
	public void paintBackground(Graphics g)
//	==================================================================
	{
		if (hasModifier(Transparent,false)) {
			return;
		}
		g.setColor(getBackground());
		g.setDrawOp(g.DRAW_OVER);
		g.fillRect(0,0,width,height);
		Object obj = getTag(TAG_BACKGROUND_IMAGE,null);
		if (obj instanceof BackgroundImage){
			((BackgroundImage)obj).display(g,x,y,width,height);
		}
	}
//	==================================================================
	public void doBorder(Graphics g)
//	==================================================================
	{
		if (borderWidth == 0 && borderStyle == 0) return;
		synchronized(Rect.gBuffer1){
			g.draw3DRect(getDim(Rect.gBuffer1),borderStyle == 0 ? BDR_OUTLINE|BF_RECT : borderStyle, hasModifier(DrawFlat,true), (borderStyle & BF_MIDDLE) == 0 ? null : getBackground(), borderColor);
		}
	}
	/*
//	==================================================================
	public void oldButWorksDoPaintChildren(Graphics g, int x, int y, int w, int h)
//	==================================================================
	{
		int flags = getModifiers(true);
		if ((flags & Invisible) != 0) return;
		Iterator kids = getChildren();
		if (!kids.hasNext()) return;
		Rect area = new Rect(x,y,w,h);
		Rect inter = new Rect(0,0,0,0);
		Rect oldClip = g.getClip(new Rect());
		Rect curClip = area.getIntersection(oldClip == null ? area:oldClip,null);
		while (kids.hasNext()){//kids.size()-1; i>=0; i--){
			Control mc = (Control)kids.next();
			Rect r = mc.getRect();
			curClip.getIntersection(r,inter);
			if (inter.width <= 0 || inter.height <= 0) continue;
			inter.x -= r.x; inter.y -= r.y;
			g.translate(r.x,r.y);
			//Rect prev = g.getClip(new Rect());

			g.setClip(inter.x,inter.y,inter.width,inter.height);
			//mc.paintBackground(g);
			mc.doPaint(g,inter);
			mc.doPaintChildren(g,inter.x,inter.y,inter.width,inter.height);
			if (oldClip != null) g.setClip(oldClip.x,oldClip.y,oldClip.width,oldClip.height);
			g.translate(-r.x,-r.y);
	  }
	}
	*/
	/*
	protected static boolean haveNativePaint = true;
//	==================================================================
	public void doPaintChildren(Graphics g, int x, int y, int w, int h)
//	==================================================================
	{
		if (haveNativePaint)
			try{
				doPaintChildren(this,getModifiers(true),g,new Rect(x,y,w,h));
				return;
			}catch(Throwable e){
				haveNativePaint = false;
			}
		//notNative_doPaintChildren(this,getModifiers(true),g,new Rect(x,y,w,h));
		doPaintChildren(this,getModifiers(true),g,new Rect(x,y,w,h),g.getClip(new Rect()));
	}
	private static native void doPaintChildren(Control who,int flags,Graphics g, Rect area);
	*/
//	==================================================================
	public void doPaintChildren(Graphics g, int x, int y, int w, int h)
//	==================================================================
	{
		Rect r = Rect.getCached(x,y,w,h);
		Rect clip = Rect.getCached();
		Rect cc = Rect.getCached();
		Rect rr = Rect.getCached();
		try{
			doPaintChildren(this,getModifiers(true),g,r,g.getClip(clip),cc,rr);
		}finally{
			clip.cache();
			r.cache();
			rr.cache();
			cc.cache();
		}
	}
/*
//	==================================================================
	public static void notNative_doPaintChildren(Control who,int flags,Graphics g, Rect area)
//	==================================================================
	{
		doPaintChildren(who,flags,g,area,g.getClip(new Rect()));
	}
*/
	public static Object getCached(Class type)
	{
		return Cache.get(type);
	}
	public static Rect getCachedRect(int x, int y, int width, int height)
	{
		return Rect.getCached(x,y,width,height);
	}
	public static Rect getCachedRect()
	{
		return Rect.getCached();
	}
	public static Dimension getCachedDim(int width, int height)
	{
		return ((Dimension)Cache.get(Dimension.class)).set(width,height);
	}
	public static Dimension getCachedDim()
	{
		return ((Dimension)Cache.get(Dimension.class));
	}
	public static void cache(Object obj)
	{
		Cache.put(obj);
	}
//	-------------------------------------------------------------------
	private static void doPaintChildren(Control who,int flags,Graphics g,Rect area,Rect clip,Rect curClip, Rect r)
//	-------------------------------------------------------------------
	{
		Control child;
		if ((flags & Invisible) != 0) return;
		if (who.children == null) return;
		int ax = area.x, ay = area.y, aw = area.width, ah = area.height;
		//Rect curClip, cc = Rect.getCached();
		//Rect r, rr = Rect.getCached();
		//curClip = cc;
		//r = rr;
		try{
			if (clip == null) curClip.set(area);
			else area.getIntersection(clip,curClip);
			
			for (child = who.children; child != null; child = child.next){
				if ((child.modifiers & Invisible) != 0) {
					continue;
				}
				r.x = child.x; r.y = child.y; r.width = child.width; r.height = child.height;
				curClip.getIntersection(r,area);
				if (area.width <= 0 || area.height <= 0) continue;
				area.x -= r.x; area.y -= r.y;
				g.translate(r.x,r.y);
				g.setClip(area.x,area.y,area.width,area.height);
				//child.paintBackground(g);
				boolean didLock = child.lockPaint;
				BufferedGraphics bg = 
					false && 
					doubleBuffer &&
					((child.modifiers & DontBuffer) == 0) &&
					(g.getSurfaceType() != ISurface.IMAGE_SURFACE) ?
						child.bufferGraphics(g,area,0) : null;
				Graphics use = bg == null ? g : bg.getGraphics();
				if (didLock) child.readLock();
				PixelBuffer pb = null;
				try{
							pb = doPaint(child,use,area);
				}finally{
					if (didLock) child.unlock();
				}
				int rx = r.x, ry = r.y, rw = r.width, rh = r.height;
				int cx = curClip.x, cy = curClip.y, cw = curClip.width, ch = curClip.height;
				doPaintChildren(child,flags|child.modifiers,
						pb == null ? use : pb.getCurrentDrawingBuffer(),
						area,area,curClip,r);//inter.x,inter.y,inter.width,inter.height);
				if (pb != null){
					pb.putDrawingBuffer(pb.PUT_SET);
					pb.draw(use,area.x,area.y,0);
					pb.free();
				}
				curClip.set(cx,cy,cw,ch);
				r.set(rx,ry,rw,rh);
				if (bg != null) child.releaseBufferedGraphics(bg);
				if (clip != null) g.setClip(clip.x,clip.y,clip.width,clip.height);
				g.translate(-r.x,-r.y);
			}
			area.set(ax,ay,aw,ah);
		}finally{
			//cc.cache();
			//r.cache();
		}
	}
	/**
	 * Repaint the "data" part of the Control.

	This is only used for Controls with the HasData modifier. These are controls that distinguish between
	their data (e.g. contained text) and non-data (e.g. border) on-screen display.

		
	 */
//	-------------------------------------------------------------------
	protected final void doPaintData()
//	-------------------------------------------------------------------
	{
		if ((modifiers & Transparent) != 0){
			repaintNow();
			return;
		}
		if (!Gui.requestPaint(this)) {
			return;
		}
		int flags = getModifiers(true);
		if ((flags & Invisible) != 0) return;
		BufferedGraphics bg = bufferGraphics(null,0,0,width,height,0);
		if (bg == null) return;
		if (!bg.canCopyFromOriginal()) {
			bg.cancel();
			repaintNow();
			return;
		}else
			bg.copyFromOriginal();
		try{
			Graphics g = bg.getGraphics();
			g.setFont(getFont());
			boolean didLock = lockPaint;
			if (didLock) readLock();
			try{
				doPaintData(g);
			}finally{
				if (didLock) unlock();
			}
		}finally{
			releaseBufferedGraphics(bg);
		}
	}

	/**
	 * Repaint the "data" part of the Control.

	This is only used for Controls with the HasData modifier. These are controls that distinguish between
	their data (e.g. contained text) and non-data (e.g. border) on-screen display.
		
	 * @param g The Graphics object to paint to.
	 */
//	-------------------------------------------------------------------
	protected void doPaintData(Graphics g)
//	-------------------------------------------------------------------
	{
		
	}
	/**
	 * Repaint the "data" part of the Control now.

	This is only used for Controls with the HasData modifier. These are controls that distinguish between
	their data (e.g. contained text) and non-data (e.g. border) on-screen display.

	If the Control does not have the HasData modifier, then the entire control will be repainted via repaintNow().
	 */
//	===================================================================
	public void repaintDataNow()
//	===================================================================
	{
		if ((modifiers & HasData) == 0) {
			repaintNow(); 
		}else{
			doPaintData();
		}
	}
	/**
	 * The default version of this method simpy repaints the control - however for controls
	 * which may display multiple images a better refresh scheme would be needed.
	 * @param image the image to refresh.
	 * @param options refresh options.
	 */
//	===================================================================
	public void refresh(IImage image, int options)
//	===================================================================
	{
		repaintNow();
	}

	/*###############################################################
	 * 
	 * Dealing with pen/mouse events
	 * 
	 ###############################################################*/
	/**
	* Use this to control the resolution of the pen drags - it is the minimum delay in milliseconds
	* between processing pen drags. Set it to zero to get the highest
	* resolution for dragging - best for when creating applications for drawing on the screen.
	* Set it to a higer value to get less drag events and improve response. By default it is 100.
	**/
	public int dragTime = 100;
	/**
	Returns true if a ControlPopupForm is attached to this Control.
	**/
//	===================================================================
	public boolean hasPopupFormAttached()
//	===================================================================
	{
		if (listeners == null) return false;
		for (int i = 0; i<listeners.size(); i++)
			if (listeners.get(i) instanceof ControlPopupForm) return true;
		return false;
	}
	/**
	* This is used to indicate that the control is going to show a pop-up Frame of some sort (e.g. for a Menu)
	* in reaction to the specified PenEvent.
	**/
//	===================================================================
	public boolean willShowFrame(PenEvent ev)
//	===================================================================
	{
		if (ev.type == ev.PEN_DOWN) return hasPopupFormAttached();
		return false;
	}
	/**
	 * If this returns true then the Pen Hold Down indicator will be
	 * displayed before a Popup menu is shown.
	 * @param x the x position on the Control.
	 * @param y the y position on the Control.
	 * @return true if the Control wants the Pen Hold Down indicator shown.
	 */
	protected boolean showHoldDownIndicator(int x, int y)
	{
		return ((menuState != null) && ((modifiers & WantHoldDown) != 0));
	}
	/**
	 * @deprecated - use getPosInParent(Container,Point) instead.
	 * @param parent
	 * @return
	 */
	public Point getPosInParent(Container parent)
	{
		return getPosInParent(parent,Point.getCached(0,0));
	}
//	==================================================================
	public Point getPosInParent(Container parent,Point dest)
//	==================================================================
	{
		Point p = getLocation(dest);
		int x = p.x,  y = p.y;
		for (Control c = getParent(); c != parent && c != null; c = c.parent){
			c.getLocation(p);
			x += p.x; y += p.y;
		}
		return p.set(x,y);
	}
//	==================================================================
	public void makeFrameTopMost()
//	==================================================================
	{
		//if (true) return;
		Frame f = getFrame();
		if (this instanceof Frame) f = (Frame)this;
		if (f == null) return;
		Container mc = f.parent;
		if (mc == null) return;
		mc.makeFrameTopMost();
		ControlIterator ci = (ControlIterator)getCached(ControlIterator.class);
		try{
			for (Iterator it = mc.getChildren(ci,true); it.hasNext();){
				Object obj = it.next();
				if (obj == f) return; //Already on top.
				if (obj instanceof Frame) break;
			}
		}finally{
			cache(ci);
		}
		mc.remove(f);
		mc.add(f);
		int got = f.modify(Invisible,0);
		Rect r = Gui.getAppRect(f);
		f.eraseSavedScreen();
		Gui.refreshScreen(getWindow(),r);
		f.resetRect();
		f.restore(got,Invisible);
		f.repaintNow();//Gui.refreshScreen(r);
	}
	/**
	 * This is used to transfer a pen press to a Control that is contained
		within this child control tree of this Control.
	 * @param toChild The child control to pass the pen press to.
	 */
//	===================================================================
	public final void transferPenPress(Control toChild)
//	===================================================================
	{
		if (toChild == null) return;
		int dx = 0, dy = 0;
		for (Control c = toChild; c != this && c != null; c = c.getParent()){
			dx += c.x;
			dy += c.y;

		}
		transferPenPress(toChild,-dx,-dy);
	}
	/**
	 * This is used to transfer a pen press to another Control.
	* @param toWho The Control to send it to.
	* @param dx The change in the x co-ordinate for the event.
	* @param dy The change in the y co-ordinate for the event.
	*/
//	===================================================================
	public final void transferPenPress(Control toWho,int dx,int dy)
//	===================================================================
	{
		modify(0,GotPenDown);
		getWindow().transferPenPress(Gui.currentPenPress(),toWho,dx,dy);
	}
	
//	==================================================================
	public void onLabelPenEvent(PenEvent ev)
//	==================================================================
	{
		if (!(this instanceof Container)){
			ev.x = 2; ev.y = 2;
			onPenEvent(ev);
		}else{
			Control c = getNextKeyFocus(null,true);
			if (c != null) Gui.takeFocus(c,ByKeyboard);
		}
	}
	/**
	 * Call this in response to a Pen Down event to convert Drag
	 * events to simple Move events, and to let the release event
	 * get sent to whichever control it is released on, instead
	 * of the initially pressed control.<p>
	 * This is used for things like menu/menu bars.
	 */
	public static void setFreeDragging()
	{
		Window.penDragToMove();
	}
	/**
	 * Called to handle an incoming PenEvent. By default this dispatches
		the event to the various pen handling methods like <b>penPressed()</b>
	 * @param ev The incoming PenEvent.
	 */
	
//	==================================================================
	public void onPenEvent(PenEvent ev)
//	==================================================================
	{
		EventThreadData data = Gui.getEventThreadData();
		if (ev.type == ev.PEN_DOWN) data.currentPenPress = ev;
		int flags = getModifiers(true);
		int myFlags = getModifiers(false);
		if (checkPenTransparent(ev)) return;
		data.penPoint.move(ev.x,ev.y);
		if (ev.type == ev.PEN_DOWN) {
			if (!firstPress) return;
			holdDownCancelled = false;
			firstPress = false;
			if (willShowFrame(ev)) {
				Gui.freezeSIP(this);
			}
			Frame myFrame = getFrame();
			if (!amOnTopFrame()){
				//if (firstPress)
				//if (willShowFrame(ev)) ewe.sys.Vm.freezeSIP(this);
				boolean ma = menuIsActive();
				if (!Gui.pressedOutsideTopFrame(this,new Point(ev.x,ev.y))) 
					return;
				//firstPress = false;
				if (ma && !menuIsActive()) return;
				if (!amOnTopFrame() && (myFrame == null || !myFrame.isControlPanel)){
					return;
				}
			}
			if (myFrame == null || !myFrame.isControlPanel)		
				makeFrameTopMost();
			//Gui.frameOnTop(this);
			if (((flags & Disabled) != 0) && ((flags & AlwaysEnabled) == 0)){
				if ((flags & Invisible) == 0) Sound.beep(Sound.BEEP_GUI);
				return;
			}else if ((myFlags & NoFocus) == 0 || Gui.focusedControl() == null) {
				Gui.takeFocus(this,ByMouse);
				//debugControl.setText("Focused control: "+Gui.focusedControl().name);
			}
			penStatus = (penStatus|GotPenDown)&~DidHoldDown;
			//modify(GotPenDown,DidHoldDown);
			dragging = reusedDrag;
			dragging.cancelled = false;
			dragging.didDrag = false;
			dragging.set(ev);
			dragging.resolution = startDragResolution;
			int dx = pressPoint.x - data.penPoint.x;
			int dy = pressPoint.y - data.penPoint.y;
			if (dx > 3 || dx < -3 || dy > 3 || dy < -3) lastClick -= doubleClickTime*2;
			pressPoint.move(data.penPoint.x,data.penPoint.y);
			penPressed(data.penPoint);
			if ((myFlags & WantHoldDown) != 0) {
				boolean wasC = holdDownCancelled;
				holdDownCancelled = false;
				if (!wasC) 
					//holdTicker.requestTick(this,holdDownPause);
					HoldDownDisplay.hd.start(this,ev.x,ev.y,false);
			}
		}else if (ev.type == ev.PEN_UP) {
			//holdId = null;
			//holdTicker.cancelTick(this);
			if (Window.dragAndDropData != null){
				Window.dragOverNew(this);
				boolean accepted = false;
				final Control c = Window.dragAndDropSource instanceof Control ? ((Control)Window.dragAndDropSource) : null;
				final DragContext dc = c == null ? null : c.dragging;
				if (acceptsData(Window.dragAndDropData,dc)){
					DragContext dc2 = dc;
					if (dc2 == null) {
						dc2 = new DragContext();
						dc2.dropOptions = null;
					}
					if (dc2.dropOptions != null){ 
						// The only way to be in here is if dc is the actual c.dragging - in which case c != null.
						/*
						PenEvent.resetCursor();
						final Menu m = new Menu(); 
						for (int i = 0; i<dc.dropOptions.length; i++){
							m.addItem(dc.dropOptions[i]);
						}
						m.addItem("-");
						m.addItem(new MenuItem("Cancel","cancel"));
						final Object theData = Window.dragAndDropData;
						final Point dp = new Point(data.penPoint.x,data.penPoint.y);
						ddMenuUp = true;
						m.modify(KeepSIP,0);
						m.exec(this,data.penPoint,new EventListener(){
							public void onEvent(Event ev){
								if (ev.target != m) return;
								if (ev.type == MenuEvent.SELECTED){
									m.close();
									ddMenuUp = false;
									MenuItem got = (MenuItem)((MenuEvent)ev).selectedItem;									
									if (got.action.equals("cancel")){
										c.dataTransferCancelled(theData);
										dataTransferCancelled(null);
									}else{
										dc.dropAction = got.action;
										if (dataDroppedOn(theData,dp,dc)){
											c.dataAccepted(Control.this,theData,got.action);
										}else {
											c.dataTransferCancelled(theData);
											dataTransferCancelled(null);
										}
									}
								}else if (ev.type == MenuEvent.ABORTED){
									m.close();
									ddMenuUp = false;
									c.dataTransferCancelled(theData);
									dataTransferCancelled(null);
								}
							}
							});*/
						return;
					}else{
						String action = "copy";
						if (c != null){
							if (c.dragging.dropAction == null) 
								action = c.dragging.dropAction = c.dragging.isCopy ? "copy" : "move";
						}
						if (dataDroppedOn(Window.dragAndDropData,data.penPoint,c == null ? null : c.dragging))
							if (c != null){
								c.dataAccepted(this,Window.dragAndDropData,action);
								accepted = true;

							}
					}
				}
				if (!accepted) {
					c.dataTransferCancelled(Window.dragAndDropData);
					dataTransferCancelled(null);
				}
				Window.resetCursor();
				return;
			}
			if (((penStatus & GotPenDown) == 0) && (!WindowEventThread._freeDragging)) {
				return; //(!hasModifier(GotPenDown,false)) return;
			}
			penStatus &= ~GotPenDown;//modify(0,GotPenDown);
//	..................................................................
			long cc = System.currentTimeMillis();
			boolean doDouble = false;
			if ((int)(cc-lastClick) <= doubleClickTime && lastClickedControl == this) {
				lastClickedControl = null;
				doDouble = true;
			}else{
				lastClick = cc;
				lastClickedControl = this;
			}
//	..................................................................
			if (dragging != null && dragging.didDrag && hasModifier(WantDrag,false)){
				dragging.curPoint.move(data.penPoint.x,data.penPoint.y);
				dragging.penEvent = ev;
				stopDragging(dragging);
			}else {
				if ((ev.modifiers & ev.RIGHT_BUTTON) != 0){
					penRightReleased(data.penPoint);
				}else{
					if ((penStatus & DidHoldDown) != 0) //(hasModifier(DidHoldDown,false)) 
						penReleased(data.penPoint);
					else if (doDouble) penDoubleClicked(data.penPoint);
					else penClicked(data.penPoint);
				}				
			}
		}else if (ev.type == ev.PEN_DRAG){
			if (Window.nextMouseIsMove() && dragTime > 0){
				if (System.currentTimeMillis() - ev.timeStamp > dragTime) return;
			}
			if (Window.dragAndDropData != null){
				Window.dragOverNew(this);
				dataDraggedOver(Window.dragAndDropData,data.penPoint,ev);
				return;
			}
			if ((penStatus & GotPenDown) == 0) return;
			if (!dragging.hasDragged(data.penPoint)) return;

			//holdTicker.cancelTick(this);
			if (hasModifier(WantDrag,false) && !menuIsActive()){
				if (dragging.cancelled) return;
				dragging.resolution = dragResolution;
				dragging.curPoint.move(data.penPoint.x,data.penPoint.y);
				dragging.penEvent = ev;
				long now = System.currentTimeMillis();
				if (!dragging.didDrag) {
					dragging.lastDragTime = now;
					startDragging(dragging);
				}else{
					if (now < dragging.lastDragTime || now-dragging.lastDragTime>=dragging.rate){
						dragging.lastDragTime = now;
						dragged(dragging);
					}
				}
				dragging.prevPoint.set(dragging.curPoint);
			}
			dragging.didDrag = true;
		}else if (ev.type == ev.PEN_MOVED_ON && ((myFlags & MouseSensitive) != 0)){
			if ((flags & Disabled) == 0){
				penStatus |= PenIsOn;
				if (amOnTopFrame()) repaintNow();
			}
		}else if (ev.type == ev.PEN_MOVED_OFF && ((myFlags & MouseSensitive) != 0)){
			if ((flags & Disabled) == 0){
				penStatus &= ~PenIsOn;
				if (amOnTopFrame()) repaintNow();
			}
		}else if (ev.type == ev.SCROLL_UP || ev.type == ev.SCROLL_DOWN){
			if (this instanceof ScrollClient){
				ScrollClient sc = (ScrollClient)this;
				if (sc.getServer() != null){
					sc.doScroll(sc.needScrollBar(IScroll.Vertical,height) ? IScroll.Vertical : IScroll.Horizontal,ev.type == ev.SCROLL_UP ? IScroll.ScrollLower : IScroll.ScrollHigher, 0);
					return;
				}
			}
			if (parent != null) parent.onPenEvent(ev);
		}else if (ev.type == ev.MOUSE_HOVER_OVER || ev.type == ev.MOUSE_STOP_HOVER){
			mouseHover(data.penPoint,ev.modifiers,ev.type == ev.MOUSE_HOVER_OVER);
		}
	}
	
//	-------------------------------------------------------------------
	private final boolean fixFocus(Event ev)
//	-------------------------------------------------------------------
	{
		if (hasModifier(TakeControlEvents,false) && (ev.target != null) && (ev.target instanceof Control))
			if (contains((Control)ev.target) || ev.target == this){
				if (!(ev instanceof ControlEvent)) return false;
				if (hasModifier(TakeControlEvents,false) && (ev.target != null) && (ev.target instanceof Control))
				if (ev.type != ControlEvent.FOCUS_IN && ev.type != ControlEvent.FOCUS_OUT) return false;
				if (contains(((ControlEvent)ev).oldOrNewFocus) || ((ControlEvent)ev).oldOrNewFocus == this) return true;
			}
		return false;
	}
	
	protected void menuStateChanged(boolean menuOpened)
	{
		
	}
//	==================================================================
	public void onControlEvent(ControlEvent ev)
//	==================================================================
	{
		if (ev.type == ev.POPUP_CLOSED || ev.type == ev.MENU_SHOWN) 
			menuStateChanged(ev.type == ev.MENU_SHOWN);
		else
		if (hasModifier(TakeControlEvents,false) && (ev.target != null) && (ev.target instanceof Control))
			if (contains((Control)ev.target)){
				if (!fixFocus(ev)) {
					ev.target = this;
					if ((ev instanceof MenuEvent) && (ev.type == MenuEvent.SELECTED || ev.type == MenuEvent.ABORTED)){
						Object it = ((MenuEvent)ev).selectedItem;
						if (it instanceof MenuItem) lastSelected = (MenuItem)it;
					}
					sendToListeners(ev);
				}
			}
	}
//	-------------------------------------------------------------------
	protected MenuItem lastSelected = null;
//	-------------------------------------------------------------------
	/**
	* Find the menu item which was last selected. This may be from a sub menu.
	**/
//	===================================================================
	public MenuItem getLastSelected()
//	===================================================================
	{
		return lastSelected;
	}
	
	protected boolean isOnMe(int x, int y)
	{
		if (x<0 || x>=width) return false;
		return !(y<0 || y>= height);
	}
	/**
	 * Checks if the Point p - which is relative to the top-left of this Control, is within the bounds
		of the Control.
	 * @param p The point to check.
	 * @return true if the point is within the Control.
	 */
//	==================================================================
	protected boolean isOnMe(Point p)
//	==================================================================
	{
		if (p.x<0 || p.x>=width) return false;
		return !(p.y<0 || p.y>= height);
	}
	protected boolean isOnMe(Point p, boolean orOnMyPrompt)
	{
		if (p.x>=0 && p.x < width && p.y >= 0 && p.y < height) {
			return true;
		}
		if (!orOnMyPrompt || promptControl == null) return false;
		Point d = Point.getCached(0,0);
		try{
			getPosInFrame(d);
			int xx = p.x+d.x, yy = p.y+d.y;
			promptControl.getPosInFrame(d);
			return (xx >= d.x && xx < d.x+promptControl.width && yy >= d.y && yy < d.y+promptControl.height);
		}finally{
			d.cache();
		}
	}
	/**
	Called when the Pen or Mouse button is pressed on the control.
	This is called by the onPenEvent(PenEvent ev) method.<p>
	Within this method you can use the Control.Gui.currentPenPress() value
	to find out more about the PenEvent that caused this - including any
	modifiers (e.g. SHIFT keys) or determine if the left or right mouse
	button was pressed.<p>
 @param p The point on the Control where the pen was pressed.
 */
public void penPressed(Point p) {}
/**
	Called when the Pen or Mouse button is held down on the control. This will
	only get called if the WantHoldDown modifier has been set for this control.
	This is called by the onPenEvent(PenEvent ev) method.<p>
	Within this method you can use the Control.Gui.currentPenPress() value
	to find out more about the PenEvent that caused this - including any
	modifiers (e.g. SHIFT keys) or determine if the left or right mouse
	button was pressed.<p>
	This method is usually used to display a context sensitive menu. In fact
	the default code for it is:
	<pre>
		if (!menuIsActive()) doMenu(p);
	</pre>
 @param p The point on the Control where the pen was held.
 */
public void penHeld(Point p){if (!menuIsActive()) doMenu(p);}
/**
	Called when the right Mouse button is released on the control.
	This is called by the onPenEvent(PenEvent ev) method.<p>
	Within this method you can use the Control.Gui.currentPenPress() value
	to find out more about the PenEvent that caused this - including any
	modifiers (e.g. SHIFT keys) or determine if the left or right mouse
	button was pressed.<p>
	This method is usually used to display a context sensitive menu. In fact
	the default code for it is:
	<pre>
		if (!menuIsActive()) 
			if (doMenu(p)) return;
		penReleased(p);
	</pre>
 * @param p The point on the Control where the pen was released.
 */
public void penRightReleased(Point p)
{
	if (!menuIsActive()) 
		if (doMenu(p)) return;
	penReleased(p);
}
	
public void mouseHover(Point p, int modifiers, boolean isHover)
{
	
}
	/**
	Called when the Pen or Mouse button is released on the control.
	This is called by the onPenEvent(PenEvent ev) method.<p>
	Within this method you can use the Control.Gui.currentPenPress() value
	to find out more about the PenEvent that caused this - including any
	modifiers (e.g. SHIFT keys) or determine if the left or right mouse
	button was pressed.<p>

 @param p The point on the Control where the pen was released.
 */
public void penReleased(Point p) {}
/**
	Called if the Pen or Mouse button is quickly pressed and released.
	This is called by the onPenEvent(PenEvent ev) method.<p>
	Within this method you can use the Control.Gui.currentPenPress() value
	to find out more about the PenEvent that caused this - including any
	modifiers (e.g. SHIFT keys) or determine if the left or right mouse
	button was pressed.<p>
	By default this method simply calls the penReleased() method.

 @param p The point on the Control where the pen was clicked.
 */
public void penClicked(Point p) {penReleased(p);}
/**
	Called if the Pen or Mouse button is double clicked on the Control.
	This is called by the onPenEvent(PenEvent ev) method.<p>
	Within this method you can use the Control.Gui.currentPenPress() value
	to find out more about the PenEvent that caused this - including any
	modifiers (e.g. SHIFT keys) or determine if the left or right mouse
	button was pressed.<p>

	By default this method simply calls the penClicked() method.
 @param p The point on the Control where the pen was double-clicked.
 */
public void penDoubleClicked(Point p) {penClicked(p);}
	
/*###############################################################
 * 
 * Dealing with HotKeys
 * 
 ###############################################################*/
	
/**
 * This sets both the border style and width of the Control
 * @param style The border style. One or more of the BDR_XXX or EDGE_XXX constants ORed together.
 * @param width The distance between the edge of the Control and the data within it.
 * @return this Control
 */
//===================================================================
public final Control setBorder(int style,int width)
//===================================================================

{
	borderWidth = width;
	borderStyle = style;
	return this;
}
	/**
	 * Set the hot-key for this Control.
	 * @param modifiers One or more of the IKeys.SHIFT, IKeys.CONTROL and IKeys.ONLY ORed together.
	 * @param key The character for the hot-key.
	 * @return this Control
	 */
//	===================================================================
	public final Control setHotKey(int modifiers,char key)
//	===================================================================
	{
		hotKey = KeyEvent.toKey(modifiers,Vm.getLocale().changeCase(key,false));
		return this;
	}
	/**
	 * Set the hot-key for this Control.
	 * @param modifiers One or more of the IKeys.SHIFT, IKeys.CONTROL and IKeys.ONLY ORed together.
	 * @param key The KeyEvent key code for the hot-key.
	 * @return this Control
	 */
//	===================================================================
	public final Control setHotKey(int modifiers,int key)
//	===================================================================
	{

		hotKey = KeyEvent.toKey(modifiers,key);
		return this;
	}
	/**
	* This method checks to see if a KeyEvent should be considered the Control's hot-key. If it
	* is it calls doAction(ByKeyboard).
	* @param from The Control (parent or sibling) that is passing the event to this Control.
	* @param ev The KeyEvent representing the key press.
	* @return true if the hot-key was accepted and handled.
	*/
//	-------------------------------------------------------------------
	protected boolean doHotKey(Control from,KeyEvent ev)
//	-------------------------------------------------------------------
	{
		int flags = getModifiers(true);
		boolean en = (((flags & Disabled) == 0)  || ((flags & AlwaysEnabled) != 0));
		if (ev.isHotKey(hotKey) && en){
			doAction(ByKeyboard);
			notifyAction();
			return true;
		}
		return false;
	}
//	===================================================================
	public boolean tryNext(boolean forwards)
//	===================================================================
	{
		if (parent == null) return false;
		Control c = parent.getNextKeyFocus(this,forwards);
		if (c == null) return false;
		Gui.takeFocus(c,ByKeyboard);
		return true;
	}

	/**
	 * Format the String as a true Hotkey encoded string using the hotkey associated with this control (if any).
	 * @param label The label to be displayed.
	 * @return A hotkey encoded label for display.
	 * @deprecated use makeHot(String,CharArray) instead.
	 */
//	===================================================================
	public final String makeHot(String label)
//	===================================================================
	{
		return Gui.makeHot(label,hotKey);
	}
	/**
	 * Format the String as a true Hotkey encoded string using the hotkey associated with this control (if any).
	 * @param label The label to be displayed.
	 * @param dest an optional destination CharArray.
	 * @return A hotkey encoded label for display in a CharArray.
	 */

	public final CharArray makeHot(String label, CharArray dest)
	{
		return Gui.makeHot(label,hotKey,dest);
	}
	public int [] exitKeys;
	
//	===================================================================
	public void doActionKey(int key) {doAction(ByKeyboard); notifyAction();}
//	===================================================================
	
	/**
	* Determine which sub-control should receive the keyboard focus. This method is normally
	* called when the user moves the focus off of one control using TAB or the cursor keys.
	* A control which has no children should not look for sub controls.
	* @param sourceChild the child control from which the call came from, or null if the call
	* came from the parent of this control, or this control itself.
	* @param forwards true if the user wants to go to the next control, false if the user wants
	* to go to the previous one.
	* @return the control (which may be this control) that should receive the focus, or null
	* if none should.
	*/
//	===================================================================
	public Control getNextKeyFocus(Control sourceChild,boolean forwards)
//	===================================================================
	{
		if ((getModifiers(true) & (Disabled|DisplayOnly)) != 0) return null;
		if ((modifiers & TakesKeyFocus) != 0) return this;
		return null;
	}

	private Gate exitLock;
	/**
	* If this returns false it means it is already in exitEntry.
	**/
//	===================================================================
	public final boolean exitEntry(int exitKey,int modifiers)
//	===================================================================
	{
		synchronized(this){
			if (exitLock == null) {
				exitLock = new Gate();
				exitLock.multipleEntry = false;
			}
		}
		if (!exitLock.grab()) return false;
		try{
			Control c = Gui.focusedControl();
			if (c == this || contains(c)) Gui.takeFocus(null,ByRequest);
			ControlEvent ce = new ControlEvent(ControlEvent.EXITED,this);
			ce.exitKey = exitKey;
			ce.exitKeyModifiers = modifiers;
			postEvent(ce);
		}finally{
			exitLock.unlock();
		}
		return true;
	}

//	-------------------------------------------------------------------
	protected final boolean checkExitKey(KeyEvent ev)
//	-------------------------------------------------------------------
	{
		if (exitKeys != null){
			for (int i = 0; i<exitKeys.length; i++){
				if (exitKeys[i] == ev.key){
					exitEntry(ev.key,ev.modifiers);
					return true;
				}
			}
		}
		if (parent != null) return parent.checkExitKey(ev);
		return false;
	}

//	-------------------------------------------------------------------
	protected boolean isSomeonesHotKey(KeyEvent ev)
//	-------------------------------------------------------------------
	{
		if (ev.type != ev.KEY_PRESS) return false;
		if (doHotKey(this,ev)) return true;
		if (parent == null) return false;
		return parent.doHotKey(this,ev);
	}
	/**
	* This checks the incoming KeyEvent to see if it should cause this Control to show a
	* menu. If it does then the Menu is shown and this returns true. Otherwise it returns false.
	**/
//	===================================================================
	public boolean checkMenuKey(KeyEvent ev,Point where)
//	===================================================================
	{
		if (ev.type != ev.KEY_PRESS || ev.key != IKeys.MENU || menuIsActive()) return false;
		return doMenu(where);
	}
	/**
	This is the standard Control onKeyEvent() - call this only after you have checked for
	other keys that your control definitely wants to trap. This method will check for an
	exit key, a menu key, will check for hot keys and check for keys that may cause a
	shift to another control.
	**/
//	===================================================================
	public void standardOnKeyEvent(KeyEvent ev)
//	===================================================================
	{
		if (ev.type != ev.KEY_PRESS) return;
		ev.consumed = true;
		if (checkExitKey(ev)) return;	
		if (checkMenuKey(ev,zeroPoint)) return;
		
		if ((ev.modifiers & IKeys.CONTROL) != 0) {
			ev.controlToLetter();
			if (!doHotKey(this,ev)){
						if (parent != null) 
							if (!parent.doHotKey(this,ev))
								ev.consumed = false;
							else
								return;
					
			}else return;
		}else if (ev.key == IKeys.ENTER || ev.key == IKeys.ESCAPE){
			if (isSomeonesHotKey(ev))
				return;
		}
		{
			if (ev.key == IKeys.PREV_CONTROL || ev.key == IKeys.UP || ev.key == IKeys.LEFT || (ev.key == IKeys.TAB && ((ev.modifiers & IKeys.SHIFT) != 0))){
				tryNext(false);
			}else if (ev.key == IKeys.NEXT_CONTROL || ev.key == IKeys.DOWN || ev.key == IKeys.RIGHT || (ev.key == IKeys.TAB && ((ev.modifiers & IKeys.SHIFT) == 0))){
				tryNext(true);
			}else if (ev.isActionKey()){
				int flags = getModifiers(true);
				if (((flags & Disabled) == 0)  || ((flags & AlwaysEnabled) != 0))
					doActionKey(ev.key);
			}else if (!doHotKey(this,ev)){
				if (parent != null) 
					if (!parent.doHotKey(this,ev))
						ev.consumed = false;
			}
		}
	}
//	==================================================================
	public void onKeyEvent(KeyEvent ev)
//	==================================================================
	{
		standardOnKeyEvent(ev);
	}
	/*###############################################################
	 * 
	 * Data dragging and transfer.
	 * 
	 ###############################################################*/
//	==================================================================
	public int holdDownPause = 500; //Fifth of a second.
	public int holdTick = 50; //Twenty times a second.
	public  DragContext dragging;
//	==================================================================

	private static boolean holdDownCancelled = false;
	//private Object holdId;
	
	public static void cancelHoldDown()
	{
		holdDownCancelled = true;
	}
	void heldDown()
	{
		int myFlags = getModifiers(false);		
		EventThreadData data = Gui.getEventThreadData();
		if ((penStatus & GotPenDown) != 0) {//((myFlags & GotPenDown) != 0) {
			if (((myFlags & WantDrag) != 0) && dragging.didDrag){
				if (!isOnMe(data.penPoint) && Window.dragAndDropData == null){
					dragging.curPoint.move(data.penPoint.x,data.penPoint.y);
					//dragging.penEvent = ev;
					dragged(dragging);
				}
			}else if (amOnTopFrame()){//Gui.focusedControl() == this){
				int s = Gui.getAsyncKeyState(IKeys.PEN);
				if ((s & 0x8000) != 0){
					wantAnotherPenHeld = false;
					penHeld(data.penPoint);
					penStatus |= DidHoldDown;
					//modify(DidHoldDown,0);
					if (wantAnotherPenHeld)
						HoldDownDisplay.hd.start(this,data.penPoint.x,data.penPoint.y,true);
						//holdTicker.requestTick(Control.this,holdTick);
				}
			}

		}
	}
//	==================================================================
	public void ticked(Object id,long elapsed)
//	==================================================================
	{
		/*
		if (id == holdTicker){
			if (holdDownCancelled){
				holdDownCancelled = false;
				return;
			}
			final int myFlags = getModifiers(false);
			try{
				new WindowThreadTask(){
					protected void doRun(){
						EventThreadData data = Gui.getEventThreadData();
						if ((penStatus & GotPenDown) != 0) {//((myFlags & GotPenDown) != 0) {
							if (((myFlags & WantDrag) != 0) && dragging.didDrag){
								if (!isOnMe(data.penPoint) && Window.dragAndDropData == null){
									dragging.curPoint.move(data.penPoint.x,data.penPoint.y);
									//dragging.penEvent = ev;
									dragged(dragging);
								}
							}else if (amOnTopFrame()){//Gui.focusedControl() == this){
								int s = Gui.getAsyncKeyState(IKeys.PEN);
								if ((s & 0x8000) != 0){
									wantAnotherPenHeld = false;
									penHeld(data.penPoint);
									penStatus |= DidHoldDown;
									//modify(DidHoldDown,0);
									if (wantAnotherPenHeld) holdTicker.requestTick(Control.this,holdTick);
								}
							}
			
						}
					}
				}.run(this,false);
			}catch(Exception e){}
		}//else super.ticked(id,elapsed);
		*/
	}
	/**
	* This is called by a drop from a drag and drop operation. By default
	* it will call takeData() unless it is from itself, in which case
	* the data is rejected and this returns false.
	**/
//	===================================================================
	public boolean dataDroppedOn(Object data,Point p,DragContext dc)
//	===================================================================
	{
		if (Window.dragAndDropSource == this) return false;
		return takeData(data,dc);
	}
	/**
	 * This is called when data that was previously dragged into the control
		has now been dragged off.
	 * @param data the data being dragged.
	 */
//	===================================================================
	public void dataDraggedOff(Object data){}
//	===================================================================
//	===================================================================
	public static DragContext getDragAndDropContext()
//	===================================================================
	{
		if (Window.dragAndDropSource instanceof Control)
			return ((Control)Window.dragAndDropSource).dragging;
		return null;
	}

	/**
	 * This is called when data is being dragged into the control.
	Within this method the control should call either willAcceptDrop()
	or dontAcceptDrop() to indicate its willingness to accept the data.<p>
	By default this method calls <b>dataDraggedOver(Object data)</b> -
	which itself by default calls <b>acceptsData()</b> to determine whether
	to call willAcceptDrop() or dontAcceptDrop().
	 * @param data The data being dragged.
	 * @param p The current location of the pen/mouse.
	 * @param ev The PenEvent that prompted this call.

	 */
//	===================================================================
	public void dataDraggedOver(Object data,Point p,PenEvent ev)
//	===================================================================
	{
		dataDraggedOver(data);
	}
	/**
	 * This is called when data is being dragged into the control.
	Within this method the control should call either willAcceptDrop()
	or dontAcceptDrop() to indicate its willingness to accept the data.<p>
	By default this method calls <b>acceptsData()</b> to determine whether
	to call willAcceptDrop() or dontAcceptDrop().
	 * @param data The data being dragged.
	 */
//	===================================================================
	public void dataDraggedOver(Object data)
//	===================================================================
	{
		if (Window.dragAndDropSource == this) willAcceptDrop();
		else if (acceptsData(data,getDragAndDropContext())) willAcceptDrop();
		else dontAcceptDrop();
	}
	/**
	 * This should indicate whether this control will accept the data either via
		 a drag and drop or via a paste operation.
	 * @param data The data to be transferred.
	 * @param how if the transfer is via DragAndDrop this will be a DragContext object,
		otherwise it will be null to indicate a clipboard operation.
	 * @return true if the data will be accepted by this control.
	 */
//	===================================================================
	public boolean acceptsData(Object data,DragContext how) 
//	===================================================================
	{return false;}
	/**
	 * This tells the Control to take the data which has either come from a clipboard operation
		or from a DragAndDrop.
	 * @param data The data to accept.

	 * @param how if the transfer is via DragAndDrop this will be a DragContext object,
		otherwise it will be null to indicate a clipboard operation.
	 * @return true if the data was successfully accepted.
	 */
//	===================================================================
	public boolean takeData(Object data,DragContext how)
//	===================================================================
	{
		if (!acceptsData(data,how)) return false;
		if (this instanceof Selectable){
			if (!((Selectable)this).replaceSelection(data)) return false;
			notifyDataChange();
			return true;
		}else
			return false;
	}

	/**
	 * Should be called from the dataDraggedOver() method to indicate that the Control is willing
		to accept the data.
	 */
//	-------------------------------------------------------------------
	protected void dontAcceptDrop()

//	-------------------------------------------------------------------
	{
		getWindow().setCursor(Cursor.DONT_DROP_CURSOR);
		//ewe.sys.Vm.setCursor(this,ewe.sys.Vm.DONT_DROP_CURSOR);
	}
	/**
	 * Should be called from the dataDraggedOver() method to indicate that the Control is not willing
		to accept the data.
	 */
//	-------------------------------------------------------------------
	protected void willAcceptDrop()
//	-------------------------------------------------------------------
	{
		//ewe.sys.Vm.setCursor(this,Window.dragAndDropCursor);
		getWindow().setCursor(Window.dragAndDropCursor);
	}
	/*
	public void penPressed(PenEvent ev) {penPressed(new Point(ev.x,ev.y));}
	public void penReleased(PenEvent ev){penReleased(new Point(ev.x,ev.y));}
	public void penClicked(PenEvent ev) {penClicked(new Point(ev.x,ev.y));}
	public void penDoubleClicked(PenEvent ev) {penDoubleClicked(new Point(ev.x,ev.y));}
	*/
	/**
	 * A quick way to provide DragAndDrop data. This will be called from tryDragAndDrop() which itself
		may be called from the startDragging() method.<p>
		To setup the DragContext you should set the dataToDrag, isMultiple and isCopy and (optionally)
		the dragCursor value.
	 * @param dc the DragContext to setup for the drag and drop operation.
	 * @return true if the data was successfully provided to drag and drop in the DragContext. false
		if there is no data to drag.
	 */
//	-------------------------------------------------------------------
	protected boolean getDataToDragAndDrop(DragContext dc)
//	-------------------------------------------------------------------
	{return false;}

	/**
	* Call this from startDragging() if you want to try to do a drag and drop.
	* @param dc the DragContext to setup for the drag and drop operation.
	* @return 
	*/
//	-------------------------------------------------------------------
	protected boolean tryDragAndDrop(DragContext dc)
//	-------------------------------------------------------------------
	{
		dc.isCopy = ((dc.modifiers & IKeys.CONTROL) != 0);
		if (!getDataToDragAndDrop(dc)){
			dc.cancelled = true;
			return false;
		}
		if (dc.dragCursor != 0) startDragAndDrop(dc.dataToDrag,dc.dragCursor,dc.isCopy);
		else startDragAndDrop(dc.dataToDrag,dc.isMultiple,dc.isCopy);
		return true;
	}
	/**
	 * This is called to indicate the start of a pen/mouse drag operation. This
		will only get called if the WantDrag modifier has been set for this control.
	Within this method you can call startDragAndDrop() or tryDragAndDrop() to
	begin a data DragAndDrop operation.
	 * @param dc A DragContext indicating information about the dragging.
	 */
//	===================================================================
	public void startDragging(DragContext dc) 
//	===================================================================
	{dragged(dc);}
	/**
	 * Called when the pen was released during a drag operation. This will only
		be called if a data DragAndDrop was not initiated and if the cancelled member
		of the DragContext was not set true during the operation.
	 * @param dc The DragContext for the dragging operation.
	 */
//	===================================================================
	public void stopDragging(DragContext dc) 
//	===================================================================
	{penReleased(dc.curPoint);}
	/**
	 * This is called during the drag operation. It is also called by default
		by the startDragging() method if you do not override it.
	 * @param dc The DragContext for the dragging operation.
	 */
//	===================================================================
	public void dragged(DragContext dc) {}
//	===================================================================
	/**
	* A temporary holding spot for data in the process of being moved.
	**/
	protected static Object beforeRemoved;
	/**
	* This is called either when data that was cut from this control was
	* pasted into another control, OR when data that was dragged from this
	* control is dropped into another control.<p>
	* @param byWho The Control that accepted the data.
	* @param data The data being transferred.
	* @param action This is either "copy" or "move" indicating if the data
	is meant to be copied to the other control or moved to the other Control.
	*/
//	-------------------------------------------------------------------

	protected void dataAccepted(Control byWho,Object data,String action)
//	-------------------------------------------------------------------
	{
		if (action.equalsIgnoreCase("copy")) dataTransferCancelled(data);
	}
	/**
	* This gets called either when that was cut from this control was not
	* placed into another control (either it was rejected by a paste or a
	* subsequent copy/cut into the clipboard occured before the data was pasted)
	* or a drag and drop was initiated but not completed.
	* @param data The data that was being transferred. If it is null it indicates
	that this Control was a <b>destination</b> for a Drag and Drop operation, but the
	operation was cancelled by the user.
	*/
//	-------------------------------------------------------------------
	protected void dataTransferCancelled(Object data)
//	-------------------------------------------------------------------
	{
	}
	/**
	* This tells the control that data is being removed from it either via
	* a cut operation or by a drag operation that is not explicitly a copy operation.
	* If the object is a Selectable object, then by default this will call
	* delete selection. If you want more advanced operations (e.g. graying
	* out the selection) you need to override this. If DragContext is null, it
	* implies that it is via a "cut" operation, otherwise it is via a drag.
	**/
//	-------------------------------------------------------------------
	protected void dataBeingRemoved(Object data,DragContext dc)
//	-------------------------------------------------------------------
	{
		if (this instanceof Selectable && canEdit())
			((Selectable)this).deleteSelection();
	}
	/**
	* Called by a Control from within the startDragging() method to indicate that
	* a DragAndDrop operation should begin.
	* @param data The data to drag.
	* @param dragCursor The cursor code for the special cursor to use. This should
	be one of the ewe.sys.Vm.XXX_CURSOR constants OR a value returned by ewe.sys.Vm.createCursor().
	* @param isCopy true if the operation is to be considered a copy operation. false if it is to be
	considered a move operation.
	* @return true if successful.
	*/
//	-------------------------------------------------------------------
	protected boolean startDragAndDrop(Object data,int dragCursor,boolean isCopy)
//	-------------------------------------------------------------------
	{
		beforeRemoved = null;
		dragging.didDrag = true;
		dragging.isCopy = isCopy;
		if (!isCopy) dataBeingRemoved(data,dragging);
		return getWindow().beginDragAndDrop(this,data,dragCursor);
	}
	/**
	* Called by a Control from within the startDragging() method to indicate that
	* a DragAndDrop operation should begin.
	* @param data The data to drag.
	* @param isMultiple should be true if the data should be considered as multiple data. This
	will affect the cursor that is used.
	* @param isCopy true if the operation is to be considered a copy operation. false if it is to be
	considered a move operation. This will affect the cursor that is used.
	* @return true if successful.
	*/
//	-------------------------------------------------------------------
	protected boolean startDragAndDrop(Object data,boolean isMultiple,boolean isCopy)
//	-------------------------------------------------------------------
	{
		beforeRemoved = null;
		dragging.didDrag = true;
		dragging.isCopy = isCopy;
		dragging.isMultiple = isMultiple;
		if (!isCopy) dataBeingRemoved(data,dragging);
		return getWindow().beginDragAndDrop(this,data,isMultiple,isCopy);
	}
	/**
		Called when the Pen or Mouse button is pressed on the control.
		This is called by the onPenEvent(PenEvent ev) method.<p>
		Within this method you can use the Control.Gui.currentPenPress() value
		to find out more about the PenEvent that caused this - including any
		modifiers (e.g. SHIFT keys) or determine if the left or right mouse
		button was pressed.<p>
	 @param p The point on the Control where the pen was pressed.
	 */
	
	/*###############################################################
	 * 
	 * Menu Handling
	 * 
	 ###############################################################*/
	/**
	* This holds a MenuState Object used for controlling the Menu associated with 
	* the Control. You should not set this directly - it is set when you call <b>setMenu()</b>
	**/
	public MenuState menuState;
	/**
	* This is called before a menu is displayed. You can use it to adjust the items in the menu. If it returns
	* false the menu will not be displayed.
	**/
//	===================================================================
	public boolean checkMenu(Menu m)
//	===================================================================
	{
		return checkClipboardOperations(m);
	}
	/**
	 * This is called to enable/disable the clipboard options depending on the state of the Control.
	 * @param m The Menu which contains the clipboard operations.
	 * @return always true.
	 */
//	===================================================================
	public boolean checkClipboardOperations(Menu m)
//	===================================================================
	{
		int flags = getModifiers(true);
		boolean editable = (((flags & (Disabled|NotEditable|DisplayOnly)) == 0) || ((flags & NotAnEditor) != 0));
		if (clipItems == null) return true;
		clipItems[0].modifiers &= ~MenuItem.Disabled;
		if (editable){
			clipItems[1].modifiers &= ~MenuItem.Disabled;
			clipItems[2].modifiers &= ~MenuItem.Disabled;
		}else{
			clipItems[1].modifiers |= MenuItem.Disabled;
			clipItems[2].modifiers |= MenuItem.Disabled;
		}
		Object obj = getClipObject();
		if (obj == null) clipItems[2].modifiers |= MenuItem.Disabled;
		else if (!acceptsData(obj,null))
			clipItems[2].modifiers |= MenuItem.Disabled;
		if (this instanceof Selectable)
			if (!((Selectable)this).hasSelection()) {
				clipItems[0].modifiers |= MenuItem.Disabled;
				clipItems[1].modifiers |= MenuItem.Disabled;
			}
		return true;
	}
	/**
	 * This adds the standard clipboard menu to a menu to be used by this control.
	 * @param addTo The menu under construction or null for a new menu.
	 * @return the addTo parameter or a new Menu if it was null.
	 */
//	===================================================================
	public Menu getClipboardMenu(Menu addTo)
//	===================================================================
	{
		if (clipItems == null){
			clipItems = new MenuItem[]
			{
				MenuItem.getIconized("Copy$c","eve/copysmall.png"),
				MenuItem.getIconized("Cut$x","eve/cutsmall.png"),
				MenuItem.getIconized("Paste$v","eve/pastesmall.png"),
				//,new MenuItem("Exchange",0,null)
			};
		}
		if (addTo == null) addTo = new Menu();
		addTo.addSection(clipItems,true);
		return addTo;
	}

//	===================================================================
	public String clipboardToString(Object what)
//	===================================================================
	{
		if (what == null) return new String();
		if (what instanceof Object []){
			Object [] w = (Object []) what;
			String s = new String();
			for (int i = 0; i<w.length; i++)
				s += clipboardToString(w[i])+"\n";
			return s;			
		}else return what.toString();
	}
	/**
	* Copy the selected contents of the control to clipboard. If cut is true
	* the selected contents will be removed from the control if it is editable.
	**/
//	===================================================================
	public void toClipboard(boolean cut)
//	===================================================================
	{
		setClipObject(clipboardTransfer(getClipObject(),true,cut));
	}
	/**
	* Paste the contents of the clipboard into the control, overwriting any
	* exisiting selected area. If exchange is true the selected area in the


	* control will replace the current clipboard data.
	**/
//	===================================================================
	public void fromClipboard()
//	===================================================================
	{
		clipboardTransfer(getClipObject(),false,false);
		notifyDataChange();
	}

	/**
	 * Return whether the Control currently has an active popup menu.
	 */
	public boolean menuIsActive()
	{
		if (menuState == null) return false;
		return menuState.isOpen();
	}
	/**
	* If a menu for the control is being displayed, this will close it.
	**/
//	===================================================================
	public void closeMenu()
//	===================================================================
	{
		if (menuState != null && menuState.isOpen())
			menuState.closeMenu();
	}
	/**
	 * Associate a context sensitive menu with the Control. If there is already
		a menu associated with the Control it will be replaced by this new one.<p>
		You can set a Menu for most Controls including Buttons.
	 * @param menu The Menu to set for the Control. This can be null.
	 */
//	===================================================================
	public void setMenu(Menu menu)
//	===================================================================
	{
		if (menu == null){
			if (menuState != null){
				menuState.control = null;
				menuState.menu = null;
				menuState = null;
				return;
			}
		}
		if (menuState == null) menuState = new MenuState();
		menuState.menu = menu;
		menuState.control = this;
		menu.modify(KeepSIP,0);
	}
	/**
	 * Return the Menu associated with the Control if any.
	 * @return the Menu associated with the Control if any.
	 */
//	===================================================================
	public Menu getMenu()
//	===================================================================
	{
		return menuState == null ? null:menuState.menu;
	}
	/**
	 * This attempts to display a Menu for the Control if one exists AND is not already displayed.
		<p>By default it will call <b>doShowMenu(pen)</b> if the menu is not already active.
	 * @param pen The point where the pen/mouse was pressed - or null if you are requesting the control to show
		its menu at no particular point.
	 * @return true if the Menu was displayed by the method call.
	 */
//	===================================================================
	public boolean tryStartMenu(Point pen)
//	===================================================================
	{
		if (menuIsActive()) return false;
		if (hasModifier(DisablePopupMenu,false)) return false;
		if (pen == null) pen = new Point(0,0);
		return doShowMenu(pen);
	}
	/**
	 * This is a request to display a menu based on the pen being held or the mouse being right clicked on
		 the Control. By default this method calls <b>tryStartMenu(p);</b>
	 * @param p The point where the pen was pressed or held.
	 * @return true if a menu was displayed.
	 */
	public boolean doMenu(Point p){return tryStartMenu(p);}
	/**
	 * @param p
	 * @return
	 */
	/**
	 * Used to show the menu associated with the Control. This will be called by <b>tryStartMenu()</b> which
		is what you should use instead.
	 * @param pen The point where the pen was pressed.
	 * @return true always.
	 */
//	-------------------------------------------------------------------
	protected boolean doShowMenu(Point pen)
//	-------------------------------------------------------------------
	{
		if (menuState == null) return false;
		menuState.doShowMenu(pen);
		return true;
	}
	
	/*###############################################################
	 * 
	 * Generating and Sending Events
	 * 
	 ###############################################################*/
//	------------------------------------------------------------------
	protected Vector listeners;
	/**
	 * This adds an EventListener that can listen for Events that are <b>generated</b>
	 * by this Control - not Events that are received by this control.
	 * @param list the EventListener to add.
	 */
	public void addListener(EventListener list)
//	===================================================================
	{
		if (listeners == null) listeners = new Vector(1);
		listeners.remove(list);
		listeners.add(list);
	}
	/**
	 * This removes an EventListener added by addListener().
	 * @param list the EventListener to add.
	 */
	public void removeListener(EventListener list)
//	===================================================================
	{
		if (listeners == null) return;
		listeners.remove(list);
	}
	/**
	 * Send an Event to the the Listeners but note that PenEvents are not
	 * sent to Listeners.
	 * @param ev the Event to send.
	 */
	public void sendToListeners(Event ev)
//	===================================================================
	{
		if (listeners != null){// && (ev instanceof ControlEvent))
			for (int i = 0; i<listeners.size(); i++){
				EventListener c = (EventListener)listeners.get(i);
				if ((c instanceof Control) && (ev instanceof PenEvent)) continue;
				c.onEvent(ev);
			}
		}
	}
	/**
	 * Generate and send a standard "action" event for this Control.
	 */
	public void notifyAction()
	{
		postEvent(new ControlEvent(ControlEvent.PRESSED,this));
	}
	/**
	 * The fieldTransfer used by the fromField() and toField() methods
	 * that take no parameters.
	 */
	public FieldTransfer fieldTransfer;

	/**
	 * Transfer data from a Field (possibly in an Object) to
	 * be displayed on this Control.
	 * @param ft the FieldTransfer Object to use.
	 */
	public void fromField(FieldTransfer ft)
//	===================================================================
	{
		if (ft == null) return;
		if (ft.dataInterface != this) return;
		if (ft.converter == null) ft.converter = Gui.getDefaultConverter();
		ft.transfer(ft.FROM_OBJECT);
		updateData();
	}
	/**
	 * Transfer data from this Control to a Field (possibly in an Object).
	 * @param ft the FieldTransfer Object to use.
	 */
	public void toField(FieldTransfer ft)
//	===================================================================
	{
		if (ft == null) return;
		if (ft.dataInterface != this) return;
		if (ft.converter == null) ft.converter = Gui.getDefaultConverter();
		ft.transfer(ft.TO_OBJECT);
		//ewe.sys.Vm.debug("Transfered: "+ft.fieldName);
	}
	/**
	 * Transfer data from a Field (possibly in an Object) to
	 * be displayed on this Control
	 * using the FieldTransfer saved in the member field "fieldTransfer".
	 */
	public final void fromField() {fromField(fieldTransfer);}
	/**
	 * Transfer data from this Control to a Field (possibly in an Object)
	 * using the FieldTransfer saved in the member field "fieldTransfer".
	 */
	public final void toField() {toField(fieldTransfer);}
//	===================================================================
	
//	-------------------------------------------------------------------
	private void doChangeTransfer()
//	-------------------------------------------------------------------
	{
		if (fieldTransfer != null){
			if ((modifiers & (NotEditable|DisplayOnly|Disabled)) != 0) fromField();
			else toField();
		}
	}
	/**
	* Generate and dispatch a new DataChangeEvent.
	**/
//	===================================================================
	public final void notifyDataChange() 
//	===================================================================
	{
		notifyDataChange(new DataChangeEvent(DataChangeEvent.DATA_CHANGED,this));
	}
	/**
	* Generate and dispatch a new ControlEvent with type ControlEvent.STRUCTURE_CHANGED.
	**/
//	===================================================================
	public final void notifyStructureChanged() 
//	===================================================================
	{
		postEvent(new ControlEvent(ControlEvent.STRUCTURE_CHANGED,this));
	}
	
	public void notifyDataChange(DataChangeEvent dce)
	{
		RuntimeException re = null;
		try{
			doChangeTransfer();
		}catch(RuntimeException e){
			re = e;
		}
		postEvent(dce);
		if (re != null) throw re;
	}
	/**
	 * Do the action as appropriate for this Control. By default this does nothing.
	 * @param how one of the ByXXX values.
	 */
	public void doAction(int how){}
//	===================================================================
	/*###############################################################
	 * 
	 * Handling incoming Events
	 * 
	 ###############################################################*/
	/*
//	==================================================================
	public void show(Rect where)
//	==================================================================
	{
			setRect(where.x,where.y,where.width,where.height);
			shown();
	}
	*/
	/**
	 * This is called when a Control is shown in a Window. Always call
	 * super.shown() if you override this.
	 */
	public void shown()
//	==================================================================
	{
		Iterator it = getChildren((ControlIterator)Cache.get(ControlIterator.class),false);
		while(it.hasNext())
			((Control)it.next()).shown();
		Cache.put(it);
	}
	/**
	* This is called to indicate that the Form is about to close and it causes a FormClosed event to be posted.
	* If you override it you should call super.formClosing(). There is no way of stopping the 
	* Form closing at this point. 
	**/
//	===================================================================
	protected void formClosing()
//	===================================================================
	{
		Iterator i = getSubControls((ControlIterator)Cache.get(ControlIterator.class));
		while(i.hasNext()){
			Object obj = i.next();
			if (obj instanceof Control)
				((Control)obj).formClosing();
		}
		Cache.put(i);
	}
	/**
	 * This will give a PenEvent to the parent Control IF
	 * this Control is marked as PenTransparent.
	 * @return true if the PenEvent was handed off to the parent
	 * false if not.
	 */
//	==================================================================
	protected boolean checkPenTransparent(PenEvent ev)
//	==================================================================
	{
		if (!hasModifier(PenTransparent,false)) return false;
		Control c = getParent();
		if (c == null) return false;
		ev.target = c;
		ev.x += this.x;
		ev.y += this.y;
		c.onEvent(ev);
		return true;
	}
	
	private static boolean ddMenuUp = false;
	/**
	 * This is called when the popup menu associated with the control generates an event.
		By default it calls popupMenuEvent(Object selectedItem) if the selectedItem is not null.
	 * @param ev The event.
	 */
//	===================================================================
	public void popupMenuEvent(MenuEvent ev)
//	===================================================================
	{
		if (ev.selectedItem != null)
			popupMenuEvent(ev.selectedItem);
	}
	/**
	* This gets closed after the popup menu for the Control has closed. You can clear the menu
	* by doing setMenu(null).
	**/
//	-------------------------------------------------------------------
	protected void popupMenuClosed(Menu m)
//	-------------------------------------------------------------------
	{
	}
	
//	===================================================================
	public static void setClipObject(Object obj)
//	===================================================================
	{
		Vm.setClipboardText(null);
		clipObject = obj;
		if (obj instanceof String []) {
			String [] all = (String [])obj;
			String out = new String();
			for (int i = 0; i<all.length; i++){
				if (all[i] != null) out += all[i];
				if (all.length != 1) out += "\n";
			}
			Vm.setClipboardText(out);
		}else if (obj instanceof String){
			Vm.setClipboardText((String)obj);
		}
	}

//	===================================================================
	public static Object getClipObject()
//	===================================================================
	{
		if (clipOwner != null) return clipObject;
		String got = Vm.getClipboardText();
		if (got == null) got = "";
		if (got.length() == 0) return clipObject;
		if (got.indexOf('\n') == -1) return got;
		return mString.split(got,'\n');
	}
	/**
	* This is called to get data from the control to put into the clipboard.
	**/
//	-------------------------------------------------------------------
	protected Object getDataToCopy()
//	-------------------------------------------------------------------
	{
		if (!(this instanceof Selectable)) return null;
		Selectable s = (Selectable)this;
		return s.getSelection();
	}
	
	/**
	* This calls takeData(data,null) and then calls dataAccepted(this,data,action)
	* on the clipOwner.
	**/
//	-------------------------------------------------------------------
	protected void takeFromClipboard(Object clip,String action)
//	-------------------------------------------------------------------
	{
		boolean took = takeData(clip,null);
		if (clipOwner != null){
			if (took) clipOwner.dataAccepted(this,clip,action);
			else clipOwner.dataTransferCancelled(clip);
			setClipObject(null);
			clipOwner = null;
		}
	}

//	===================================================================
	public Object clipboardTransfer(Object clip,boolean toClipboard,boolean cut)
//	===================================================================
	{
		Object ret = clip;
		if (!toClipboard){ //................................Paste operation.
			if (clip == null) return clip;
			takeFromClipboard(clip,"move");

			return null;
		}
		//..................................................Cut/copy operation.
		if (clipOwner != null) clipOwner.dataTransferCancelled(clip);
		//..................................................
		ret = getDataToCopy();
		if (ret == null) return null;
		int flags = getModifiers(true);
		boolean editable = (((flags & (Disabled|NotEditable|DisplayOnly)) == 0) || ((flags & NotAnEditor) != 0));
		if (cut && editable) {
			clipOwner = this;
			dataBeingRemoved(ret,null);
		}
		return ret;
	}
	/**
	* This converts an object into either a String or an array of Strings. If the Object
	* cannot be converted into either, it will return null.
	**/
//	===================================================================
	public final static Object toTextData(Object data)
//	===================================================================
	{
		if (data == null) return null;
		if (data instanceof String []) return data;
		return data.toString();
	}
	
	/**
	 * This is called when the user selects an item from the popup menu.
	 * @param selectedItem The item selected.
	 */
//	===================================================================
	public void popupMenuEvent(Object selectedItem)
//	===================================================================
	{
		if (clipItems == null) return;
		if (selectedItem == clipItems[0]) setClipObject(clipboardTransfer(getClipObject(),true,false));
		else if (selectedItem == clipItems[1]) setClipObject(clipboardTransfer(getClipObject(),true,true));
		else if (selectedItem == clipItems[2]) clipboardTransfer(getClipObject(),false,false);
		//else if (selectedItem == clipItems[3]) setClipObject(clipboardTransfer(getClipObject(),false,true));
		if (selectedItem == clipItems[1] || selectedItem == clipItems[2]) {
			repaintNow();
			notifyDataChange();
		}
	}
	/**
	 * Post an Event using the Event Thread of the Window containing this Control.
	 * The method returns immediately even if the current Thread <b>is</b> the Window Event
	 * Thread.
	 * @param ev the Event to post.
	 * @return true if the Event was posted, false if there was no Window for the Control.
	 */
	public boolean postInWindowThread(Event ev)
	{
		Window w = getWindow();
		if (w == null) return false;
		w.queueEvent(ev);
		return true;
	}
	/**
	 * Post an Event to an EventListener via the Event Thread of the Window
	 * containing this Control and via the <b>cause</b> field of a newly created DeferredEvent.
	 * <p>
	 * A new Event is created, the target is set to  
	 * The method returns immediately even if the current Thread <b>is</b> the Window Event
	 * Thread.
	 * @param ev the Event to re-post.
	 * @param listener the EventListener to receive the Event.
	 * @return true if the Event was posted, false if there was no Window for the Control.
	 */
	public boolean deferInWindowThread(Event ev,EventListener listener)
	{
		Window w = getWindow();
		if (w == null) return false;
		w.queueEvent(new DeferredEvent(ev,listener));
		return true;
	}
	
	public void onEvent(Event ev)
	{
		if (ev.target == getMenu() && (ev instanceof MenuEvent)){
			popupMenuEvent((MenuEvent)ev);
		}else if (ev instanceof PenEvent){
			Control f = this;
			/*
			if (ev.type == PenEvent.SCROLL_UP || ev.type == PenEvent.SCROLL_DOWN){
				f = Gui.focusedControl();
				if (f == null) f = this;
			}
			*/
			f.onPenEvent((PenEvent)ev);
		}else if (ev instanceof KeyEvent){
			if (ev.type == KeyEvent.KEY_PRESS && Window.dragAndDropData != null && ((KeyEvent)ev).key == IKeys.ESCAPE && !ddMenuUp){
				try{
					Control c = (Control)Window.dragAndDropSource;
					Window.resetCursor();
					c.dataTransferCancelled(Window.dragAndDropData);
					if (Window.lastDraggedOver != null)
						Window.lastDraggedOver.dataTransferCancelled(null);
					Window.dragAndDropData = null;
					c.dragging.cancelled = true;
				}catch(Exception e){}
				return;
			}
			Control f = Gui.focusedControl();
			if (f != null) {
				f.onKeyEvent((KeyEvent)ev);
		 		if (f.hasModifier(SendUpUIEvents,false)){
					for (Control c = f.parent; c != null && !ev.consumed; c = c.parent)
						c.onKeyEvent((KeyEvent)ev);
				}
			}
		}else if (ev instanceof ControlEvent){
			onControlEvent((ControlEvent)ev);
		}else if (ev instanceof DataChangeEvent){
			if (contains((Control)ev.target) && hasModifier(TakeControlEvents,false)){
				ev.target = this;
				try{
					doChangeTransfer();
				}catch(Exception e){}
				sendToListeners(ev);
			}
		}else if (ev instanceof SurfaceTextEvent){
			if (ev.type == SurfaceTextEvent.TEXT_ENTERED){
				SurfaceTextEvent te = (SurfaceTextEvent)ev;
				if ((te.flags & te.FLAG_TEXT_WAS_ENTERED) != 0){
					setText(((SurfaceTextEvent)ev).entered);
					notifyDataChange();
				}
			}
		}
		
	}
	/**
	 * Returns true if a tooltip is displayed in the current window.
	 * @return
	 */
	public boolean tipIsDisplayed()
	{
		Window w = getWindow();
		if (w == null) return false;
		return w.tipIsDisplayed();
	}
	/**
	 * This method is called by the Window to send UI events to this Control.
	 * The method will first call checkParentInterceptsUIEvent() to see if
	 * any parent intercepts the event. If that returns false then the normal
	 * postEvent() method is called.
	 * @param ev the event to be sent to this Control.
	 */
	public void postUIEvent(Event ev)
	{
		if (checkParentInterceptsUIEvent(ev,this)) return;
		postEvent(ev);
	}
	public void postEvent(Event ev)
	{
		if (ev == null) return;
		if (!(ev instanceof KeyEvent)) 
			if (!fixFocus(ev)){
				sendToListeners(ev);
			}
		if ((ev instanceof PenEvent) && !hasModifier(SendUpUIEvents,false))  onEvent(ev);
		else if ((ev instanceof KeyEvent || ev instanceof SurfaceTextEvent))  onEvent(ev);
		else {
			for (Control c = this; c != null; c = c.parent){
				if (ev.consumed) break;
				c.onEvent(ev);
			}
		}
		//super.postEvent(ev);
		
	}
	/*###############################################################
	 * 
	 * Scrolling
	 * 
	 ###############################################################*/
	protected ScrollServer ss;
	public void setServer(ScrollServer server){ss = server;}
	public ScrollServer getServer() {return ss;}
	/**
	 * This is called to update a Control on-screen <b>after</b> it has been scrolled.
	 * In this case "being scrolled" means that the internal display co-ordinates of the
	 * Control have been altered but a repaint has not been done. This method attempts to
	 * do the most effecient repaint depending on the capabilities of the underlying graphics
	 * system.<p>
	 * If data can be moved directly on screen, then this will be done and only the 
	 * newly exposed portion will be updated via a repaintNow() operation. If no moving
	 * is possible at all - or if the SIP may be covering the control then no scrolling
	 * will be done. 
	 * @param sx The source x co-ordinate.
	 * @param sy The source y co-ordinate.
	 * @param sw The source width.
	 * @param sh The source height.
	 * @param destX The destination x co-ordinate.
	 * @param destY The destination y co-ordinate.
	 */
	public void scrollAndRepaint(int sx,int sy,int sw,int sh,int destX,int destY)
//	===================================================================
	{
		
		if (false) {
			//System.out.println("Fix scroll and repaint!");
			repaintNow();
			return;
		}
		
		//
		Window w = getWindow();
		if (w == null) return;
		ISurface surface = w.getSurface();
		if (surface == null) return;
		//
		if (!surface.canMove() || ((modifiers & Transparent) != 0)|| (Gui.getSIP(null) != 0 && !(this instanceof EditControl))){
	 		repaintNow();
	 		return;
		}
		Rect srcScroll, ss = Rect.getCached(sx,sy,sw,sh); srcScroll = ss;
		Rect dstScroll, ds = Rect.getCached(0,0,width,height); dstScroll = ds;
		Point p = Point.getCached(0,0);
		try{
			if (Gui.getRectInWindow(this,dstScroll,true,p) == null) return;
			dstScroll.x -= p.x;
			dstScroll.y -= p.y;
			//
			// First make sure the source rectangle is within what is actually
			// in the control.
			//
			srcScroll.getIntersection(dstScroll,srcScroll);
			destX += srcScroll.x-sx;
			destY += srcScroll.y-sy;
			sx = srcScroll.x; sy = srcScroll.y; sw = srcScroll.width; sh = srcScroll.height;
			//
			// If nothing on screen actually moves then just repaint the whole thing.
			//
			if (sw <= 0 || sh <= 0){
				repaintNow();
				return;
			}
			//
			// Next see where it would actually end up on screen.
			//
			srcScroll.x = destX; srcScroll.y = destY;
			srcScroll.getIntersection(dstScroll,srcScroll);
			sx += srcScroll.x-destX;
			sy += srcScroll.y-destY;
			destX = srcScroll.x; destY = srcScroll.y; sw = srcScroll.width; sh = srcScroll.height;
			//
			// If nothing on screen actually moves then just repaint the whole thing.
			//
			if (sw <= 0 || sh <= 0){
				repaintNow();
				return;
			}
			//
			surface.moveImage(p.x+sx,p.y+sy,sw,sh,p.x+destX,p.y+destY);
			//BufferedGraphics bg = getGraphics(destX,destY,sw,sh); bg.release();
			//
			//if (true) return;
			Rect r = srcScroll;
			boolean didVertical = false;
			if (destX != sx){
				didVertical = true;
				r.set(destX < sx ? destX+sw : 0,0,width-sw,height);
				repaintNow(null,r);
			}
			if (sh < height){
				r.set(didVertical ? destX: 0, destY < sy ? destY+sh : 0, didVertical ? sw : width,height-sh);
				repaintNow(null,r);
			}
		}finally{
			ss.cache();
			ds.cache();
			cache(p);
		}
	}
	/**
	* This is the sound that is played when a popup-menu or pull-down menu is first
	* shown. Set this to null to disable this sound, or change it for a different sound.
	**/
	//public static String popupSound = Device.isMobile() ? "eve/Popup.wav": null;

	//public static SoundClip popupSoundClip = null;
	/**
	* This is called when a popup-menu or pull-down menu is first shown. It can
	* also be called at any time for any other purpose. If you set Sound.popupSound to
	* null, then no beep will be played at these times.
	**/
//	===================================================================
	public static void popupBeep()
//	===================================================================
	{
		if (!Device.isMobile()) return;
		Sound.beep(GuiStyle.popupBeep);
		/*
		if (popupSound != null){
			//new ewe.fx.SoundClip(popupSound);
			if (popupSoundClip == null) 
				popupSoundClip = new ewe.fx.SoundClip(popupSound);
		}
		if (popupSoundClip != null)
			popupSoundClip.play(Vm.SND_ASYNC);
			*/
	}
	
	/*###############################################################
	 * 
	 * Cursor
	 * 
	 ###############################################################*/
	public void setCursor(int cursor)
	{
		PenEvent.setCursor(this,cursor);
	}
	/*###############################################################
	 * 
	 * Hover
	 * 
	 ###############################################################*/
	/* (non-Javadoc)
	 * @see eve.fx.Drawable#getWidth()
	 */
	public int getWidth() {
		return width;
	}
	/* (non-Javadoc)
	 * @see eve.fx.Drawable#getHeight()
	 */
	public int getHeight() {
		return height;
	}
	/* (non-Javadoc)
	 * @see eve.fx.Drawable#free()
	 */
	public void free() {
		// TODO Auto-generated method stub
		
	}
	/* (non-Javadoc)
	 * @see eve.fx.Drawable#draw(eve.fx.Graphics, int, int, int)
	 */
	public void draw(Graphics g, int x, int y, int options) {
		g.translate(x,y);
		repaintNow(g, new Rect(0,0,width,height));
		g.translate(-x,-y);
	
	}
	/* (non-Javadoc)
	 * @see eve.sys.DataInterface#setFieldTransfer(eve.sys.FieldTransfer)
	 */
	public void setFieldTransfer(FieldTransfer ft) {
		fieldTransfer = ft;
	}
	/* (non-Javadoc)
	 * @see eve.sys.DataInterface#getFieldTransfer()
	 */
	public FieldTransfer getFieldTransfer() {
		return fieldTransfer;
	}
	/* (non-Javadoc)
	 * @see eve.sys.DataInterface#updateFrom(eve.sys.FieldTransfer)
	 */
	public void updateFrom(FieldTransfer ft) {
		fromField(ft);
	}
	/* (non-Javadoc)
	 * @see eve.sys.DataInterface#updateTo(eve.sys.FieldTransfer)
	 */
	public void updateTo(FieldTransfer ft) {
		toField(ft);
	}
	
//	####################################################
//	####################################################
	public void queuePenEvent(int type,int modifiers)
	{
		Window w = getWindow();
		Rect r = Gui.getRectInWindow(this,null,false);
		SurfacePointerEvent pe = new SurfacePointerEvent();
		pe.modifiers = modifiers;
		pe.x = r.x+2;
		pe.y = r.y+2;
		pe.target = w;
		pe.type = type;
		w.queueEvent(pe);
	}
	/**
	 * Call this in a non-event thread.
	 *
	 */
	public void simulateClickOn(int modifiers)
	{
		queuePenEvent(SurfacePointerEvent.PEN_DOWN,modifiers);
		mThread.nap(20);
		queuePenEvent(SurfacePointerEvent.PEN_UP,modifiers);
		mThread.nap(20);
	}

	public void queueKeyEvent(int type, int key, int modifiers)
	{
		SurfaceKeyEvent ke = new SurfaceKeyEvent();
		ke.key = key;
		ke.modifiers = modifiers;
		Window w = getWindow();
		ke.target = w;
		ke.type = type;
		w.queueEvent(ke);
	}
	public void simulateKeyType(int key, int modifiers)
	{
		queueKeyEvent(SurfaceKeyEvent.KEY_PRESS,key,modifiers);
		mThread.nap(20);
		queueKeyEvent(SurfaceKeyEvent.KEY_RELEASE,key,modifiers);
		mThread.nap(20);
	}
	public void simulateKeyPresses(String text)
	{
		for (int i = 0; i<text.length(); i++){
			char c = text.charAt(i);
			simulateKeyType(c,0);
		}
	}
	/* (non-Javadoc)
	 * @see eve.fx.Drawable#draw(eve.fx.Graphics, int, int, int, int, int)
	 */
	public void draw(Graphics g, int x, int y, int width, int height, int options) {
		//ImageTool.drawScaled(this,g,x,y,width,height,options);
		// TODO fix this.
		
	}
	/**
	 * This method is called by a Window before a UI event is sent to the destination
	 * control. If it finds a parent with the TakeChildUIEvents modifier it will call 
	 * interceptChildUIEvent on that parent. If that method returns true the event
	 * will not be sent to the child. 
	 * @param ev the event to be sent to the child control.
	 * @param destination the child control.
	 * @return true if this method found a parent with the TakeChildUIEvents modifier
	 * set <b>and</b> the interceptChildUIEvent() method returned true.
	 */
	public final boolean checkParentInterceptsUIEvent(Event ev, Control destination)
	{
		for (Control c = parent; c != null; c = c.parent)
			if ((c.modifiers & TakeChildUIEvents) != 0)
				return c.interceptChildUIEvent(ev,destination);
		return false;
	}
	/**
	 * This method is called if this Control has the TakeChildUIEvents modifier set
	 * and one of its children is the destination for a UI event. If it returns false
	 * then the event should be delivered to the child control, otherwise it should
	 * not deliver the event.
	 * <p>
	 * If you override this method you may wish to call checkParentInterceptsUIEvent()
	 * first or second. That method will check if any parents also wish to intercept the event
	 * and calls the intereceptChildUIEvent() method if it does. If that method returns
	 * true then this method should return immediately true.
	 * @param event the event that will be sent to the destination.
	 * @param destination the destination Control.
	 * @return true if the event was handled and should not be delivered, false if it
	 * should be delivered.
	 */
	public boolean interceptChildUIEvent(Event event, Control destination)
	{
		if (checkParentInterceptsUIEvent(event,destination)) return true;
		return false;
	}
}
//####################################################
