/*
 * Created on Aug 6, 2005
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import eve.data.HasProperties;
import eve.data.PropertyList;
import eve.fx.Color;
import eve.fx.Dimension;
import eve.fx.IImage;
import eve.fx.IconAndText;
import eve.fx.Insets;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.gui.IKeys;
import eve.fx.gui.WindowConstants;
import eve.sys.Cache;
import eve.sys.Event;
import eve.sys.Handle;
import eve.sys.Reflection;
import eve.sys.TimeOut;
import eve.sys.TimedOutException;
import eve.sys.Vm;
import eve.sys.Wrapper;
import eve.ui.event.ControlEvent;
import eve.ui.event.DataChangeEvent;
import eve.ui.event.FormEvent;
import eve.ui.event.ListEvent;
import eve.ui.event.MenuEvent;
import eve.ui.event.SoftKeyEvent;
import eve.ui.event.WindowEvent;
import eve.util.mVector;

/**
 * A Form is a Control that is able to create and display itself in its own Window
 * or Frame via a show() or execute() method call.<p>
 * You would hardly use this control - usually you would use a higher version of
 * this, such as a Form.
 * 
 * @author Michael L Brereton
 *
 */
//####################################################
public class Form extends FormBase implements HasProperties{

	public boolean acceptsDroppedFiles;
	/**
	* The properties of the Form.
	*/
	protected PropertyList pl;
//	===================================================================
	public PropertyList getProperties()
//	===================================================================
	{
		if (pl == null) pl = new PropertyList();
		return pl;
	}
	/**
	* A special method used by FieldTransfer to transfer data to and from controls and properties.
	**/
//	===================================================================
	public boolean _getSetField(String fieldName,Wrapper wrapper,boolean isGet)
//	===================================================================
	{
		return PropertyList.getSetProperties(this,fieldName,wrapper,isGet);
	}

	/**
	 * Changes all necessary flags so that any new Window/Frame opened by
	 * this Form has no top title bar. 
	 */
	public void setNoTitleBar()
	{
		hasTitle = hasTopBar = false;
		windowFlagsToClear |= WindowConstants.FLAG_HAS_CLOSE_BUTTON|WindowConstants.FLAG_HAS_TITLE;
	}
	/**
	 * Set the Form to show maximized on a PDA/mobile device.
	 */
	public void maximizeOnPDA()
	{
		windowFlagsToSet |= WindowConstants.FLAG_MAXIMIZE_ON_PDA;
		if (Gui.isPDA || (Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_MAXIMIZE_APPS) != 0){
			Dimension d = Gui.getScreenSize();
			setPreferredSize(d.width,d.height);
		}
	}
	/**
	 * Prepare a Form for display at full screen on a PDA. It calls
	 * maximizeOnPDA() and then sets the FLAG_FULL_SCREEN flags and removes
	 * the FLAG_HAS_TITLE and FLAG_HAS_CLOSE_BUTTON flags. It also sets
	 * hasTopBar and hasTitle to false and sets exitButtonDefined and 
	 * hideSoftKeyBarIfNotFound to true. This means that you MUST provide
	 * some way for the user to exit the Form yourself since no extra controls
	 * for this will be provided by the Form itself.
	 */
	public void fullScreenOnPDA()
	{
		maximizeOnPDA();
		windowFlagsToSet |= WindowConstants.FLAG_FULL_SCREEN;
		windowFlagsToClear |= WindowConstants.FLAG_HAS_TITLE|WindowConstants.FLAG_HAS_CLOSE_BUTTON;
		hasTopBar = hasTitle = false;
		exitButtonDefined = true;
		hideSoftKeyBarIfNotFound = true;
	}
	/**
	 * Normally an Editor will call maximizeOnPDA() when created.
	 * This method will reverse that change. Call it immediately on construction.
	 */
	public void makeSmall()
	{
		windowFlagsToSet &= ~WindowConstants.FLAG_MAXIMIZE_ON_PDA;
		if (Gui.isPDA  || (Vm.getParameter(Vm.VM_FLAGS) & Vm.VM_FLAG_MAXIMIZE_APPS) != 0){
			tags.clear(TAG_PREFERREDSIZE);
		}
	}

/**
 * This creates a new Window for the Form on request (when being shown). You can use this to
	setup the window icon.
 * @return a new Window for the Form.
 */
public Window createWindow()
{
	Window w = new Window();
	w.modify(modifiers&ModifiersInheritedFromWindow,0);
	return w;
}
/**
 * Set this true if the Form should display the Soft keys of its child form.
 */
public boolean inheritSoftKeys = false;
public boolean resizable = !Gui.isPDA;//true;
public boolean moveable = !Gui.isPDA;
public boolean hasTitle = true;
public boolean resizeOnSIP = false;
public boolean keepFrame = true;
public boolean noBorder = Gui.hasSoftKeys;
public boolean hasTopBar = !Gui.hasSoftKeys;
public boolean exitSystemOnClose = false;
/**
 * On systems that use the SoftKeyBar - if a control becomes active but there
 * is no assigned keybar for the control (and there is no default one for the
 * Form) then normally whatever keybar was there before remains unchanged. If
 * this is set true then if a keybar is not found then SoftKeyBar.hide() is
 * called instead. Note that if SoftKeyBar.hideRemovesSoftKeyBar is false 
 * (the default) then the SoftKeyBar is not removed - it is just made empty.
 * If you set SoftKeyBar.hideRemovesSoftKeyBar true - then the SoftKeyBar
 * will be removed entirely.
 */
public boolean hideSoftKeyBarIfNotFound = false;
/**
 * Set this true to supress the <b>automatic</b> addition of a SIP filler to the Form.
 */
public boolean dontAddSIPFiller = false;
/**
 * This can be used to monitor the shown state of the Form. If it indicates
 * Running then it is still shown and has not been closed yet.
 */
public Handle handle = new Handle();
public Control
yes = null, no = null, cancel = null, ok = null,
apply = null, reset = null, deflt = null, back = null;

/**
 * Set this to be an ImageData or (better) a DeviceIcon (created by Device.createIcon())
 */
public Object windowIcon;
/**
* Set this to be greater than zero to limit the number of buttons displayed at the bottom of the Form.
**/
public int buttonsPerRow = 0;
/**
* If this is set to a control then this control will appear to the right of the
* title bar. If it is activated it will cause the form to exit
* with the value IDOK.
*/
public Control titleOK = null;
/**
* If this is set to a control then this control will appear to the right of the
* title bar. If it is activated it will cause the form to exit
* with the value IDCANCEL.
*/
public Control titleCancel = null;
/**
* The buttons to be displayed at the bottom of the Form (if any) are stored here.
**/
protected Vector buttons = null;
/**
* Use this to add controls to the top of the form, below the
* title bar. It is initially null, but you can create a CellPanel, 
* add controls to it and set topControls to point to it.
* This panel will stretch horizontally across the top of the
* form but it will not stretch vertically. You can use it to
* add a menu bar, for example, to your form.
*/
public CellPanel topControls;
/**
* Use this to add controls to the title bar of the form.
* It is initially null, but you can create a CellPanel, add controls
* to it and then set titleControls to point to it. These controls
* will appear to the right of the title, but to the left of any
* titleOK or titleCancel buttons.
*/
public CellPanel titleControls;
/**
* This is the panel that will contain any of the special buttons you add.
*/
public CellPanel buttonsPanel;

public static final String WINDOW_TITLE_DONT_CHANGE = "!";
/**
If this is not null, then the window that the Form is displayed in will have its title
set to this value regardless of whether it is the top level form or not. If it is null
then the window title will be set to the form title. If it is WINDOW_TITLE_DONT_CHANGE
then the window title will not be affected.
**/
public String windowTitle = null;
/**
* This is set true when doButtons() is called to add close buttons to the
* form. If it is false when the form is displayed, then the Form will
* automatically add an OK button to the title of the form.
*/
public boolean exitButtonDefined = false;
/**
* These are flags that you want to set for any new Window being created to display this Form.
It should consist of WindowConstants.FLAG_xxx values OR'ed together.
**/
public int windowFlagsToSet = 0;
/**
* These are flags that you want to clear for any new Window being created to display this Form.
It should consist of WindowConstants.FLAG_xxx values OR'ed together.
**/
public int windowFlagsToClear = 0;
/**
* Set this to be a valid Window.TaskBarIconInfo if you want to display an icon in the taskbar
* when this Form is displayed. It will replace the application icon with this and then restore it
* when it closes. If there is no application icon, it will add a new icon.
**/

/**
 * Tests if the Form has an exit button defined for it.
 * @return true if the Form has an exit button defined for it.
 */
//==================================================================
public boolean hasExitButton()
//==================================================================
{
	return (titleOK != null || titleCancel != null || exitButtonDefined);
}
//===================================================================
public MenuItem makeMenuItemForForm(String label, String action, String iconName)
//===================================================================
{
	return makeMenuItemForForm(label,action,loadImage(iconName));
}
//===================================================================
public MenuItem makeMenuItemForForm(String label, String action, IImage icon)
//===================================================================
{
	if (action == null) action = label;
	MenuItem mi = new MenuItem(label);
	mi.action = action;
	mi.label = label;
	if (icon != null) {
		mi.image = new IconAndText(icon,label,getFontMetrics());
		((IconAndText)mi.image).textColor = null;
	}
	return mi;
}

//===================================================================
public Control makeButtonForForm(String text, IImage icon, int hotkey)
//===================================================================
{
 	Control c = getButton(text).setHotKey(0,hotkey);
	c.image = icon;
	if (c instanceof Button) ((Button)c).textPosition = RIGHT;
	if (hasModifier(MouseSensitive,false)) c.modify(MouseSensitive,0);
	return c;	
}

//===================================================================
public Control makeButtonForForm(String text, String iconName, int hotkey)
//===================================================================
{
	return makeButtonForForm(text,loadImage(iconName),hotkey);
}

/**
Create and return the default OK or Cancel button.
* @param whichButton one of OKB, DEFOKB, CANCELB, or DEFCANCELB.
* @param textToUse the text to use with the button, or null to use the icon only.
* @return the button created. You should assign this button to either the "ok" or "cancel"
field.
*/
//===================================================================
public Control makeDefaultButton(int whichButton, String textToUse)
//===================================================================
{
	if ((whichButton & (OKB|DEFOKB)) != 0)
		return makeButtonForForm(textToUse,tick,IKeys.ENTER);
	else if ((whichButton & (CANCELB|DEFCANCELB)) != 0)
		return makeButtonForForm(textToUse,cross,IKeys.ESCAPE);
	else
		return null;
}
/**
Create and return the default OK or Cancel button.
* @param whichButton one of OKB, DEFOKB, CANCELB, or DEFCANCELB.
* @param useText true to use the standard OK/Cancel lables, false to use the icon only.
* @return the button created. You should assign this button to either the "ok" or "cancel"
field.
**/
//===================================================================
public Control makeDefaultButton(int whichButton, boolean useText)
//===================================================================
{
	if ((whichButton & (OKB|DEFOKB)) != 0)
		return makeButtonForForm(useText ? "OK" : null,tick,IKeys.ENTER);
	else if ((whichButton & (CANCELB|DEFCANCELB)) != 0)
		return makeButtonForForm(useText ? "Cancel" : null,cross,IKeys.ESCAPE);
	else
		return null;
}
/**
* This modifies the setup for the Form to be one suitable to run full screen on a SmartPhone 
* compatible IF the application actually is running on a SmartPhone device. 
* <p>
* Setup the title and other Form options before calling this method.
*
* @return true if the application IS running on a SmartPhone device.
*/
//===================================================================
public boolean modifyForSmartPhone()
//===================================================================
{
	if (!Gui.hasSoftKeys) return false;
	windowTitle = title;
	moveable = resizable = hasTopBar = false;
	windowFlagsToClear |= Window.FLAG_HAS_CLOSE_BUTTON;
	return true;
}

//===================================================================
public boolean placeCancelOnLeft()
//===================================================================
{
	return reverse();
}
//-------------------------------------------------------------------
private Control addButton(int what,int mask,String text,IImage icon,int hot)
//-------------------------------------------------------------------
{
	int whichButton = (what & mask);
	if (whichButton == 0) return null;
	Control c = makeButtonForForm(text,icon,hot);
	if (c == null) return null;
	exitButtonDefined = true;
	if (buttons == null) buttons = new Vector();
	buttons.add(c);
	return c;
}

private static boolean reverse()
{
	return (Gui.getGuiFlags() & Window.GUI_FLAG_REVERSE_OK_CANCEL) != 0;
}
/**
* This is used to add a row of buttons at the bottom of the form, using
* The value of "which" can be the logical OR of any of the pre-defined
* button values (e.g. YESB|NOB|OKB).
*/
//==================================================================
public void doButtons(int which)
//==================================================================
{
	boolean first = !reverse();
	if ((which & (YESB|NOB)) == (YESB|NOB))
		if ((which & (CANCELB|DEFCANCELB)) != 0)
			if (!first){
				cancel = addButton(which,DEFCANCELB|CANCELB,cancelText,cross,IKeys.ESCAPE);
				no = addButton(which,NOB,"No",stop,'n');
				yes = addButton(which,YESB,"Yes",tick,IKeys.ENTER);
			}
	for (int i = 0; i<2; i++){
		if (first){
			if (ok == null) ok = addButton(which,OKB,"OK",tick,'o');
			if (ok == null) ok = addButton(which,DEFOKB,"OK",tick,IKeys.ENTER);
			if (yes == null) yes = addButton(which,YESB,"Yes",tick,reverse() ? IKeys.ENTER : 'y');
		}else{
			if (no == null) no = addButton(which,NOB,"No",stop,reverse() ? IKeys.ESCAPE : 'n');
			if (cancel == null) cancel = addButton(which,CANCELB,cancelText,cross,'c');
			if (cancel == null) cancel = addButton(which,DEFCANCELB,cancelText,cross,IKeys.ESCAPE);
		}
		first = !first;
	}
	if ((which & CANCELB) == 0)
	if (back == null) back = addButton(which,BACKB,"<< Back",null,'b');
}
static final String cancelText = "Cancel";
/**
* This is used to add custom buttons to the bottom of the bar.
*/
//==================================================================
public void addButton(Control b)
//==================================================================
{
	if (b == null) return;
	if (buttons == null) buttons = new Vector();
	if (hasModifier(MouseSensitive,false)) b.modify(MouseSensitive,0);
	buttons.add(b);
}
/**
* Used to create a new Button for the Form. By default it simply creates an Button().
**/
//==================================================================
public Control getButton(String text)
//==================================================================
{
	return new Button(text);
}

public static int BUTTONS_TO_SOFT_KEY_FIRST_BUTTON_SEPARATE = 0x1;
public static int BUTTONS_TO_SOFT_KEY_MENU_ALWAYS = 0x2;
public static int BUTTONS_TO_SOFT_KEY_USE_FIRST_SOFT_KEY = 0x4;

//===================================================================
public static Vector gatherButtons(Container buttons,Vector destination)
//===================================================================
{
	if (destination == null) destination = new Vector();
	Iterator it = buttons.getCachedSubControls();
	for (; it.hasNext();){
		Object c = it.next();
		if (c instanceof Button) destination.add(c);
		else if (c instanceof Container) gatherButtons((Container)c, destination);
	}
	Cache.put(it);
	return destination;	
}
/**
Add a container which contains Button objects to the SoftKeyBar.
**/
//===================================================================
public boolean buttonsToSoftKeyBar(Container buttons, String menuName, int options)
//===================================================================
{
	if (buttons == null) return false;
	return buttonsToSoftKeyBar(gatherButtons(buttons,null),menuName,options);
}
/**
Add a Vector of  Button objects to a Menu.
**/
//===================================================================
public Menu buttonsToMenu(Vector buttons, Menu destination)
//===================================================================
{
	if (destination == null) destination = new Menu();
	Menu m = destination;
	if (buttons == null || buttons.size() == 0) return m;
	for (int i = 0; i<buttons.size(); i++){
		Button mb = (Button)buttons.get(i);
		m.addItem(createMenuItem(mb));
	}
	return m;
}

/**
Add a Vector of  Button objects to the SoftKeyBar.
**/
//===================================================================
public boolean buttonsToSoftKeyBar(Vector buttons, String menuName, int options)
//===================================================================
{
	boolean firstButtonSeparate = (options & BUTTONS_TO_SOFT_KEY_FIRST_BUTTON_SEPARATE) != 0;
	boolean menuAlways = (options & BUTTONS_TO_SOFT_KEY_MENU_ALWAYS) != 0;
	int dest = (options & BUTTONS_TO_SOFT_KEY_USE_FIRST_SOFT_KEY) != 0 ? 1 : 2;
	//
	if (!SoftKeyBar.usingSoftKeys()) return false;
	if (buttons == null || buttons.size() == 0) return false;
	//
	SoftKeyBar sk = makeSoftKeys();
	//
	if (firstButtonSeparate){
		sk.setKey(1,(Button)buttons.get(0));
		((Button)buttons.get(0)).addListener(this);
		if (buttons.size() == 1) return true;
	}
	//
	Menu m = null;
	for (int i = firstButtonSeparate ? 1 : 0; i<buttons.size(); i++){
		Button mb = (Button)buttons.get(i);
		mb.addListener(this);
		if (i == buttons.size()-1 && m == null && !menuAlways){
			sk.setKey(dest,mb);
			return true;
		}else{
			if (m == null) m = new Menu();
			m.addItem(sk.createMenuItem(mb));
		}
	}
	sk.setKey(dest,menuName,m);
	return true;
}

/**
 * This is called before the Form is displayed to setup the buttons as requested by doButtons().
	If no exit button has been defined for the Form, then titleOK is set to a new button.
 */
//-------------------------------------------------------------------
protected void checkButtons()
//-------------------------------------------------------------------
{
	if (buttons != null) {
		if (buttons.size() != 0){
			if (!inheritSoftKeys && Gui.hasSoftKeys && getSoftKeyBarFor(null) == null){
				buttonsToSoftKeyBar(buttons,
					(no != null && cancel != null) ? "No/Cancel" : "Actions"
					,BUTTONS_TO_SOFT_KEY_FIRST_BUTTON_SEPARATE);
			}else{
				CellPanel p = new CellPanel(); 
				p.defaultTags.set(TAG_INSETS,new Insets(0,1,0,1));
				p.modify(AlwaysEnabled|NotAnEditor,0); // Just in case a dialog pops up with global disabling.
				for (int i = 0; i<buttons.size(); i++){
					p.addNext((Control)buttons.get(i));
					if ((buttonsPerRow > 0) && (((i+1)%buttonsPerRow) == 0))
						p.endRow();
				}
				p.endRow();
				CellPanel p2 = buttonsPanel = new CellPanel(); 
				p.defaultTags.set(TAG_INSETS,new Insets(2,2,2,2));
				p2.addLast(p).setControl(CENTER);// p2.borderStyle = Graphics.EDGE_SUNKEN;
			}
		}
	}
	if (!hasExitButton()){
		if (!inheritSoftKeys && Gui.hasSoftKeys){
			if (getSoftKeyBarFor(null) == null){
				SoftKeyBar sk = makeSoftKeys();
				sk.setKey(1,"Close|"+EXIT_IDCANCEL,close,null);
			}
		}else{
			Button b = new Button(close);//getButton("OK");
			b.setBorder(EDGE_ETCHED,2);
			b.backGround = Color.DarkBlue;
			b.insideColor = getBackground();
			titleCancel = b;
		}		
	}
	if (titleOK != null) titleOK.modify(AlwaysEnabled|NotAnEditor,0);
	if (titleCancel != null) titleCancel.modify(AlwaysEnabled|NotAnEditor,0);
}
//==================================================================
public void make(boolean reMake)
//==================================================================
{
	if (made) return;
	if (sipFiller == null && !dontAddSIPFiller) {
		createSipFiller(this);
	}
	if (buttonsPanel != null) {
		addLast(buttonsPanel).setCell(HSTRETCH);
	}
	super.make(reMake);
	if (hasModifier(KeepSIP,false))
		modifyAll(KeepSIP,0);
	//toControls();
}
/**
 * Use this to indicate to the Form the ScrollablePanel that is considered
 * the "MainScrollBarPanel" for the Form. This is the one that will be enabled
 * with enableScrolling(). This method does NOT add it to the Form itself,
 * you have to add it to the Form yourself and place the controls within it
 * as normal. This method just places it in the Form's PropertyList and tags
 * it as the MainScrollBarPanel.
 * @param mainScroller the MainScrollBarPanel for the Form.
 */
public void setMainScrollBarPanel(ScrollablePanel mainScroller)
{
	getProperties().set("MainScrollBarPanel",mainScroller);
}
/**
If a ScrollablePanel has been setup in the Editor, and the property "MainScrollBarPanel" has
been added to the PropertyList of this Editor, then this will either enable or disable the
scrolling of the scrollbar in different directions.
@param horizontally true to enable the scrolling horizontally, false to disable it.
@param vertically true to enable the scrolling vertically, false to disable it.
@return true if a "MainScrollBarPanel" was defined, false if it was not.
**/
public boolean enableScrolling(boolean horizontally,boolean vertically)
{
	Object sbp = PropertyList.getValue(this,"MainScrollBarPanel",null);
	if (sbp instanceof ScrollablePanel){
		int ex = HEXPAND|VEXPAND;
		if (!horizontally) ex |= HCONTRACT;
		if (!vertically) ex |= VCONTRACT;
		((ScrollablePanel)sbp).setClientConstraints(ex);
		return true;
	}else
		return false;
}

static class waitState
{
	Window shownIn;
	boolean isShowing;
	boolean wasWaiting;
}

static waitState waiting = new waitState();
//static boolean wasWaiting = false;
/**
 * This displays a wait cursor until the next Form is shown.
 */
//===================================================================
public static synchronized void showWait()
//===================================================================
{
	showWait(null);
}
/**
 * This displays a wait cursor until the next Form is shown.
 */
//===================================================================
public static synchronized void showWait(Window win)
//===================================================================
{
	if (waiting.isShowing) return;
	if (win == null) win = Window.getActiveWindow();
	waiting.isShowing = true;
	waiting.shownIn = win;
	Gui.showWait(win,true);
}
/**
* This requests that a wait Cursor be shown IF it was shown before the last
* Form was displayed - otherwise it has no effect. continueWait() is called
* automatically when a MessageBox exits.
**/
//===================================================================
public static void continueWait()
//===================================================================
{
	if (waiting.wasWaiting) showWait(waiting.shownIn);
	waiting.wasWaiting = false;
}
/**
* This cancels the display of the wait Cursor IF it was requested.
**/
//===================================================================
public static void cancelWait()
//===================================================================
{
	if (waiting.isShowing) Gui.showWait(waiting.shownIn,false);
	waiting.wasWaiting = waiting.isShowing = false;
}

/**
 * Use this to explicitly set the Control that will receive the focus
 * first when the window for this Form is shown.
 */
public Control firstFocus;

public String title = "untitled";

public boolean dismantleOnClose = false;
/**
* The Frame the Form is currently displayed in.
**/
protected FormFrame formFrame;

public int exitValue = 0;
/**
* This is called when the Form is about to be displayed. It creates and
* returns a FormFrame with the contents set to be the Form.
*/
//==================================================================
public FormFrame getFormFrame(int options)
//==================================================================
{
	if (keepFrame && formFrame != null) return formFrame;
	return formFrame = makeFrame(options);
}
/**
* This is used to actually create the frame for the form. The 
* implementation of this is simply: {return new FormFrame(this,options);}
* but it can be overridden to return a different FormFrame.
*/
//-------------------------------------------------------------------
protected FormFrame makeFrame(int options) 
//-------------------------------------------------------------------
{
	return new FormFrame(this,options);
}

public float frameAlpha = 1.0f;

//===================================================================
public synchronized void dismantle(Control stopAt)
//===================================================================
{
	if (stopAt == this) return;
	if (!keepFrame && formFrame != null)
		formFrame = null;
	super.dismantle(stopAt);
}

private boolean pushedSoftkey;

protected void setupFrame(Frame f,int options)
{
	
}
/**
 * Use this to update the Window title for the Form (if you change the Form's
 * title or Form's windowTitle).
 *
 */
public void fixWindowTitle()
{
	if (title == null || title.length() == 0) return;
	Window w = getWindow();
	if (w == null) return;
	w.name = title;
	if (windowTitle == null){
		w.name = title;
		w.setTitle(title);
	}else if (windowTitle != WINDOW_TITLE_DONT_CHANGE){
		w.name = windowTitle;
		w.setTitle(windowTitle);
	}
}
/**
 * All show()/exec()/execute() methods eventually call this method.
	It is used to create and display the Frame/Window for the Form. It also creates the Handle
	for the Form and sets it to Handle.Running.
 * @param parent The parent Frame (may be null).
 * @param modal true if the Form is to be displayed modally.
 * @param options one of the Gui.XXX_FRAME options along with the Gui.NEW_WINDOW option.
 * @return The FormFrame created and displayed.
 */
//-------------------------------------------------------------------
protected synchronized FormFrame doShowExec(Frame parent,boolean modal,int options)
//-------------------------------------------------------------------
{
	options = Gui.fixShowOptions(parent,options);
	FormFrame f = getFormFrame(options);
	f.resizeOnSIP = resizeOnSIP;
	f.backGround = getBackground();
	setupFrame(f,options);
	f.addListener(this);
	checkButtons();
	pushedSoftkey = false;
	if (softkeyBars != null){
		SoftKeyBar got = (SoftKeyBar)softkeyBars.get("default");
		SoftKeyBar.push(got);
		pushedSoftkey = true;
	}else{
	}
	if (modal) Gui.execFrame(f,parent,options);	
	else Gui.showFrame(f,parent,options);
	//if (listener != null) addListener(listener);
	//fixWindowTitle();
	/*
	if (taskbarIcon != null) {
		if (mApp.taskbarWindow == null)
			mApp.taskbarWindow = new TaskbarWindow(mApp.mainApp.windowTitle);
		String nm = getClass().getName();
		getProperties().set("oldIcon",mApp.taskbarWindow.getIcon());
		mApp.taskbarWindow.addIcon(nm,taskbarIcon.nativeIcon);
		mApp.taskbarWindow.setIconAndTip(nm,taskbarIcon.tip);
	}
	*/
	//
	waiting.wasWaiting = waiting.isShowing;
	if (waiting.isShowing) {
		Gui.showWait(waiting.shownIn,false);
		waiting.isShowing = false;
	}
	//
	if (handle == null) handle = new Handle();
	handle.setFlags(Handle.Running,Handle.Stopped);
	return f;
}
/**
* This closes the form unconditionally and sets the exitValue to be exitCode.
*/
//==================================================================
public void close(int exitCode)
//==================================================================
{
	exitValue = exitCode;
	Control p = null;
	Frame f = getFrame();
	if (f != null) Gui.hideFrame(f);
	try{
	if (pushedSoftkey) SoftKeyBar.pop();
	formClosing();
	if (exitSystemOnClose){
		Application.exit(exitCode);
	}
	if (listeners != null) listeners.clear();
	for (Control c = this; c != null; c = c.getParent())
		if (c instanceof Form){
			stopFormHandle(((Form)c).handle,exitCode);
		}
	if (this instanceof MessageBox) continueWait();
	if (dismantleOnClose) dismantle();
	}catch(Exception e){
		e.printStackTrace();
	}
}
/**
* Call this ONLY after a show() call. It waits until the Form has been shown.
* If the form is being placed in a new Window this will occur when the Window manager
* sends a PAINT message to the Window. If the Form is within an already displayed Frame
* then this will return immediately.
**/
//===================================================================
public boolean waitUntilShown(int timeOut)
//===================================================================
{
	try{
		boolean got = handle.waitOn(SHOWN_FLAG,timeOut);
		return got;
	}catch(Throwable t){
		return false;
	}
	/*
	Frame f = getFrame();
	if (f == null) return false;
	if (!Gui.isWindowFrame(f)) return true;
	Window w = f.getWindow();
	if (w == null) return true;
	return w.waitUntilPainted(timeOut);
	*/
}

public static final int SHOWN_FLAG = 0x100;

public void setAsShown()
{
	if (handle != null) handle.setFlags(SHOWN_FLAG,0); 
}
public void shown()
{
	setAsShown();
	super.shown();
}
/**
* This is called to indicate that the Form is about to close and it causes a FormEvent.CLOSED event to be posted.
* If you override it you should call super.formClosing(). There is no way of stopping the 
* Form closing at this point. 
**/
//-------------------------------------------------------------------
protected void formClosing()
//-------------------------------------------------------------------
{
	/*
	if (taskbarIcon != null && mApp.taskbarWindow != null)
		mApp.taskbarWindow.restoreIcon(getProperties().getValue("oldIcon",null));
		*/
	postEvent(new FormEvent(FormEvent.CLOSED,this));
	super.formClosing();
}
/**
* This is called by the exit() method to see if the form can be
* closed. It should return true if it can be closed and false
* if it cannot.
*/
//-------------------------------------------------------------------
protected boolean canExit(int exitCode)
//-------------------------------------------------------------------
{
	return true;
}
/**
* This will call the canExit() method and if it returns true, it will
* close the form.
*/
//==================================================================
public boolean exit(final int exitCode)
//==================================================================
{
	if (!canExit(exitCode)) return false;
	Control c = Gui.focusedControl();
	boolean wasMe = (c != null && c.isChildOf(this));
	if (false){
		close(exitCode);
	}else{
		Window w = getWindow();
		try{
			new WindowThreadTask(){
				public void doRun(){
					//if (h.shouldStop) exit(IDCANCEL);
					//else exit(IDOK);
					close(exitCode);
				}
			}.execute(this);
		}catch(IllegalStateException e){
			close(exitCode);
		}
	}
	//close(exitCode);
	//if (wasMe) Gui.releaseFocus(null,ByRequest);
	return true;
}

/**
* This displays the Form modally within an already displayed parent Frame and then waits for
* the Form to close.
* @param parent The parent Frame for the Form. A null parent will indicate that a new Window
* should be created.
* @param options this should be one of: 
* <br>Gui.CENTER_FRAME - Size the Frame to the preferred size of its contents and then
* display it centered in the parent frame (this is the most common option).
* <br>Gui.FILL_FRAME - Fill the entire parent frame. 
* <br>Gui.FILL_WIDTH - Fill the entire width of the parent frame. 
* <br>Gui.FILL_HEIGHT - Fill the entire height of the parent frame. 
* <br><b>If options is 0 then the Frame will not be sized, positioned or painted</b> although it will
* be added as a child of the parent Frame. In this case you must size and position the returned
* FormFrame yourself and then call repaintNow().
* @return The return value of the Form as sent to exit() or close().
*/
//===================================================================
public final int execute(Frame parent,int options)
//===================================================================
{
	exec(parent,options);
	return waitUntilClosed();
}
//===================================================================
public final int execute(Frame parent)
//===================================================================
{
	exec(parent,Gui.CENTER_FRAME);
	return waitUntilClosed();
}
/**
* This displays the Form modally within an already displayed parent Frame.
* @param parent The parent Frame for the Form. A null parent will indicate that a new Window
* should be created.
* @param options this should be one of: 
* <br>Gui.CENTER_FRAME - Size the Frame to the preferred size of its contents and then
* display it centered in the parent frame (this is the most common option).
* <br>Gui.FILL_FRAME - Fill the entire parent frame. 
* <br>Gui.FILL_WIDTH - Fill the entire width of the parent frame. 
* <br>Gui.FILL_HEIGHT - Fill the entire height of the parent frame. 
* <br><b>If options is 0 then the Frame will not be sized, positioned or painted</b> although it will
* be added as a child of the parent Frame. In this case you must size and position the returned
* FormFrame yourself and then call repaintNow().
* @return The FormFrame that was created and displayed.
*/
//===================================================================
public final FormFrame exec(Frame parent,int options) 
//===================================================================
{
	if (parent == null) return doShowExec(parent,true,Gui.NEW_WINDOW);//defaultShowOptions);
	else return doShowExec(parent,true,options);
}
/**
 * This is used to exec a Form using the Handle from another Form (usually got
 * from closeForTransfer()).
 * @param transferredHandle The Handle to use for this Form.
 * @return The FormFrame created.
 */
//===================================================================
public FormFrame exec(Handle transferredHandle)
//===================================================================
{
	FormFrame ff = exec();
	handle = transferredHandle;
	return ff;
}
/**
 * This is used to exec a Form using the Handle from another Form (usually got
 * from closeForTransfer()).
 * @param transferredHandle The Handle to use for this Form.
  @param parent The parent Frame to exec() this Form in.
* @param options this should be one of: 
* <br>Gui.CENTER_FRAME - Size the Frame to the preferred size of its contents and then
* display it centered in the parent frame (this is the most common option).
* <br>Gui.FILL_FRAME - Fill the entire parent frame. 
* <br>Gui.FILL_WIDTH - Fill the entire width of the parent frame. 
* <br>Gui.FILL_HEIGHT - Fill the entire height of the parent frame. 
* <br><b>If options is 0 then the Frame will not be sized, positioned or painted</b> although it will
* be added as a child of the parent Frame. In this case you must size and position the returned
* FormFrame yourself and then call repaintNow().
* @return The FormFrame created.
*/
//===================================================================
public FormFrame exec(Handle transferredHandle,Frame parent,int options)
//===================================================================
{
	FormFrame ff = exec(parent,options);
	handle = transferredHandle;
	return ff;
}

/**
* This displays the Form modally in a new Window and waits for the Form to close. If the platform does not support
* multiple windows it will be displayed centered within the main application window.
* @return The return value of the Form as sent to exit() or close().
*/
//===================================================================
public final int execute()
//===================================================================
{
	exec();
	return waitUntilClosed();
}
/**
* This will display the form either modally or not.
**/
/*
//===================================================================
public final FormFrame exec(boolean modally)
//===================================================================
{
	if (modally) return exec();
	else return show();
}
*/
/**
* This displays the Form modally in a new Window. If the platform does not support
* multiple windows it will be displayed centered within the main application window.
* @return The FormFrame that was created and displayed.
*/
//===================================================================
public final FormFrame exec()
//===================================================================
{
	return doShowExec(null,true,Gui.NEW_WINDOW);//defaultShowOptions);
}
/**
* This waits indefinitely for the form to close and returns the exitValue.
* @return the exitValue of the Form.
* @exception IllegalStateException if the Form has no handle (i.e. has not been opened).
*/
//===================================================================
public final int waitUntilClosed() throws IllegalStateException
//===================================================================
{
	Handle handle = this.handle;
	if (handle == null) throw new IllegalStateException();
	while(true){
		try{
			handle.waitUntilStopped();
			break;
		}catch(InterruptedException e){}
	}
	if (!(handle.returnValue instanceof Integer)) return exitValue;
	return ((Integer)handle.returnValue).intValue();
}

/**
 * Use this to wait on the Handle of a Form, without keeping a reference to the Form.
 * This allows the Handle to be transferred to another Form while letting this Form
 * be destroyed/garbage collected.
 * @param h The handle of the Form (from the "handle" variable).
 * @param t A TimeOut to wait form.
 * @return The final exited value of the Form as placed in the Handle.
 * @exception InterruptedException If the Thread was interrupted.
 * @exception TimedOutException If the TimeOut expired.
 */
//===================================================================
public static final int waitUntilClosed(Handle h,TimeOut t) throws InterruptedException, TimedOutException
//===================================================================
{
	if (!h.waitUntilStopped(t))
		throw new TimedOutException();
	if (!(h.returnValue instanceof Integer)) return 0;
	return ((Integer)h.returnValue).intValue();
}
/**
 * Use this to wait on the Handle of a Form, without keeping a reference to the Form.
 * This allows the Handle to be transferred to another Form while letting this Form
 * be destroyed/garbage collected.
 * @param h The handle of the Form (from the "handle" variable).
 * @return The final exited value of the Form as placed in the Handle.
 */
//===================================================================
public static final int waitUntilClosed(Handle h)
//===================================================================
{
	while(true){
		try{
			h.waitUntilStopped();
			break;
		}catch(InterruptedException e){
			
		}
	}
	if (!(h.returnValue instanceof Integer)) return 0;
	return ((Integer)h.returnValue).intValue();
}
/**
* This waits and returns true if it closed before the timeout. 
* @exception IllegalStateException if the Form has no handle (i.e. has not been opened).
**/
/*
//===================================================================
public final boolean waitUntilClosed(TimeOut howLong) throws IllegalStateException, Interrupte
//===================================================================
{
	if (handle == null) throw new IllegalStateException();
	return handle.waitOnFlags(Handle.Stopped,howLong);
}
*/
/**
 * This closes the Form but does not set the handle to a stopped state, instead it returns
 * the still running handle. This allows the handle to be transferred to another Form using
 * setHandle(), so that any Threads waiting for this Form to close will then wait for the
 * other Form to close instead.<p>
 * Transferring the handle to another Form is done using the exec(Handle) method or
 * exec(Handle,Parent,int) method.
* @param dismantle dismantle this Form after closing.
* @param showWait call the showWait() method to display a wait cursor until the next Form is
* displayed.
* @return The running handle for this open Form.
*/
//===================================================================
public Handle closeForTransfer(boolean dismantle,boolean showWait)
//===================================================================
{
	Handle h = handle;
	handle = null;
	close(0);
	if (dismantle) this.dismantle();
	if (showWait) this.showWait();
	return h;
}
/**
* This closes the Form but does not set the handle to a stopped state, instead it returns
 * the still running handle. This allows the handle to be transferred to another Form using
 * setHandle(), so that any Threads waiting for this Form to close will then wait for the
 * other Form to close instead.<p>
 * Transferring the handle to another Form is done using the exec(Handle) method or
 * exec(Handle,Parent,int) method.<p>
This Form will be dismantled after closing and the wait cursor will be displayed until the
new Form is displayed.
* @return The running handle for this open Form.
 */
//===================================================================
public Handle closeForTransfer()
//===================================================================
{
	return closeForTransfer(true,true);
}
/**
* This displays the Form non-modally in a new Window. If the platform does not support
* multiple windows it will be displayed centered within the main application window.
* @return The FormFrame that was created and displayed.
*/
//===================================================================
public final Frame show()
//===================================================================
{
	return doShowExec(null,false,Gui.NEW_WINDOW);//defaultShowOptions);
}
/**
* This displays the Form non-modally within an already displayed parent Frame.
* @param parent The parent Frame for the Form. A null parent will indicate that a new Window
* should be created.
* @param options this should be one of: 
* <br>Gui.CENTER_FRAME - Size the Frame to the preferred size of its contents and then
* display it centered in the parent frame (this is the most common option).
* <br>Gui.FILL_FRAME - Fill the entire parent frame. 
* <br>Gui.FILL_WIDTH - Fill the entire width of the parent frame. 
* <br>Gui.FILL_HEIGHT - Fill the entire height of the parent frame. 
* <br><b>If options is 0 then the Frame will not be sized, positioned or painted</b> although it will
* be added as a child of the parent Frame. In this case you must size and position the returned
* FormFrame yourself and then call repaintNow().
* @return The FormFrame that was created and displayed.
*/
//===================================================================
public final Frame show(Frame parent,int options) 
//===================================================================
{
	if (parent == null) return doShowExec(parent,false,Gui.NEW_WINDOW);//defaultShowOptions);
	else return doShowExec(parent,false,options);
}
/**
* This can be used to stop the Handle of a Form to indicate that it should be considered
* closed.
* @param h The Handle of the open Form.
* @param retCode The exit code the Form should report. If the retCode
* is IDCANCEL then the Aborted flag is also set otherwise the Success
* flag is set.
*/
//===================================================================
public static void stopFormHandle(Handle h,int retCode)
//===================================================================
{
	if (h == null) return;
	h.returnValue = new Integer(retCode);
	int v = retCode == IDCANCEL ? Handle.Aborted|Handle.Stopped : Handle.Succeeded;
	h.setFlags(v,Handle.Running);
}
private Hashtable softkeyBars;

/**
Create a SoftKeyBar that applies to the entire Form and set the left
and right key functions.
* @return the SoftKeyBar.
*/
//===================================================================
public SoftKeyBar makeSoftKeys(Object left, String leftLabel, Object right, String rightLabel)
//===================================================================
{
	SoftKeyBar sk = makeSoftKeys();
	if (left != null) sk.setKey(1,left,leftLabel);
	if (right != null) sk.setKey(2,right,rightLabel);
	return sk;
}
/**
Create a SoftKeyBar that applies to the entire Form and return it for
you to set the keys in it. If you want a SoftKey which changes for
a particular control, use setSoftKeyBarFor().
**/
//===================================================================
public SoftKeyBar makeSoftKeys()
//===================================================================
{
	SoftKeyBar sb = getSoftKeyBarFor(null);
	if (sb == null) setSoftKeyBarFor(null,sb = new SoftKeyBar());
	return sb;
}
/**
 * Set the SoftKeyBar for a particular Control or the default SoftKeyBar.
 * @param c the particular Control or null to set the default SoftKeyBar.
 * @param bar The SoftKeyBar to use.
 */
//===================================================================
public void setSoftKeyBarFor(Control c, SoftKeyBar bar)
//===================================================================
{
	if (c == this) c = null;
	if (softkeyBars == null) softkeyBars = new Hashtable();
	if (bar == null) return;
	bar.addListener(this);
	softkeyBars.put(c == null || c == this ? "default" : (Object)c,bar);
}
/**
 * Set the SoftKeyBar for a particular set of Controls.
 * @param controls a Vector holding a set of Controls.
 * @param bar The SoftKeyBar to use for each of the Controls.
 */
//===================================================================
public void setSoftKeyBarForAll(Vector controls, SoftKeyBar bar)
//===================================================================
{
	if (softkeyBars == null) softkeyBars = new Hashtable();
	if (bar != null) bar.addListener(this);
	for (int i = 0; i<controls.size(); i++){
		Control c = (Control)controls.get(i);
		softkeyBars.put(c == null ? "default" : (Object)c,bar);
	}
}
/**
 * Get the SoftKeyBar for a particular Control or the default SoftKeyBar.
 * @param c the particular Control or null to get the default SoftKeyBar.
 */
//===================================================================
public SoftKeyBar getSoftKeyBarFor(Control c)
//===================================================================
{
	if (softkeyBars == null) return null;
	return (SoftKeyBar)softkeyBars.get(c == null || c == this ? "default" : (Object)c);
}
/**
 * Remove the SoftKeyBar for a particular Control or the default SoftKeyBar.
 * @param c the particular Control or null to remove the default SoftKeyBar.
 */
//===================================================================
public void removeSoftKeyBarFor(Control c)
//===================================================================
{
	if (softkeyBars == null) return;
	softkeyBars.remove(c == null ? "default" : (Object)c);
}
//public boolean 
/**
* Make sure you call super.onControlEvent() if you override
* this method.
*/
//==================================================================
public void onControlEvent(ControlEvent ev)
//==================================================================
{
	if (ev.type == ev.PRESSED){
		boolean gotIt = true;
		if (ev.target == yes) exit(IDYES);
		else if (ev.target == no) exit(IDNO);
		else if (ev.target == ok) exit(IDOK);
		else if (ev.target == cancel) exit(IDCANCEL);
		else if (ev.target == back) exit(IDBACK);
		else if (ev.action != null) gotIt = handleAction(ev);
		else gotIt = false;
		if (gotIt) ev.consumed = true;
		super.onControlEvent(ev);
	}else if (ev.type == ev.FOCUS_IN){
		if (ev.target instanceof Control && ((Control)ev.target).isChildOf(this)){
			//fixWindowTitle();
			if (softkeyBars != null){
				SoftKeyBar skb = (SoftKeyBar)softkeyBars.get(ev.target);
				if (skb == null) skb = (SoftKeyBar)softkeyBars.get("default");
				if (skb != null) skb.display();
				else if (hideSoftKeyBarIfNotFound)
					SoftKeyBar.hide();
				//System.out.println("Focus: "+ev.target.getClass()+", "+skb);
				//if (skb == null) skb = SoftKeyBar.empty;
				//skb.display();
			}
		}
		super.onControlEvent(ev);
	}else
		super.onControlEvent(ev);
}
/**
Handle an action type event. This method is called by the default handleAction(ControlEvent ev) method and also by
the default menuItemSelected() method and onSoftKey() method.
By default this method checks if the action is equal to any of the EXIT_XXX constants
and if it is it will call exit() with the appropriate exit code.
* @param action the action value.
* @return true to tell the Form the event has been handled, false if not.
*/
//===================================================================
public boolean handleAction(String action)
//===================================================================
{
	if (action.equals(EXIT_IDCANCEL)) exit(IDCANCEL);
	else if (action.equals(EXIT_IDOK))exit(IDOK); 
	else if (action.equals(EXIT_IDYES))exit(IDYES); 
	else if (action.equals(EXIT_IDNO))exit(IDNO); 
	else if (action.equals(EXIT_IDBACK))exit(IDBACK); 
	else return false;
	return true;
}
/**
Handle a ControlEvent that is of type PRESSED and for which the "action" field is
not null. This method is called by the default onControlEvent(). By default this method
returns a call to handleAction(String action).
* @param ev The Control Event.
* @return true to tell the Form the event has been handled and not to pass the event
to its parents, false to continue normal event dispatching.
*/
//===================================================================
public boolean handleAction(ControlEvent ev)
//===================================================================
{
	return handleAction(ev.action);
}
//===================================================================
public void menuItemSelected(MenuItem selected)
//===================================================================
{
	if (selected.action != null) handleAction(selected.action);
}
/**
 * The default onEvent for a Form will call this method
 * if a SoftKeyEvent is sent to the Form. By default it calls handleAction(String action).
 * @param whichKey the SoftKey that was pressed, either 1 or 2.
 * @param action the action associated with the key or menu item.
 * @param selected if a MenuItem was selected, this will be the selected item.
 */
//===================================================================
public void onSoftKey(int whichKey, String action, MenuItem selected)
//===================================================================
{
	if (selected == null) handleAction(action);
	else menuItemSelected(selected);
}
/**
 * This method is called on a SoftKeyEvent. By default it will call se.fireProxyAction()
 * and then call onSoftKey(int,String,MenuItem);
 * @param se the SoftKeyEvent.
 */
//===================================================================
public void onSoftKey(SoftKeyEvent se)
//===================================================================
{
	se.fireProxyAction();
	onSoftKey(se.whichKey, se.action, se.selectedItem);
}

protected Vector fieldHandlers;

public synchronized void addFieldHandler(FieldHandler handler)
{
	if (handler != null) fieldHandlers = mVector.add(fieldHandlers,handler); 
}
//===================================================================
public void onEvent(Event ev)
//===================================================================
{
	if (fieldHandlers != null){
		for (Iterator it = fieldHandlers.iterator(); it.hasNext();){
			FieldHandler fh = (FieldHandler)it.next();
			fh.gotEvent(ev,this);
		}
	}
	if (ev instanceof DataChangeEvent){
		onDataChangeEvent((DataChangeEvent)ev);
	}else if (ev instanceof WindowEvent){
		WindowEvent we = (WindowEvent)ev;
		if (ev.type == WindowEvent.DATA_DROPPED){
			Rect r = Gui.getRectInWindow(this,null,true);
			if (r != null && r.isIn(we.x,we.y)){
				if (we.data instanceof String) filesDropped(new String[]{(String)we.data});
				else if (we.data instanceof String []) filesDropped((String [])we.data);
			}
		}
	}else if (ev instanceof SoftKeyEvent){
		onSoftKey((SoftKeyEvent)ev);
	}else if (ev instanceof MenuEvent && !(ev instanceof ListEvent) && ev.type == MenuEvent.SELECTED){
		MenuEvent me = (MenuEvent)ev;
		if (me.selectedItem instanceof MenuItem){
			MenuItem mi = (MenuItem)me.selectedItem;
			if (mi.data instanceof Control){
				((Control)mi.data).notifyAction();
			}else menuItemSelected(mi);
		}
	}
	super.onEvent(ev);
}
//===================================================================
public void onDataChangeEvent(DataChangeEvent ev){}
//===================================================================
/**
 * This is called when files are dropped on the Form and acceptsDroppedFiles is true.
 * @param fileNames the list of fileNames dropped. 
 */
//===================================================================
public void filesDropped(String [] fileNames) {}
//===================================================================
//===================================================================
public MenuItem createMenuItem(String label, String action, String iconName)
//===================================================================
{
	return createMenuItem(label,action,loadImage(iconName));
}
//===================================================================
public MenuItem createMenuItem(String label, String action, IImage icon)
//===================================================================
{
	if (action == null) action = label;
	MenuItem mi = new MenuItem();
	mi.action = action;
	mi.label = label;
	if (icon != null) {
		mi.image = new IconAndText(icon,label,getFontMetrics());
		((IconAndText)mi.image).textColor = null;
	}
	return mi;
}
/**
Return a MenuItem that will be used as a proxy for a specific buton. When
the menu item is selected the specified button will fire of an action event (ControlEvent.PRESSED)
**/
//===================================================================
public MenuItem createMenuItem(Button button)
//===================================================================
{
	MenuItem mi = createMenuItem(button.text, button.action, button.image);
	mi.data = button;
	button.addListener(this);
	return mi;
}
/**
 * This will return the Saver associated with the Form if one was assigned.
 */
//===================================================================
public IFileSaver getFileSaver()
//===================================================================
{
	if (pl == null) return null;
	return (IFileSaver) pl.getValue("FileSaver",null);
}
/**
* This saves the FileSaver as a property called "FileSaver" in the Form's property list and
* sets up the saver as a listener for DataChangeEvents.
**/
//===================================================================
public void setFileSaver(IFileSaver fs,boolean checkOnExit)
//===================================================================
{
	getProperties().set("FileSaver",fs);	
	if (fs != null) {
		addListener(fs);
		if (checkOnExit) getProperties().setBoolean("CheckFileSaverOnExit",true);
	}
}
//===================================================================
public Form getTopmostForm()
//===================================================================
{
	return Gui.getTopmostForm(this);
}
/**
* If this Form is embedded within another Form, then this will close
* the topmost Form.
**/
//===================================================================
public void closeAll(int exitCode)
//===================================================================
{
	getTopmostForm().close(exitCode);	
}
private SIPFiller sipFiller;
/**
 * Return the SIP filler for the Form if one is present.
 */
SIPFiller getSipFiller()
{
	return sipFiller;
}
/**
 * Create and return a SIPFiller for the Form - or return null
 * if one was alredy created. Add this SIPFiller appropriately in
 * your Form.
 * @param addTo the container within the Form to add the SIP filler. If
 * this is null it is not added to any container and you must add the SIP filler with the constraints:
 * <pre>
 * <b>addLast(sipFiller).setCell(HSTRETCH).setPreferredSize(10,0);</b>
 * </pre>
 * @return and SIPFiller to be used by the Form, or null if one
 * was already added.
 */
public SIPFiller createSipFiller(Panel addTo)
{
	if (sipFiller == null){
		sipFiller = new SIPFiller();
		if (addTo != null)
			addTo.addLast(sipFiller).setCell(HSTRETCH).setPreferredSize(10,0);
		return sipFiller;
	}
	return null;
}
void fillSip(int desktopHeight) throws Exception
{
	if (sipFiller == null) throw new NullPointerException();
	Window w = getWindow();
	Point p = Gui.getPosInParent(sipFiller,w,Point.getCached(0,0));
	int y = p.y;
	//StringBuffer sb = new StringBuffer();
	Rect r = w.getWindowRect(Rect.getCached(),false); y += r.y;
	//sb.append(r.toString()+"\n");
	w.getWindowRect(r,true); y += r.y; //y += r.y;
	//sb.append(r.toString()+"\n");
	//sb.append(y+", "+desktopHeight+", "+Window.visibleHeight);
	//eve.sys.Device.messageBox("Sip",sb.toString(),eve.sys.Device.MB_TYPE_OK);
	int need = y-desktopHeight;
	if (need >= 0){
		sipFiller.setFixedSize(sipFiller.width,sipFiller.height+need);
	}
	((CellPanel)sipFiller.parent).relayout(true);
	//System.out.println("SF: "+sipFiller.getRect());
	p.cache();
	r.cache();
}
boolean unfillSip()
{
	//Vm.debugStackTrace();
	sipFiller.setFixedSize(sipFiller.width,0);
	((CellPanel)sipFiller.parent).relayout(true);
	return true;
}

/**
 * A very useful method for creating a Form with a tabbed panel. It places it in
	the most appropriate place for the platform.
* @param useExpandingTabs set this true if you wish for the created tabbed panel to be the autoexpanding type.
 * @return An array of Controls. The Control at index 0 will be the tabbed panel you should use. If you
	specified an expanding tabbed panel, then the panel at index 1 will be the panel for adding your main controls - otherwise
	it will be null.
*/
//===================================================================
public Control [] addTabbedPanel(boolean useExpandingTabs)
//===================================================================
{
	boolean tb = toolsOnBottom();
	Control [] ret = new Control[2];
	TabbedPanel t;
	if (useExpandingTabs){
		Object [] all = TabbedPanel.getExpandingTabbedPanel(!tb);
		t = (TabbedPanel)all[0];
		SplittablePanel s = (SplittablePanel)all[1];
		addLast(s);
		CellPanel cv = (CellPanel)all[2];
		if (tb){
			CellPanel actual = new CellPanel();
			cv.addLast(actual);
			if (!dontAddSIPFiller){
				sipFiller = new SIPFiller();
				cv.addLast(sipFiller).setCell(HSTRETCH).setPreferredSize(10,0);
			}
			cv = actual;
		}
		ret[1] = cv;
	}else{
		t = new TabbedPanel();
		t.tabLocation = tb ? t.SOUTH : t.NORTH;
		addLast(t);
	}
	ret[0] = t;
	SIPButton.placeIn(t);
	return ret;
}
/**
 * On a PocketPC (or simulated PocketPC) or other device, 
 * this method puts in a SIP filler at the bottom of the Form that will expand
 * automatically when the SIP appears, thereby pushing all other Controls out
 * of the way of the SIP. <b>This is done if you use addToolBar() or addTabbedPanel()
 * so do not call this method if you have already called either of those methods.</b>
 * <p>
 * Unlike addToolBar() and addTabbedPanel() this method is conditional and will only
 * add a SIP filler on a PocketPC device - on all other devices or platforms this method
 * has no effect and returns null. 
 * @return an array of 2 CellPanels. The CellPanel at index 1 is the CellPanel that you
 * should add your main controls to. The CellPanel at index 0 is the CellPanel that
 * appears (and stays below) the SIP and which holds the SIP Button. You can use this
 * CellPanel to add tool buttons, menu bars or whatever control that will fit. You
 * can also leave it empty if you wish.
 */
public CellPanel[] addSIPAdjustment()
{
	if (!toolsOnBottom()) return null;
	return addToolbar();
}
/**
 * A very useful method for creating a Form with a toolbar or menubar. It places it in
	the most appropriate place for the platform. If the platform supports a SIP button on the
	screen (e.g. PocketPC) then a SipButton placeholder will be put in the toolbar and the
	SIP button will always be visible once the Form is displayed.
 * @return An array of CellPanels. The panel at index 0 will be the panel you should use for
	your toolbar. The panel at index 1 will be the panel for adding your main controls.
 */
//===================================================================
public CellPanel [] addToolbar()
//===================================================================
{
	CellPanel [] ret = new CellPanel[2];
	ret[0] = new CellPanel();
	ret[1] = new CellPanel();
	if (toolsOnBottom()){
		addLast(ret[1]);
		if (!dontAddSIPFiller)
			addLast(sipFiller = new SIPFiller()).setCell(HSTRETCH).setPreferredSize(10,0);
		addLast(ret[0]).setCell(HSTRETCH);
		ret[0].setBorder((EDGE_ETCHED & ~BF_RECT)|BF_TOP,2);
		ret[0].setAsBelowSip();
		/*
		new mThread(){
			public void run(){
				try{
					sleep(3000);	
					fillSip();
				}catch(Exception e){}
			}
		}.start();
		*/
	}else{
		ret[0].setBorder((EDGE_ETCHED & ~BF_RECT)|BF_BOTTOM,2);
		addLast(ret[0]).setCell(HSTRETCH);
		addLast(ret[1]);
	}
	ret[0] = SIPButton.placeIn(ret[0]);
	return ret;
}

/**
* Returns true if tools are best added to the bottom of Forms on this platform.
**/
//===================================================================
public boolean toolsOnBottom()
//===================================================================
{
	return SIPButton.hasSipButton() && !Gui.hasSoftKeys;	
}

/**
* This will set the title of the Form if it has not already been set.
**/
//===================================================================
public void defaultTitleTo(String newTitle)
//===================================================================
{
	if (title.equals("untitled") || title == null) title = newTitle;
}
/**
 * Display the Form modally in a new Window and return the Handle that can be used to monitor the Form's open
 * state using waitUntilClosed().
 * @param dismantleOnClose The state to set the dismantleOnClose flag.
 * @return the Handle that can be used to monitor the Form's open
 * state using waitUntilClosed().
 */
//===================================================================
public Handle exec(boolean dismantleOnClose)
//===================================================================
{
	this.dismantleOnClose = dismantleOnClose;
	exec();
	return handle;
}
/**
 * Take any buttons added to the button Vector via doButtons() or addButton()
 * and put them in the destination Vector (or a new Vector). If there are no
 * buttons this will return null.
 * @param dest an optional destination Vector. If null a new one will be created.
 * @return the Vector with the buttons in it, or null if there were no buttons.
 */
public Vector takeButtons(Vector dest)
{
	if (buttons == null || buttons.size() == 0) return null;
	if (dest == null) dest = new Vector();
	dest.addAll(buttons);
	buttons = null;
	return dest;
}
/**
 * Take any buttons added to the button Vector via doButtons() or addButton()
 * and put them in the destination Panel (or a new ButtonBar). If there are no
 * buttons this will return null.
 * @param dest an optional destination Pane. If null a ButtonBar will be created.
 * @return the Panel with the buttons in it, or null if there were no buttons.
 */
public Panel takeButtons(Panel dest)
{
	if (buttons == null || buttons.size() == 0) return null;
	if (dest == null) dest = new ButtonBar();
	for (int i = 0; i<buttons.size(); i++)
		dest.addNext((Control)buttons.get(i));
	buttons = null;
	return dest;
}
/**
 * Take any buttons added to the button Vector via doButtons() or addButton()
 * and put them in a new ButtonBar. If there are no
 * buttons this will return null.
 * @return a ButtonBar with the buttons in it, or null if there were no buttons.
 */
public ButtonBar takeButtons()
{
	return (ButtonBar)takeButtons((Panel)null);
}
/**
 * Used by the VM to execute a Form directly from the command line.
 * @param formClass the name of the Form class.
 * @param args the other arguments on the command line.
 */
public static void formMain(String formClass, String[] args)
{
	Application.startApplication(args);
	Class c = Reflection.forName(formClass);
	Form f = c != null ? (Form)Reflection.newInstance(c) : null; 
	if (f == null){
		new MessageBox("Form Error","Cannot create or execute: "+formClass,MBOK).execute();
		Application.exit(-1);
	}
	Application.exit(f.execute());
}
}
//####################################################
