package eve.ui;
import eve.fx.BufferedGraphics;
import eve.fx.Dimension;
import eve.fx.FontMetrics;
import eve.fx.Graphics;
import eve.fx.Metrics;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.Sound;
import eve.fx.gui.IKeys;
import eve.fx.gui.SurfaceTextEvent;
import eve.sys.Event;
import eve.sys.Vm;
import eve.ui.event.ControlEvent;
import eve.ui.event.DataChangeEvent;
import eve.ui.event.KeyEvent;
import eve.util.Range;

/**
* An mInput is a single line input.
* <p>
* Call <b>getText()</b> and <b>setText()</b> to get/set the text data.
* <p>
* Set <b>isPassword</b> true to indicate that '*' characters should be displayed
* instead of the input text.
* <p>
* Change the <b>columns</b> value to increase the preferred width of the input. This
* value is measured in characters.
**/
//##################################################################
public class Input extends EditControl implements Selectable {
//##################################################################

public int spacing = 3;
{
	columns = 10;
	borderStyle = GuiStyle.inputEdge;
	modify(HasData|WantDrag|WantHoldDown|ShowSIP|TakesKeyFocus,0);
	//setMenu(getClipboardMenu(null));
	setCursor(Vm.IBEAM_CURSOR);
}
/**
* A percent figure. Defaults to 30%
*/
public int minXScroll = 30;
/**
* Set this true so that only '*' characters are displayed.
**/
public boolean isPassword = false;
/**
* You can have the mInput fire an action event when the user single clicks the mInput by setting
* this value to 1, or double clicks by setting the value to 2. A value of 0 (the default) sets
* the mInput to not fire any actions with mouse clicks.
**/
public int clicksToFireAction = 0;
/**
* This is the maximum number of allowed characters, a value of -1 (the default) puts no limit
* on the number of characters.
**/
public int maxLength = -1;
/**
* If this is true then the mInput will fire an action event when enter is pressed, but will not
* automatically pass focus to the next control as it would usually do.
**/
public boolean wantReturn = false;
/**
* If this is true then a DataChange will be fired on each key.
**/
public boolean dataChangeOnEachKey = false;
/**
* This is the default edge to use on mInputs. By default it is EDGE_SUNKEN|BDR_OUTLINE, but
* you can change it to something different (e.g. BF_BOTTOM - to emulate a PalmOS type input).
**/
//public static int inputEdge = EDGE_SUNKEN|BDR_OUTLINE;
/**
* This is a flag used with a generated DataChangeEvent to indicate that the data was changed by the
* enter key being pressed. This would indicate that an Action event will follow
* the DataChangeEvent.
**/
public static final int DATA_CHANGED_BY_ENTER = 0x1;
//==================================================================
public Input() {text = "";}
public Input(String txt) {text = txt;}
//==================================================================

private static Menu menu;

//-------------------------------------------------------------------
protected boolean doShowMenu(Point p)
//-------------------------------------------------------------------
{
	if (menu == null) menu = getClipboardMenu(null);
	menu.keepFrame = false;
	if (getMenu() == null) setMenu(menu);
	return super.doShowMenu(p);
}
//-------------------------------------------------------------------
protected void popupMenuClosed(Menu m)
//-------------------------------------------------------------------
{
	setMenu(null);
	menu = null;
	super.popupMenuClosed(m);
}
//===================================================================
public void setText(String txt)
//===================================================================
{
	int value = modify(Invisible,0);
	try{
		noSelection();
		super.setText(txt);
		newCursorPos(0);
	}catch(Throwable t){
		t.printStackTrace();
	}finally{
		restore(value,Invisible);
	}
	if ((modifiers & Invisible) == 0){
		repaintDataNow();
	}
}
//==================================================================
protected void calculateSizes()
//==================================================================
{
	if (columns == 0) columns = 20;
	Dimension sz = Dimension.getCached(0,0); 
	Metrics.getAverageSize(getFontMetrics(),1,columns,spacing,spacing,sz);
	preferredWidth = sz.width;
	preferredHeight = sz.height;
	sz.cache();
}
//==================================================================
public void cursorToEnd(){newCursorPos(text.length());}
protected int leftMost,numDisplayed;
protected int startSel = 0, endSel = 0, cursorPos = 0;
protected boolean cursorOn = false;
protected boolean hasCursor = false;//true;
private boolean inCursor = false;
//==================================================================

//-------------------------------------------------------------------
protected Rect getDataRect(Rect dest) {return Rect.unNull(dest).set(spacing,spacing,width-spacing*2,height-spacing*2);}
//-------------------------------------------------------------------
//-------------------------------------------------------------------
protected boolean getCharRect(int which,Rect dest)
//-------------------------------------------------------------------
{
	dest.x = dest.y = 0;
	dest.width = dest.height = 0;
	if (which < leftMost || which > text.length()) return false;
	FontMetrics fm = getFontMetrics();
	dest.height = fm.getHeight();
	char[] chars = Vm.getStringChars(getDisplay(leftMost == 0 ? text : text.substring(leftMost)));
	dest.x = which == leftMost ? 0 : fm.getTextWidth(chars,0,which-leftMost);
	dest.width = which-leftMost == chars.length ? 0 : fm.getTextWidth(chars,0,which-leftMost+1)-dest.x;
	/* This only works for regular fonts.
	if (false){
		int dx = 0;
		int dw = 0;
		String s = getDisplay(text);
		for (int i = leftMost; i<=which; i++){
			dx += dw;
			if (i == text.length()) dw = 0;
			else dw = fm.getCharWidth(s.charAt(i));
		}
		System.out.println(dx+", "+dw+" = "+dest.x+", "+dest.width);
	}
	*/
	if (dest.x >= width+spacing) return false;
	return true;
}
//===================================================================
public void ticked(Object id,long elapsed)
//===================================================================
{
	if (id == blinkId){
		//System.out.println("BI: "+System.currentTimeMillis());
		if (!tipIsDisplayed() && Gui.requestPaint(this)) paintCursor();
		blinkId = Window.requestCursorBlink(this,500);
	}else 
		super.ticked(id,elapsed);
}


//
//===================================================================
public void gotFocus(int how)
//===================================================================
{
	boolean newFocus = !inFocus;
	inFocus = true;
	justGotFocus = (how == ByPen || how == ByDeferredPen);
	if (inputFlags == -1){
		inputFlags = 0;
		if (Gui.useNativeTextInput) inputFlags |= getBestPassiveFlags();
	}
	//
	if ((modifiers & DisplayOnly) == 0)
		hasCursor =  true;
	//
	if ((inputFlags & FLAG_PASSIVE) != 0) {
		if (((inputFlags & FLAG_INPUT_ON_FOCUS) != 0) && newFocus)
			if (startActiveInput(how != ByFrameChange)){
				super.gotFocus(how);
				return;
			}else{
				startNativeOnPaint = true;
			}
		repaintNow();
		super.gotFocus(how);
		return;
	}
	//
	boolean ch = !text.equals(oldText);
	oldText = text;
	if (how != ByFrameChange)
		if (checkSipCoverage()){
			super.gotFocus(how);
			return;
		}
	if (ch) notifyDataChange();
	startLocalInput(how != ByFrameChange);
	super.gotFocus(how);
}

//-------------------------------------------------------------------
protected void stopAllInput()
//-------------------------------------------------------------------
{
	cursorOn = false;
	blinkId = null;
	doingLocalInput = false;
	noSelection();
	newCursorPos(0);
}
//===================================================================
public void lostFocus(int how)
//===================================================================
{
	inFocus = false;
	blinkId = null;
	cursorOn = false;
	doingLocalInput = false;
	if (menuIsActive()) return;
	hasCursor = false;
	noSelection();
	newCursorPos(0);
	if (!text.equals(oldText)){
		oldText = text;
		notifyDataChange();
	}
}
//===================================================================
public void stopActiveInput()
//===================================================================
{
	if ((inputFlags & FLAG_PASSIVE) != 0)
		stopAllInput();
}

//===================================================================
public void updateText(boolean generateEvent)
//===================================================================
{
	updateText(generateEvent,false);
}
//-------------------------------------------------------------------
protected void updateText(boolean generateEvent, boolean fromEnter)
//-------------------------------------------------------------------
{
	boolean doEvent = (!text.equals(oldText)) && generateEvent;
	oldText = text;
	if (doEvent){
		DataChangeEvent dev = new DataChangeEvent(DataChangeEvent.DATA_CHANGED,this);
		if (fromEnter) dev.flags |= DATA_CHANGED_BY_ENTER;
		notifyDataChange(dev);
	}
}

/**
This is the password character to use for the input. By default it is '*'
**/
public char passwordCharacter = PasswordDisplayMaker.defaultPasswordCharacter;
/**
* This String will be used to display characters when the isPassword option is used.
**/
//===================================================================
public static String hidden = "***********************";

private PasswordDisplayMaker myPassword;
/**
If isPassword is false, this returns 0, otherwise it will return the password character
to use for display.
**/
//===================================================================
public char getPasswordCharacter()
//===================================================================
{
	return isPassword ? passwordCharacter : 0;
}
/**
* This gets the String to represent on screen the data String provided.
* If the mInput is <b>not</b> a password, this will just return the provided
* String itself. Otherwise it returns a String of "hidden" characters equal
* in length to the data String.
* @param s The data String to display. 
*/
//===================================================================
public String getDisplay(String s)
//===================================================================
{
	if (s == null) return "";
	if (!isPassword) return s;
	if (myPassword == null) myPassword = new PasswordDisplayMaker(passwordCharacter);
	return myPassword.getDisplay(s.length(),passwordCharacter);
}
//-------------------------------------------------------------------
protected void doPaintData(Graphics gr)
//-------------------------------------------------------------------
{
	int flags = getModifiers(true);
	boolean transparent = hasModifier(Transparent,false);
	int w = width-4, h = height-4;
	if (hasModifier(PaintOutsideOnly,false)) return;
	fix();
	if (width <= 0 || height <= 0) return;
	BufferedGraphics bg = bufferGraphics(gr,2,2,w,h,0);
	if (bg == null) return;
	bg.translateToZero();
	Graphics g = bg.getGraphics();
	getColors(hasCursor,flags);
	g.setColor(colors[1]);
	if (!transparent) g.fillRect(0,0,w,h);
	g.setColor(colors[0]);
	try{
		if (!doingNativeInput){
			if (leftMost < text.length()){
				String s = text.substring(leftMost,text.length());//leftMost+numDisplayed);
				g.setFont(getFont());
				g.drawText(getDisplay(s),spacing-2,spacing-2);
			}
			int s = startSel, e = endSel;
			boolean showSelected = true;
			if (s >= e) showSelected = false;
			else if (e-1 < leftMost) showSelected = false;
			if (showSelected){
				if (s <= leftMost) s = leftMost;
				synchronized(Rect.gBuffer1){
					Rect _rect = Rect.gBuffer1;
					getCharRect(s,_rect);
					int x = _rect.x;
					getCharRect(e-1,_rect);
					int ww = _rect.x-x+_rect.width;
					g.setColor(colors[3]);
					g.fillRect(x+spacing-2,spacing-2,ww,_rect.height);
					g.setColor(colors[2]);
					g.setFont(getFont());
					String st = text.substring(s,e);
					g.drawText(getDisplay(st),x+spacing-2,spacing-2);
				}
			}
		}
		cursorOn = false;
		paintConditionalChar(gr,true);
	}finally{
		if (gr != null && startNativeOnPaint){
			startNativeOnPaint = false;
			startActiveInput(true);
		}
		bg.release();
	}
}
//===================================================================
public synchronized void doPaint(Graphics g,Rect area)
//===================================================================
{
	TextInputParameters tip = getNativeInputParameters();
	if (tip != null){
		tip.repaint(this,area);
		return;
	}
	int flags = getModifiers(true);
	if (!((flags & Invisible) == 0)) return;
	doBackground(g);
	if (!hasModifier(PaintDataOnly,false))
		synchronized(Rect.gBuffer1){
		g.draw3DRect(
			getDim(Rect.gBuffer1),
			borderStyle, //Graphics.EDGE_SUNKEN|Graphics.BDR_OUTLINE,
			(flags & DrawFlat) != 0,
			null,
			borderColor);
		}
	doPaintData(g);
}
//BufferedGraphics myCursor;
//OperationTimer ot = new OperationTimer();
//==================================================================
public synchronized void paintCursor()
//==================================================================
{
	paintCursor(null);
}
//==================================================================
public synchronized void paintCursor(Graphics gr)
//==================================================================
{
	paintCursor(gr,false);
}
/**
 * Clear the cursor.
 */
public synchronized void clearCursor()
{
	paintCursor(null,true);
}
/**
 * Paint/toggle the cursor. If clearIt is true then the cursor is always
 * cleared (even if it may not have been there before).
 * @param gr a Graphics to draw to. This may be null.
 * @param clearIt true to clear the cursor.
 */
protected synchronized void paintCursor(Graphics gr,boolean clearIt)
{
	if (doingNativeInput || inCursor || !hasCursor || !amOnTopFrame()) return;
	if (inCursor) return;
	inCursor = true;
	Rect r = Rect.getCached();
	Rect oc = Rect.getCached();
	try{
		int flags = getModifiers(true);
		getColors(hasCursor,flags);
		if (!getCharRect(cursorPos,r)) return;
		BufferedGraphics bg = bufferGraphics(gr,r.x+spacing,spacing,2,r.height,0,null);//myCursor);
		if (bg == null) return;
		Graphics gg = bg.getGraphics();
		boolean co = cursorOn;
		cursorOn = false;
		repaintNow(gg,null);
		cursorOn = co;
		bg.translateToZero();
		if (!clearIt){
			if (!cursorOn){
				gg.setDrawOp(Graphics.DRAW_XOR);
				gg.setColor(colors[0]);
				gg.fillRect(0,0,2,r.height);
				cursorOn = true;
			}else{
				cursorOn = false;
			}
		}else
			cursorOn = false;
		bg.release();
	}finally{
		inCursor = false;
		r.cache();
		oc.cache();
	}

}
//==================================================================
public boolean paintLastChar(Graphics g) {return paintLastChar(g,false);}
public boolean paintLastChar(Graphics g,boolean eraseIt)
//==================================================================
{
	FontMetrics fm = getFontMetrics();
	if (!fm.isRegular()) return false;
	int l = text.length();
	if (l == 0) return true;
	Rect r = new Rect(0,0,0,0);
	getCharRect(l-1,r);
	getColors(hasCursor,0);
	BufferedGraphics bg = bufferGraphics(g,r.x+spacing,r.y+spacing,r.width,r.height,0);
	if (bg == null) return false;
	try{
		g = bg.getGraphics();
		g.setColor(colors[1]);
		//if (!enabled(this) || !editable(this));
		boolean transparent = hasModifier(Transparent,false);
		if (!transparent) 
			g.fillRect(r.x+spacing,r.y+spacing,r.width,r.height);
		if (!eraseIt) {
			g.setColor(colors[0]);
			g.setFont(getFont());
			String lc = getDisplay(text.substring(l-1,l));
			g.drawText(lc,r.x+spacing,r.y+spacing);
			//g.drawRect(r.x+spacing,r.y+spacing,r.x+spacing+10,r.y+spacing+10);
			//g.drawLine(0,10,width,10);
		}
		synchronized(Rect.gBuffer1){
			g.draw3DRect(
				getDim(Rect.gBuffer1),
				borderStyle, //Graphics.EDGE_SUNKEN|Graphics.BDR_OUTLINE,
				(modifiers & DrawFlat) != 0,
				null,
				borderColor);
			}
	}finally{
		bg.release();
	}
	return true;
}
private static char[] charBuff = new char[3];

//===================================================================
public boolean paintConditionalChar(Graphics gr,boolean underlined)
//===================================================================
{
	BufferedGraphics bg = bufferGraphics(gr);
	if (bg == null) return false;
	Graphics g = bg.getGraphics();
	try{
		if (conditionalCharIndex < 0 || conditionalCharIndex >= text.length()) return true;
		Rect r = new Rect(0,0,0,0);
		if (getCharRect(conditionalCharIndex,r)){
			getColors(hasCursor,0);
			g.setColor(colors[1]);
			g.fillRect(r.x+spacing,r.y+spacing,r.width,r.height);	
			g.setColor(colors[0]);
			g.setFont(getFont());
			charBuff[0] = charBuff[1] = text.charAt(conditionalCharIndex);
			charBuff[2] = 0;
			g.drawText(charBuff,0,underlined ? 3:1,r.x+spacing,r.y+spacing);
			return true;
		}else
			return false;
	}finally{
		bg.release();
	}
}

//==================================================================
protected boolean fix()
//==================================================================
{
	int lastLeft = leftMost;
	if (!hasCursor) leftMost = cursorPos = 0;
	if (cursorPos < 0 || cursorPos > text.length()) cursorPos = text.length();
	int w = width-spacing*2-2;
	numDisplayed = 0;
	int cw = getFontMetrics().getCharWidth('0');
	if (cw == 0) cw = 10;
	int toAdd = ((width/cw)*minXScroll)/100;
	if (toAdd < 1) toAdd = 1;
	synchronized(Rect.gBuffer1){
		Rect r = Rect.gBuffer1;
		if (leftMost > text.length()) leftMost = 0;
		while(cursorPos < leftMost) leftMost -= toAdd;
		if (leftMost < 0) leftMost = 0;
		while(cursorPos > leftMost) {
			getCharRect(cursorPos,r);
			if (r.x+r.width < w) break;
			leftMost += toAdd;
		}
	}
	if (cursorPos < leftMost) leftMost = cursorPos;
	/*
	boolean gotCursor = false;
	FontMetrics fm = getFontMetrics();
	String s = getDisplay(text);
	getCharRect(cursorPos,_
	*/
	/*
	for (int i = leftMost; i<=text.length(); i++){
		int need = 2;
		if (i != text.length()) need = fm.getCharWidth(s.charAt(i));
		if (need+w > width){
			if (gotCursor) break;
			else if (numDisplayed != 0){
				w -= fm.getCharWidth(s.charAt(leftMost));
				leftMost++;
				numDisplayed--;
			}
		}
		if (i != text.length()) {
			numDisplayed++;
			w += need;
		}
		if (i == cursorPos) gotCursor = true;
	}
	*/
	return leftMost != lastLeft;
}
/**
* This is used to validate any new text. This implementation will first call the validator's
* isValidText() method (if a validator has been set) and then it will check that the text
* does not exceed the maxLength value (if maxLength is not negative).
**/
//===================================================================
public boolean validateText(String what)
//===================================================================
{
	if (validator != null && !validator.isValidText(what)) {
		Sound.beep();
		return false;
	}
	if (maxLength >= 0 && what.length() > maxLength){
		Sound.beep();
		return false;
	}
	return true;
}
public void newText(String what,int newCursorPos) {newText(what,newCursorPos,true);}
public void newText(String what,int newCursorPos,boolean repaintAll)
{
	int flags = getModifiers(true);
	if (!(((flags & (NotEditable|DisplayOnly)) == 0) || ((flags & NotAnEditor) != 0))) {
		Sound.beep();
		return;
	}
	if (!validateText(what)) return;
	text = what;
	if (dataChangeOnEachKey) updateText(true);
	try{
		boolean fixed = fix();
		if (cursorOn) paintCursor();
		cursorPos = newCursorPos;
		if (!repaintAll && cursorPos == text.length()){
			if (fixed) {
				repaintDataNow();
			}else {
				if (!paintLastChar(null)) repaintDataNow();
				//System.out.println("Last char only!");
			}
		}else
			repaintDataNow();
		if (conditionalCharIndex != -1) paintConditionalChar(null,true);
	}finally{
	}
}
protected void moveCursorPos(int where,boolean takeSelection)
{
	clearCursor();
	boolean lostSelection = false;
	boolean wasLeft = false;
	boolean moveLeft = where < cursorPos;

	if (takeSelection) {
		if (!hasSelection()) startSel = endSel = cursorPos;
		if (cursorPos <= startSel) wasLeft = true;
	}
	if (!takeSelection) lostSelection = noSelection();
	cursorPos = where;
	if (cursorPos < 0) cursorPos = 0;
	if (cursorPos >= text.length()) cursorPos = text.length();
	if (takeSelection){
		if (wasLeft){
			if (moveLeft) startSel = cursorPos;
			else if (cursorPos > endSel) endSel = cursorPos;
			else startSel = cursorPos;
		}else{
			if (!moveLeft) endSel = cursorPos;
			else if (cursorPos < startSel) startSel = cursorPos;
			else endSel = cursorPos;
		}
	}
	if (fix() || lostSelection || takeSelection) repaintDataNow();
	else paintCursor();
}
protected void newCursorPos(int where,boolean takeSelection)
{
	if (cursorOn) paintCursor();
	boolean lostSelection = false;
	if (!takeSelection) lostSelection = noSelection();
	cursorPos = where;
	if (cursorPos < 0) cursorPos = 0;
	if (cursorPos >= text.length()) cursorPos = text.length();
	if (fix() || lostSelection || takeSelection) repaintDataNow();
	else paintCursor();
}
protected void newCursorPos(int where) {newCursorPos(where,true);}


//===================================================================
public boolean deleteSelection()
//===================================================================
{
	if (startSel == endSel) return false;
	String newText = text.substring(0,startSel);
	if (endSel < text.length()) newText += text.substring(endSel,text.length());
	text = newText;
	cursorPos = startSel;
	if (cursorPos > text.length()) cursorPos = text.length();
	noSelection();
	if (dataChangeOnEachKey) updateText(true);
	return true;
}
//===================================================================
public boolean replaceSelection(Object with)
//===================================================================
{
	deleteSelection();
	String in = clipboardToString(with);
	text = text.substring(0,cursorPos)+in+text.substring(cursorPos,text.length());
	cursorPos += in.length();
	if (dataChangeOnEachKey) updateText(true);
	return true;
}
//===================================================================
public Object getSelection()
//===================================================================
{
	if (!hasSelection() || isPassword) return new String();
	return text.substring(startSel,endSel);
}

//-------------------------------------------------------------------
protected void startLocalInput(boolean selectAll)
//-------------------------------------------------------------------
{
	doingLocalInput = true;
	boolean passive = (inputFlags & FLAG_PASSIVE) != 0;
	if (selectAll) this.selectAll();
	if (passive) repaintNow();
	blinkId = Window.requestCursorBlink(this,500);
}
//
// This is used to start the native input if possible. If it is not possible
// then local input is used instead.
//
//-------------------------------------------------------------------
protected void setNativeInputFlags(TextInputParameters tip)
//-------------------------------------------------------------------
{
	if (isPassword) {
		tip.flags |= TextInputParameters.FLAG_PASSWORD;
		tip.passwordCharacter = passwordCharacter;
	}
	if (dataChangeOnEachKey) tip.flags |= tip.FLAG_EVENT_ON_EACH_KEY;
}

//===================================================================
public void onKeyEvent(KeyEvent ev)
//===================================================================
{
	try{
	if (ev.type == KeyEvent.KEY_PRESS && (ev.modifiers & IKeys.CONTROL) != 0)
		if (isSomeonesHotKey(ev)) return;
	ev.keypadToKey();
	if (ev.type != ev.KEY_PRESS){
		super.onKeyEvent(ev);
		return;
	}
	//
	ev = checkInputMethod(ev);
	//
	if (ev.key == IKeys.MENU && parent instanceof ComboBox){
		((ComboBox)parent).doActionKey(ev.key);
		return;
	}
	if ((ev.modifiers & (IKeys.CONTROL|IKeys.SPECIAL)) == (IKeys.CONTROL)){
		if (ev.key != 24 && ev.key != 3 && ev.key != 22){
			super.onKeyEvent(ev);
			return;
		}else{
			ev.modifiers |= IKeys.SPECIAL;
		}
	}
	int flags = getModifiers(true);
	boolean editable = (((flags & (Disabled|NotEditable|DisplayOnly)) == 0) || ((flags & NotAnEditor) != 0));
	boolean shifted = ((ev.modifiers & IKeys.SHIFT) == IKeys.SHIFT);
	boolean ctrled = ((ev.modifiers & IKeys.CONTROL) == IKeys.CONTROL);
	if (validator != null) {
		if (!validator.isValidKeyPress(ev)) {
			Sound.beep();
			return;
		}
	}
	boolean passive = (inputFlags & FLAG_PASSIVE) != 0;
	if (passive && !doingLocalInput){
		editable = false;
		if (ev.key == IKeys.ENTER || ev.key == ' ' || ev.isActionKey()) {
			startActiveInput(true);
			return;
		}else{
			super.onKeyEvent(ev);
			return;
		}
	}
	//
	if ((ev.modifiers & IKeys.SPECIAL) != 0){
		if (ev.key == IKeys.ESCAPE){
			if (isSomeonesHotKey(ev)) return;
			text = oldText == null ? "" : oldText;
			if (!passive) selectAll();
			else stopActiveInput();
			postEvent(new ControlEvent(ControlEvent.CANCELLED,this));
			checkExitKey(ev);
			return;
		}else if (ev.key == IKeys.ENTER || ev.isActionKey()) {
			if (isSomeonesHotKey(ev)) return;
			if (!checkExitKey(ev))
				if (!wantReturn) 
					if (tryNext(true)) return;
			selectAll();
			updateText(true,true);
			notifyAction();
			//oldText = text;
		}else if (ev.key == IKeys.BACKSPACE && cursorPos != 0 && editable){
			if (deleteSelection()) 	repaintDataNow();
			else if (cursorPos == text.length()) {
				if (cursorOn) paintCursor();
				boolean didPaint = paintLastChar(null,true);
				text = text.substring(0,cursorPos-1);
				newCursorPos(cursorPos-1,false);
				if (!didPaint) repaintDataNow();
				if (dataChangeOnEachKey) updateText(true);
			}else
				newText(text.substring(0,cursorPos-1)+text.substring(cursorPos,text.length()),cursorPos-1);
		}else if (ev.key == IKeys.DELETE || ev.key == 24){
			if ((cursorPos < text.length() || startSel != endSel) && editable){
				if (shifted || ev.key == 24) {
					toClipboard(true);
					fix();
					repaintDataNow();
				}else if (deleteSelection()) 	repaintDataNow();
				else 
					newText(text.substring(0,cursorPos)+text.substring(cursorPos+1,text.length()),cursorPos);
			}
		}else if (ev.key == IKeys.INSERT || ev.key == 3 || ev.key == 22){
			if ((ctrled && ev.key == IKeys.INSERT) || (ev.key == 3 && ctrled)){
				if (hasSelection())
					toClipboard(false);
			}else if ((shifted && ev.key == IKeys.INSERT)|| (ev.key == 22 && ctrled)){
				fromClipboard();
				fix();
				repaintDataNow();
			}else if (ev.key == IKeys.INSERT && !shifted && !ctrled){
				clearCursor();
				//inputState ^= STATE_OVERWRITE;
			}
			
			/*
			if ((ev.key == 3 || ctrled) && hasSelection()){
				toClipboard(false);
			}else if ((shifted || ev.key == 22) && editable){
				fromClipboard();
				fix();
				repaintDataNow();
			}
			*/
		}else if (ev.key == IKeys.END){
			moveCursorPos(text.length(),shifted);
		}else if (ev.key == IKeys.HOME){
			moveCursorPos(0,shifted);
		}else if (ev.key == IKeys.LEFT){
			moveCursorPos(cursorPos-1,shifted);
		}else if (ev.key == IKeys.RIGHT){
			moveCursorPos(cursorPos+1,shifted);
		}else{
			super.onKeyEvent(ev);
		}
	}else if (ev.key >= 32){
		if (!editable){
			Sound.beep();
			return;
		}
		char ch = (char)ev.key;
		/* Don't really need to do this - may leave out later.
		*/
		if ((ev.modifiers & IKeys.SHIFT) == IKeys.SHIFT)
			ch = Character.toUpperCase(ch);
		if (textCase == CASE_UPPER)
			ch = Character.toUpperCase(ch);
		else if (textCase == CASE_LOWER)
			ch = Character.toLowerCase(ch);
		else if (textCase == CASE_NUMBERS && !Character.isDigit(ch)){
			Sound.beep();
			return;
		}
		boolean redoData = deleteSelection();
		if (textCase == CASE_SENTENCE && cursorPos == 0)
			ch = Character.toUpperCase(ch);
		String nt = "";
		int nc = cursorPos+1;
		boolean stayHere = conditionalCharIndex != -1;
		if ((ev.modifiers & IKeys.REPLACEMENT) == 0){
			keepConditionalChar();
			nt = text.substring(0,cursorPos)+ch+text.substring(cursorPos,text.length());
			stayHere = false;
		}else{
			//int dest = conditionalCharIndex == -1 ? 
			nt = text.substring(0,conditionalCharIndex)+ch+text.substring(conditionalCharIndex+1,text.length());
		}
		if ((ev.modifiers & IKeys.CONDITIONAL) == 0){
			keepConditionalChar();
		}else{
			conditionalCharIndex = stayHere ? cursorPos-1 : cursorPos;
			conditionalChar = ch;
		}
		newText(nt,cursorPos+(stayHere ? 0 : 1),redoData);		
	}else{
		super.onKeyEvent(ev);
	}
	}finally{
		//ot.end();
		//System.out.println(ot);
	}
}

public int findPressedChar(Point where)
{
	int ret = leftMost;
	int x = spacing;
	FontMetrics fm = getFontMetrics();
	if (where.x < x) return ret;
	int wx = where.x;
	char[] chars = Vm.getStringChars(getDisplay(leftMost == 0 ? text : text.substring(leftMost)));
	if (chars.length == 0) return text.length();
	int top = chars.length, bottom = -1;
	while(true){
		if (bottom >= top) return top+leftMost;
		int c = ((top-bottom)/2)+bottom;
		int s = c == 0 ? 0 : fm.getTextWidth(chars,0,c);
		int w = fm.getTextWidth(chars,0,c+1)-s;
		s += spacing;
		if (wx < s) top = c;
		else if (wx >= s+w) bottom = c+1;
		else{
			int mid = s+w/2;
			if (wx > mid) c++;
			return c+leftMost;
		}
	}
	/*
	String s = getDisplay(text);
	for (int i = leftMost; i<text.length(); i++){
		int fw = fm.getCharWidth(s.charAt(i));
		x += fw/2;
		if (where.x < x) return i;
		x += (-fw/2)+fw;
	}
	*/
}
//==================================================================
public boolean hasSelection() {return startSel != endSel;}
public void selectAll() {startSel = 0; endSel = text.length(); newCursorPos(text.length());}
public boolean noSelection() {boolean ret = (startSel != endSel); startSel = endSel = 0; return ret;}
private boolean dontDrag = false;
private boolean shouldDeselect = false;
//==================================================================
public void penPressed(Point where)
//==================================================================
{
	//System.out.println(doingLocalInput+", "+inputFlags);
	if (((inputFlags & FLAG_PASSIVE) != 0)&& !doingLocalInput){
		startActiveInput(true);
		return;
	}
	//System.out.println("Pressed! JGF: "+justGotFocus);
	if (menuIsActive()) {
		//System.out.println("My menu is active");
		menuState.closeMenu();
	}
	//holdDownPause = 500;
	modify(WantDrag|WantHoldDown,0);
	shouldDeselect = true;
	if (justGotFocus) {
		shouldDeselect = false;
		modify(0,WantDrag);
		super.penPressed(where);
		justGotFocus = false;
		if (clicksToFireAction == 1) {
			selectAll();
			notifyAction();
		}
		return;
	}
	newCursorPos(findPressedChar(where));
}
//===================================================================
public void penReleased(Point where)
//===================================================================
{
	if (menuIsActive()) return;
	if (shouldDeselect && hasSelection()) {
		noSelection();
		newCursorPos(findPressedChar(where));
		repaintDataNow();
	}
}
//==================================================================
public void penDoubleClicked(Point where)
//==================================================================
{
	selectAll();
	//InputPopupForm.popupForm.popup(this,getWindow().getDim(null)); //FIX: Remove this line.
	if (hasModifier(ShowSIP,false)) Gui.setSIP(getWindow(),1);
	if (clicksToFireAction == 2) notifyAction();
	//checkSipCoverage();
}
int startDragPos;
//==================================================================
public void startDragging(DragContext dc)
//==================================================================
{
	if (menuIsActive()) return;
	/* For testing only.
	if (true){
		startDragAndDrop(getText(),true,false);
		return;
	}
	*/
	startSel = endSel = startDragPos = cursorPos;//findPressedChar(dc.curPoint);
	shouldDeselect = false;
	repaintDataNow();
}
//==================================================================
public void dragged(DragContext dc)
//==================================================================
{
	if (menuIsActive()) return;
	doDragAt(dc.curPoint);
}
//==================================================================
protected void doDragAt(Point p)
//==================================================================
{
	Rect r = getRect();
	int which = 0;
	startSel = endSel = startDragPos;
	if (p.x < 0) {
		which = leftMost-1;
		if (which < 0) which = 0;
	}else if (p.x >= r.width){
		which = leftMost+numDisplayed+1;
		if (which > text.length()) which = text.length();
	}else which = findPressedChar(p);
	if (which > startDragPos) endSel = which;
 	else startSel = which;
	newCursorPos(which,true);
}
//==================================================================
public void resizeTo(int width,int height)
//==================================================================
{
	super.resizeTo(width,height);
	if (Gui.focusedControl() == this || true) {
		leftMost = 0;
		fix();
	}
}
//===================================================================
public boolean acceptsData(Object data,DragContext how)
//===================================================================
{
	if (how != null) return false;
	return super.acceptsData(data,how);
}
//===================================================================
public void update()
//===================================================================
{
	fix();
	repaintDataNow();
}
/**
 * Use this to get the selection range. If this returns 0 then there is no selection range.
 * @return A Range which indicates the first and last index of the selection.
 */
//===================================================================
public Range getSelectionRange()
//===================================================================
{
	if (!hasSelection()) return null;
	return new Range(startSel,endSel);
}
//===================================================================
public int getCursorPosition()
//===================================================================
{
	return cursorPos;
}

//===================================================================
public void onEvent(Event ev)
//===================================================================
{
	if (inputFlags == -1){
		inputFlags = 0;
		if (Gui.useNativeTextInput) inputFlags |= getBestPassiveFlags();
	}
	if (ev instanceof SurfaceTextEvent){
		SurfaceTextEvent te = (SurfaceTextEvent)ev;
		if (te.type == SurfaceTextEvent.TEXT_ENTERED){
			nativeInputStopped();
			if ((te.flags & te.FLAG_TEXT_WAS_ENTERED) != 0){
				setText(fixCase(te.entered));
				updateText(true,true);
			}
			if ((te.flags & te.FLAG_CLOSED_BY_ENTER_KEY) != 0){
				if (wantReturn) notifyAction();
				else if ((inputFlags & FLAG_KEEP_FOCUS_AFTER_INPUT) == 0) tryNext(true);
			}
			else if ((te.flags & te.FLAG_CLOSED_BY_UP_KEY) != 0)
				tryNext(false);
			else if ((te.flags & te.FLAG_CLOSED_BY_DOWN_KEY) != 0)	
				tryNext(true);
		}else if (te.type == SurfaceTextEvent.TEXT_CHANGED){
			setText(fixCase(te.entered));
			updateText(true,true);
		}
	}else
		super.onEvent(ev);
}
/**
 * Input text embedded within another Control using a new Input. 
 * This uses the InputInsideForm class to do this.
 * @param originalText
 * @param parent
 * @param x
 * @param y
 * @param width
 * @param height
 * @param options
 * @return
 */
public static String execute(String originalText,Control parent, int x, int y, int width, int height, int options)
{
	Input in = new Input();
	in.setText(originalText);
	int got = new InputInsideForm(in,options).execute(parent,x,y,width,height);
	if (got == Form.IDCANCEL) return null;
	return in.getText();
}
/**
 * Input text embedded within another Control using this Input. 
 * This uses the InputInsideForm class to do this.
 * @param parent
 * @param x
 * @param y
 * @param width
 * @param height
 * @param options
 * @return
 */
public int execute(Control parent, int x, int y, int width, int height, int options)
{
	return new InputInsideForm(this,options).execute(parent,x,y,width,height);
}
//##################################################################
}
//##################################################################


