package eve.ui;

import eve.fx.Sound;
import eve.ui.Control;
import eve.ui.Gui;
/**
 * This is a convenience class for handling Flashing messages on a Control,
 * whether it be brief message flashing or messages that stay up.
 */
public class MessageFlasher {

	private Object flashed;
	private Control parent;

	/**
	 * This option is enabled by default. It indicates that Sound.beep() should
	 * be called when the message is shown.
	 */
	public static final int OPTION_BEEP = Gui.FLASH_BEEP;
	/**
	 * Flash options. By default OPTION_BEEP is enabled.
	 */
	public int flashOptions = OPTION_BEEP;
	/**
	 * The beep typoe to use. By default it is Sound.BEEP_DEFAULT.
	 */
	public int beepType = Sound.BEEP_DEFAULT;
	/**
	 * This is the default flash time - initially set to 750.
	 */
	public int flashTime = 750;
	/**
	 * Create a MessageFlasher for flashing messages on a parent Control.
	 * @param parent the non-null parent control.
	 */
	public MessageFlasher(Control parent)
	{
		this.parent = parent;
	}
	/**
	 * Create a MessageFlasher that immediately displays a message and
	 * will continue to display it until off() or show() or flash() is called.
	 * @param message the message to display.
	 * @param parent the non-null parent control.
	 * @param beep true to beep when the message is displayed, false not to.
	 */
	public MessageFlasher(String message, Control parent, boolean beep)
	{
		this.parent = parent;
		flashOptions &= ~OPTION_BEEP;
		if (beep) flashOptions |= OPTION_BEEP;
		show(message);
	}
	/**
	 * Create a MessageFlasher that immediately displays a Control and
	 * will continue to display it until off() or show() or flash() is called.
	 * @param toDisplay a custom Control to display.
	 * @param parent the non-null parent control.
	 * @param beep true to beep when the message is displayed, false not to.
	 */
	public MessageFlasher(Control toDisplay, Control parent, boolean beep)
	{
		this.parent = parent;
		flashOptions &= ~OPTION_BEEP;
		if (beep) flashOptions |= OPTION_BEEP;
		show(toDisplay);
	}
	
	private void beep()
	{
		if ((flashOptions & OPTION_BEEP) != 0)
			Sound.beep(beepType);
	}
	/**
	 * Hide any currently displayed message. If none is displayed this has
	 * no effect.
	 */
	public synchronized void off()
	{
		if (flashed != null){
			if (flashed instanceof Form) ((Form)flashed).close(0);
			else Gui.flashMessageOff(flashed);
			flashed = null;
		}
	}
	/**
	 * Start showing the specified message, switching off any other message
	 * that may have been on. This message will stay displayed until flash(), show()
	 * or off() is called.
	 * @param message the message to show.
	 */
	public synchronized void show(String message)
	{
		off();
		flashed = Gui.flashMessageOn(message, parent);
		beep();
	}
	/**
	 * Start showing a specific custom Control, switching off any other message
	 * that may have been on. This message will stay displayed until flash(), show()
	 * or off() is called.
	 * @param toDispaly the Control to show.
	 */
	public synchronized void show(Control toDisplay)
	{
		off();
		flashed = Gui.flashMessageOn(toDisplay, parent);
		beep();
	}
	/**
	 * Flash a message for the default length of time, switching off any previous messages.
	 * @param message the message to show.
	 */
	public synchronized void flash(String message)
	{
		flash(message,flashTime);
	}
	/**
	 * Flash a message for particular length of time, switching off any previous messages.
	 * @param message the message to show.
	 * @param howLong the length of time in milliseconds to display the message.
	 */
	public synchronized void flash(String message,int howLong)
	{
		off();
		Gui.flashMessage(message,howLong,parent,0);
		beep();
	}
}
