package eve.ui;

import java.util.Hashtable;

import eve.fx.BufferedGraphics;
import eve.fx.Color;
import eve.fx.Dimension;
import eve.fx.Drawable;
import eve.fx.Font;
import eve.fx.FontMetrics;
import eve.fx.Graphics;
import eve.fx.IImage;
import eve.fx.Metrics;
import eve.fx.Picture;
import eve.fx.PixelBuffer;
import eve.fx.Point;
import eve.fx.Rect;
import eve.fx.gui.IEventModifiers;
import eve.fx.gui.IKeys;
import eve.fx.gui.RawEvent;
import eve.fx.gui.SurfaceKeyEvent;
import eve.fx.gui.WindowSurface;
import eve.sys.mThread;

/**
 * This is the base implementation of a pure Eve keyboard based input panel.
 */
public class SIPKeypad extends Canvas {
	private WindowSurface myWindow = null;

	private RawEvent re = new RawEvent();

	private void sendKeyEvent(int key, int mods) {
		if (myWindow == null)
			myWindow = getWindow().getWindowSurface();
		re.rawEventType = re.KEY_EVENT;
		re.key = key;
		re.modifiers = mods | IEventModifiers.POST_TO_BACK;
		re.surfaceEventType = SurfaceKeyEvent.KEY_PRESS;
		myWindow.postRawEvent(re);
	}

	private short[] rects;

	private int[] keys;

	private int[] mods;

	private Object display[];

	private int numKeys;

	/**
	 * Set the keys for the keypad.
	 * 
	 * @param numKeys
	 *            the number of keys.
	 * @param keys
	 *            the value for the key event for each key.
	 * @param mods
	 *            the modifier for the key event for each key.
	 * @param rects
	 *            the area on the keypad taken by the key.
	 * @param display
	 *            Objects to display on the keypad. Each item can be a String or
	 *            an IImage or null. If it is null then the value of keys for
	 *            that item is converted to a character and displayed.
	 */
	private void setKeys(int numKeys, int[] keys, int[] mods, short[] rects,
			Object[] display) {
		this.numKeys = numKeys;
		this.keys = keys;
		this.mods = mods;
		this.display = display;
		this.rects = rects;
	}

	public void setKeys(SIPKeypadKeys sk) {
		setKeys(sk.keyCount, sk.keys, sk.mods, sk.rects, sk.displays);
	}

	private Hashtable keySets = new Hashtable();

	public void addKeySet(String name, SIPKeypadKeys keys) {
		keySets.put(name, keys);
	}

	public boolean setKeys(String keysetName) {
		SIPKeypadKeys got = (SIPKeypadKeys) keySets.get(keysetName);
		if (got == null)
			return false;
		setKeys(got);
		return true;
	}

	
	private static char[] toDraw = new char[1];
	private static final Color LockedColor = Color.LightBlue;
	
	private void paintKeyBuffered(Graphics g, FontMetrics f, int whichKey,
			boolean doBack, boolean pressed) {
		int idx = whichKey * 4;
		if ((mods[whichKey] & MODIFIER_IS_LOCKED) != 0) doBack = true;
		if (doBack && !pressed) {
			g.setColor(getBackground());
			if ((mods[whichKey] & MODIFIER_IS_LOCKED) != 0) g.setColor(LockedColor);
			g.fillRect(rects[idx], rects[idx + 1], rects[idx + 2],
					rects[idx + 3]);
		}
		g.setColor(getForeground());
		g.paintRect(rects[idx], rects[idx + 1], rects[idx + 2], rects[idx + 3],
				pressed, true);
		Object got = display[whichKey];
		if (pressed) {
			g.setColor(getBackground());
			g.setBackground(getForeground());
		}
		if (got instanceof String) {
			String s = (String) got;
			int w = f.getTextWidth(s);
			g.drawText(s, rects[idx] + (rects[idx + 2] - w) / 2, rects[idx + 1]
					+ (rects[idx + 3] - f.getHeight()) / 2);
		} else if (got instanceof Drawable) {
			Drawable ii = (Drawable) got;
			int w = ii.getWidth(), h = ii.getHeight();
			ii.draw(g, rects[idx] + (rects[idx + 2] - w) / 2, rects[idx + 1]
					+ (rects[idx + 3] - h) / 2, 0);
		} else {
			char c = (char) keys[whichKey];
			int w = f.getCharWidth(c);
			toDraw[0] = c;
			g.drawText(toDraw, 0, 1, rects[idx] + (rects[idx + 2] - w) / 2,
					rects[idx + 1] + (rects[idx + 3] - f.getHeight()) / 2);
		}
	}

	protected void flashKey(int which, boolean on) {
		int idx = which * 4;
		BufferedGraphics bg = bufferGraphics(null, rects[idx], rects[idx + 1],
				rects[idx + 2], rects[idx + 3], 0);
		if (bg == null)
			return;
		try {
			Graphics g = bg.getGraphics();
			if (g == null)
				return;
			FontMetrics fm = getFontMetrics();
			g.setFont(fm.getFont());
			paintKeyBuffered(g, fm, which, true, on);
		} finally {
			bg.release();
		}
	}

	public void penPressed(Point where) {
		int x = where.x, y = where.y;
		int idx = 0;
		for (int i = 0; i < numKeys; i++, idx += 4) {
			if (x < rects[idx])
				continue;
			if (x > rects[idx] + rects[idx + 2])
				continue;
			if (y < rects[idx + 1])
				continue;
			if (y > rects[idx + 1] + rects[idx + 3])
				continue;
			flashKey(i, true);
			mThread.nap(50);
			flashKey(i, false);
			keyPressed(i);
			return;
		}
	}

	public void doPaint(Graphics g, Rect area) {
		try {
			g.setColor(getBackground());
			g.fillRect(area.x, area.y, area.width, area.height);
			g.setColor(getForeground());
			g.setBackground(getBackground());
			Rect me = Rect.getCached();
			FontMetrics f = getFontMetrics();
			g.setFont(f.getFont());
			try {
				int idx = 0;
				for (int i = 0; i < numKeys; i++) {
					me.set(rects[idx], rects[idx + 1], rects[idx + 2],
							rects[idx + 3]);
					idx += 4;
					if (area.intersects(me)) {
						paintKeyBuffered(g, f, i, false, false);
					}
				}
			} finally {
				me.cache();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private static final int imageWidth = 32;
	private static final int imageHeight = 32;
	
	private static IImage loadAnImage(String name)
	{
		try{
			IImage got = new Picture(name);
			int w = got.getImageWidth(), h = got.getImageHeight();
			if (true){
				PixelBuffer pb = new PixelBuffer(got,new Rect(0,0,w,h),new Dimension(w*2,h*2),null);
				got = pb.toPicture();
				pb.free();
			}
			return got;
		}catch(Exception e){
			return null;
		}
	}
	private static int addLeftColumn(SIPKeypadKeys sk, int startKey, int x,
			int y, int w, int h) {
		int num = startKey;
		int hw = w / 2;
		int p = num;
		num += sk.addColumn(num, 4, x, y, w, h);
		sk.displays[p++] = loadAnImage("eve/tabkey.png");
		sk.displays[p++] = loadAnImage("eve/capslockkey.png");
		sk.displays[p++] = loadAnImage("eve/shiftkey.png");
		sk.displays[p++] = loadAnImage("eve/accentlockkey.png");
		sk.keys[0] = IKeys.TAB; sk.mods[0] |= IKeys.SPECIAL;
		sk.keys[1] = K_SHIFT_LOCK;
		sk.keys[2] = K_SHIFT;
		sk.keys[3] = K_SYMBOL_LOCK;
		//sk.displays[p++] = loadAnImage("eve/menusmall.png");
		return num - startKey;
	}
	private static int addRightColumn(SIPKeypadKeys sk, int startKey, int x,
			int y, int w, int h) {
		int num = startKey;
		int hw = w / 2;
		int p = num;
		num += sk.addColumn(num, 4, x, y, w, h);
		sk.keys[p] = IKeys.BACKSPACE;
		sk.mods[p] = IKeys.SPECIAL;
		sk.displays[p++] = loadAnImage("eve/backkey.png");
		sk.keys[p] = IKeys.ENTER;
		sk.mods[p] = IKeys.SPECIAL;
		sk.displays[p++] = loadAnImage("eve/enterkey.png");
		sk.keys[p] = K_NULL;//K_MENU; 
		sk.displays[p++] = loadAnImage("eve/menusmall.png");
		sk.keys[p] = K_HIDE;
		sk.displays[p++] = loadAnImage("eve/hidekey.png");
		return num - startKey;
	}
	
	private static final char[] lowercaseAlpha = 
		"qwertyuiop@asdfghjkl;:-zxcvbnm,./_".toCharArray();
	private static final char[] lowercaseNumeric = 
		"123456789+0=".toCharArray();
	private static final char[] uppercaseAlpha = 
		"QWERTYUIOP@ASDFGHJKL;:-ZXCVBNM,./_".toCharArray();
	private static final int[] uppercaseNumeric =
		{'#',0xa3,0xa5,'$',8364,'%','&','(',')','~','*','-'};
	
	private static final int K_FIRST = 0x70000000;
	private static final int K_SHIFT = K_FIRST;
	private static final int K_SHIFT_LOCK = K_FIRST+1;
	private static final int K_SYMBOL_LOCK = K_FIRST+2;
	private static final int K_SYMBOL_1 = K_FIRST+3;
	private static final int K_SYMBOL_2 = K_FIRST+4;
	private static final int K_SYMBOL_3 = K_FIRST+5;
	private static final int K_SYMBOL_4 = K_FIRST+6;
	private static final int K_MENU = K_FIRST+7;
	private static final int K_HIDE = K_FIRST+8;
	private static final int K_NULL = K_FIRST+9;
	private static final int K_LAST = K_FIRST+9;
	private static final int MODIFIER_IS_LOCKED = 0x10000000;
	
	private int shiftStatus = 0;
	private static final int SHIFT_SHIFT = 0x1;
	private static final int SHIFT_ALPHA_LOCK = 0x2;
	private static final int SHIFT_SYMBOL_LOCK = 0x4;
	
	private static int symbolKeyIndex;
	
	private void selectSymbol(int which)
	{
		int idx = symbolKeyIndex;
		for (int i = 0; i<4; i++, idx++){
			if (i == (which-1)){
				mods[idx] |= MODIFIER_IS_LOCKED;
			}else{
				mods[idx] &= ~MODIFIER_IS_LOCKED;
			}
			flashKey(idx,false);
		}
	}
	private static int shiftKeyIndex = 1;
	
	private void unlockKey(int key)
	{
		mods[key] &= ~MODIFIER_IS_LOCKED;
		flashKey(key,false);
	}
	private void doDeshift()
	{
		unlockKey(shiftKeyIndex);
		unlockKey(shiftKeyIndex+1);
		shiftStatus &= ~(SHIFT_SHIFT|SHIFT_ALPHA_LOCK);
	}
	private void doShift(boolean withLock)
	{
		mods[shiftKeyIndex+1] |= MODIFIER_IS_LOCKED;
		flashKey(shiftKeyIndex+1,false);
		if (withLock){
			mods[shiftKeyIndex] |= MODIFIER_IS_LOCKED;
			flashKey(shiftKeyIndex,false);
		}
		shiftStatus |= SHIFT_SHIFT;
		if (withLock) shiftStatus |= SHIFT_ALPHA_LOCK;
	}
	private boolean isUppercase()
	{
		return (shiftStatus & (SHIFT_SHIFT|SHIFT_ALPHA_LOCK)) != 0;
	}
	protected void keyPressed(int which) {
		int k = keys[which], m = mods[which];
		if (isUppercase()) m |= IKeys.SHIFT;
		if (k >= K_FIRST && k <= K_LAST ){
			switch(k){
				case K_HIDE:
					Gui.setSIP(getWindow(),0);
					return;
				case K_SYMBOL_1:
				case K_SYMBOL_2:
				case K_SYMBOL_3:
				case K_SYMBOL_4:
				{
					int ws = k-K_SYMBOL_1+1;
					if ((m & MODIFIER_IS_LOCKED) != 0) ws = 0;
					selectSymbol(ws);
					mapKeyset(ws,isUppercase());
					return;
				}
				case K_SHIFT:
				case K_SHIFT_LOCK:
				{
					if ((m & MODIFIER_IS_LOCKED) != 0){
						doDeshift();
						mapKeyset(curKeyset,false);
					}else{
						doShift(k == K_SHIFT_LOCK);
						mapKeyset(curKeyset,true);
					}
					return;
				}
				case K_SYMBOL_LOCK:
				{
					if ((m & MODIFIER_IS_LOCKED) != 0){
						shiftStatus &= ~SHIFT_SYMBOL_LOCK;
						unlockKey(which);
						flashKey(which,false);
						selectSymbol(0);
						mapKeyset(0,isUppercase());
					}else{
						shiftStatus |= SHIFT_SYMBOL_LOCK;
						mods[which] |= MODIFIER_IS_LOCKED;
						flashKey(which,false);
					}
					return;
				}
				default:
					return;
			}
		}else{
			sendKeyEvent(k,m);
			boolean deShift = false;
			if (((shiftStatus & SHIFT_SHIFT) != 0) && ((shiftStatus & SHIFT_ALPHA_LOCK) == 0)){
				deShift = true;
				doDeshift();
			}
			if (curKeyset != 0){ //Showing a symbol.
				if ((shiftStatus & SHIFT_SYMBOL_LOCK) == 0){
					selectSymbol(0);
					mapKeyset(0,isUppercase());
				}else if (deShift){
					mapKeyset(curKeyset,false);
				}
			}else if (deShift){
				mapKeyset(0,false);
			}
		}
	}
	
	private static int[] toInts(Object obj)
	{
		if (obj instanceof int[])
			return (int[])obj;
		if (obj instanceof char[]){
			char[] s = (char[])obj;
			int[] r = new int[s.length];
			for (int i = 0; i<s.length; i++)
				r[i] = (int)s[i] & 0xff;
			return r;
		}
		return null;
	}
	/*
	private static final char[] lowercaseAlpha = 
		"qwertyuiop@asdfghjkl;:-zxcvbnm,./_".toCharArray();
    */
	private static final int[] commonSymbols =
	{'(',')','?','!','~','%','&','^','#','|','@',
			'\'','\"',96,180,8220,8221,8216,8217,176,
			';',':','-',//
			//169,174,8482,
			'<','>','{','}','[',']','\\',',','.','/','_'
			//171,187
	};
	private static final int[] commonNumerics =
	{0xa3,0xa5,'$',162,8364,177,'*',215,247,'+','-','='};
	
	private static final Object[] keyMaps = {
			lowercaseAlpha, lowercaseNumeric,
			uppercaseAlpha, lowercaseNumeric,

			commonSymbols, commonNumerics,
			//new int[]{'`',180,168,8220,8221,8222,8216,8217,8218}
			//,
			
			commonSymbols, commonNumerics,

			lowercaseAlpha, lowercaseNumeric,
			uppercaseAlpha, uppercaseNumeric,
//First set of accents.
			new int[]{
224,225,226,227,228,229,0x101,0x103,0x105,230,
231,0x107,0x10d,
240,0x111,0x10f,
232,233,234,235,0x113,0x115,0x117,0x119,0x11b,
0x123,
305,236,237,238,239,0x129,0x12b,0x12f
			}
			,
			new int[]{
307,309,311,
314,316,318,
320,322,241,
324,326,328
			}
			,
			new int[]{
192,193,194,195,196,197,0x100,0x102,0x104,198,
199,0x106,0x10c,
208,0x110,0x10e,
200,201,202,203,0x112,0x114,0x116,0x118,0x11a,
0x122,
304,204,205,206,207,0x128,0x12a,0x12e
								}
			,new int[]{
306,308,310,					
313,315,317,
319,321,209,
323,325,327

			},
//Second set of accents.
			new int[]{
242,243,244,245,246,333,335,337,339,248,254,
341,343,345,347,349,351,353,
355,357,359,223,32,
249,250,251,252,361,363,365,367,369,371,373
								}
								,
								new int[]{
					253,375,255,
					378,380,382,
					32,32,32,
					32,32,32
								}
								,
								new int[]{
210,211,212,213,214,332,334,336,338,216,222,
340,342,344,346,348,350,352,
354,356,358,223,32,
217,218,219,220,360,362,364,366,368,370,372,
													}
								,new int[]{
					221,374,376,					
					377,379,381,
					32,32,32,
					32,32,32

								},
			
	};
	
	private SIPKeypadKeys theKeys;
	private int mapped = -1;
	
	private void mapKeys(int which)
	{
		if (which != mapped){
			int idx = which*2;
			int[] sk = toInts(keyMaps[idx]);
			keyMaps[idx++] = sk;
			System.arraycopy(sk,0,theKeys.keys,4,sk.length);
			int next = lowercaseAlpha.length+4+5;
			sk = toInts(keyMaps[idx]);
			keyMaps[idx++] = sk;
			System.arraycopy(sk,0,theKeys.keys,next,sk.length);
			mapped = which;
			repaintNow();
		}
	}
	private int curKeyset;
	
	private void mapKeyset(int which, boolean isUppercase)
	{
		int w = which*2;
		curKeyset = which;
		if (isUppercase) w++;
		mapKeys(w);
	}
	private static String symbolKeys[] = {
			"eve/symbols1key.png",
			"eve/esckey.png", //"eve/symbols1key.png",
			"eve/accents1key.png","eve/accents2key.png"
	};
	private static int addKeys(SIPKeypadKeys sk,int startKey, int x, int y,int w,int h)
	{
		int num = startKey;
		int hw = w/2;
		int p = num;
		num += sk.addRow(num,11,x+hw,y,w,h);
		sk.rects[p*4] -= hw; sk.rects[p*4+2] += hw; p = num; sk.rects[(p-1)*4+2] += hw-1;
		num += sk.addRow(num,12,x,y+h-1,w,h);
		p = num;
		num += sk.addRow(num,11,x+hw,y+(h-1)*2,w,h);
		sk.rects[p*4] -= hw; sk.rects[p*4+2] += hw; p = num; sk.rects[(p-1)*4+2] += hw-1;
		int ww = imageWidth+4;//(int)(w*1.2);
		symbolKeyIndex = num;
		num += sk.addRow(num,5,x,y+(h-1)*3,ww,h);
		int total = (w-1)*12+1;
		short was = sk.rects[(num-3)*4];
		sk.rects[(num-2)*4] = (short)(x+total-ww);
		sk.rects[(num-3)*4] = (short)(x+total-(ww*2)+1);
		sk.rects[(num-1)*4] = was;
		sk.rects[(num-1)*4+2] = (short)(total-((ww-1)*4));
		for (int i = 0; i<4; i++){
			sk.keys[symbolKeyIndex+i] = K_SYMBOL_1+i;
			sk.displays[symbolKeyIndex+i] = loadAnImage(symbolKeys[i]);//"S"+(i+1);
		}
		sk.keys[symbolKeyIndex+1] = IKeys.ESCAPE;
		sk.mods[symbolKeyIndex+1] = IKeys.SPECIAL;
		//sk.displays[symbolKeyIndex+1] = " ";
		sk.keys[num-1] = ' ';
		//
		x += w/4+(w-1)*12;
		num += sk.addRow(num,3,x,y,w,h); y += h-1; 
		num += sk.addRow(num,3,x,y,w,h); y += h-1; 
		num += sk.addRow(num,3,x,y,w,h); y += h-1; 
		num += sk.addRow(num,3,x,y,w,h); y += h-1; 
		return num-startKey;
	}	
	private SIPKeypadKeys makeAlphaKeypad()
	{
		font = Application.findFont("keypad", true);
		FontMetrics fm = getFontMetrics(font);
		font = Metrics.getFontForHeight(imageHeight,fm);
		fm = getFontMetrics(font);
		SIPKeypadKeys sk = new SIPKeypadKeys(60);
		int sw = imageWidth*2;
		int w = (fm.getCharWidth('W') + 6 +1) & ~1, h = fm.getHeight() + 4, hw = w/2;
		if (h < imageHeight+4) h = imageHeight+4;
		if (w < sw*2/4) w = sw*2/4;
		int x = 0, y = 3;
		int num = 0;
		num += addLeftColumn(sk,num,x,y,sw,h);
		int is = w/4;
		x += sw+is;
		num += addKeys(sk,num,x,y,w,h);
		x += (w/4*2)+(w-1)*15;
		num += addRightColumn(sk,num,x,y,sw,h);
		sk.keyCount = num;
		int full = 12*(w-1)+3*(w-1)+2*sw+(w/4)*3;
		setPreferredSize(full,h*4+6);
		return sk;
	}
	//private Object[]
	
	protected SIPKeypad()
	{
		font = Application.findFont("gui", true).changeStyle(Font.BOLD);
		Application.addFont(font, "keypad");
	}
	public SIPKeypad(int options) {
		this();
		theKeys = makeAlphaKeypad();
		setKeys(theKeys);
		mapKeys(0);
	}
}