/*
 * Created on Apr 6, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.ui;

import eve.sys.Event;
import eve.sys.EventListener;
import eve.sys.Handle;
import eve.sys.Task;

/**
 * @author Michael L Brereton
 *
 * This is a Task meant to run in a Window's active Event Thread. To use it
 * you must override the abstract public void doRun() method inherited from Task.
 */
//####################################################
public abstract class WindowThreadTask extends Task implements EventListener{

	public Window window;
	private boolean didRun;
	
	private void myRun()
	{
		synchronized(this){
			if (didRun) {
				set(Succeeded);
				return;
			}
			didRun = true;
		}
		run();
	}
	/**
	 * Start the Task running in a separate call of the Window thread by posting
	 * an Event that will eventually cause the run() method to be called.
	 */
	public synchronized Handle start()
	{
		set(Running);
		if (window == null) {
			fail(new IllegalStateException("The Window has not been set."));
			return this;
		}
		Event ev = new Event();
		ev.target = this;
		window.queueEvent(ev);
		return this;
	}
	/**
	 * Start the Task running in a separate call of the Window thread by posting
	 * an Event that will eventually cause the run() method to be called or possibly
	 * run it in the current Thread if the current thread is the active  Window thread
	 * and alwaysInSeparateThread is false.
	 * @param parent the Control in the Window.
	 * @param alwaysInSeparateThread if this is true the method always returns immediately
	 * and requests the task to run in another call of the Window thread.
	 * @return a Handle to the Task.
	 */
	public synchronized void run(Control parent, boolean alwaysInSeparateThread)
	{
		window = parent.getWindow();
		if (window == null) {
			fail(new IllegalStateException("No Window for the Control."));
			return;
		}
		if (alwaysInSeparateThread || !window.inCurrentEventThread()) start();//return start();
		else{
			myRun();
		}
	}
	/**
	 * Execute the Task, wait until it is complete and return the state of the Task. 
	 * @param parent a Control contained in the Window the Task should run in. getWindow()
	 * is called on the Control to determine the Window. This parameter can be the Window
	 * itself.
	 * @return the state of the Handle upon completion.
	 * @throws IllegalStateException if the parent is not contained within a Window.
	 */
	public synchronized int execute(Control parent) throws IllegalStateException
	{
		window = parent.getWindow();
		if (window == null) throw new IllegalStateException("No Window for the Control.");
		Thread me = Thread.currentThread();
		if (!(me instanceof WindowEventThread)){
			start();
			waitUntilCompletion();
			return check();
		}
		WindowEventThread wt = (WindowEventThread)me;
		while(true){
			boolean doNow = false;
			synchronized(wt){
				doNow = me == window.activeThread;
				wt.blockingOn = doNow ? null : this;
				Thread.interrupted();
			}
			if (doNow){
				myRun();
				return check();
			}else{
				start();
				try{
					//System.out.println("Waiting: "+wt);
					waitUntilStopped();
					synchronized(wt){
						wt.blockingOn = null;
						Thread.interrupted();
					}
					return check();
				}catch(InterruptedException e){
					//System.out.println("Interrupted!");
					continue;
				}
			}
		}
	}
	/**
	 * Try to run execute(Control parent) but catch any Exception (not Error) that may
	 * occur.
	 * @param parent a Control contained in the Window the Task should run in. getWindow()
	 * is called on the Control to determine the Window. This parameter can be the Window
	 * itself.
	 * @return the state of the Handle upon completion.
	 */
	public int tryExecute(Control parent)
	{
		try{
			return execute(parent);
		}catch(Exception e){
			fail(e);
			return Failed;
		}
	}
	public final void onEvent(Event ev)
	{
		myRun();
	}

}

//####################################################
