package eve.ui.data;
import java.util.Vector;

import eve.data.DataUnit;
import eve.fx.FontMetrics;
import eve.fx.Fx;
import eve.fx.IImage;
import eve.fx.IconAndText;
import eve.sys.Event;
import eve.sys.FieldTransfer;
import eve.sys.Reflection;
import eve.sys.Wrapper;
import eve.ui.Button;
import eve.ui.ButtonBar;
import eve.ui.ButtonPullDownMenu;
import eve.ui.CellPanel;
import eve.ui.Control;
import eve.ui.Gui;
import eve.ui.GuiStyle;
import eve.ui.List;
import eve.ui.Menu;
import eve.ui.MenuItem;
import eve.ui.MessageBox;
import eve.ui.PullDownMenu;
import eve.ui.ScrollBarPanel;
import eve.ui.SplittablePanel;
import eve.ui.event.ListEvent;
import eve.util.Copyable;
import eve.util.mVector;
/**
 * A ListEditor is used to edit a list (contained in a Vector) of Objects
 * - usually LiveData objects.<p>
 * The ListEditor can work in different ways:
 * <ul>
 * <li>The list can be displayed alone and when an item is selected for editing,
 * or when a new item is created, then an Editor is displayed for editing the
 * item. This method is useful for small screen devices.</li>
 * <li>The list can be displayed along with (above, below, left of or right of)
 * an Editor that displays the contents of the selected item and allows the
 * item to be immediately edited. This is useful for desktop systems.
 * </li>
 * </ul>
 * There are buttons that can be displayed which allow for the following
 * functions:
 * <ul>
 * <li>Add a new item.</li>
 * <li>Move a new item, to top, bottom, up one, down one.</li>
 * <li>Delete an item</li>
 * <li>Edit the contents of the default item (i.e. set the initial contents
 * of a new item)</li>
 * </ul>
 * These buttons can be placed at the side of, or below the list.   
 *
 */
//##################################################################
public class ListEditor extends Editor{
//##################################################################

/**
* These are the items being edited.
**/
public Vector items = new Vector();

public boolean confirmDelete = true;//false;

//-------------------------------------------------------------------
static IImage getIcon(String image)
//-------------------------------------------------------------------
{
	return loadImage(image.endsWith(".png") ?  image : "eve/"+image+"small.png");
}

public static IImage upIcon = getIcon("uparrow");
public static IImage downIcon = getIcon("downarrow");
public static IImage topIcon = getIcon("toparrow");
public static IImage bottomIcon = getIcon("bottomarrow");
public static IImage editIcon = getIcon("edit");

//===================================================================
public Editor getEditor(Object ld,int editor)
//===================================================================
{
	if (editor == -1) return null;
	if (ld instanceof LiveData) return ((LiveData)ld).getEditor(editor);
	Editor ed = new Editor();
	ed.setFields(ld);
	return ed;
}
/**
 * This constructor does no setup whatsoever - you will have to call setup()
 * yourself.
 */
public ListEditor(){}

public ListEditor(LiveData forWhat,int display,int editor,int defaultEditor,boolean vertical,int buttonsAndOptions)
//===================================================================
{
	model = forWhat;
	if ((buttonsAndOptions & OPTION_DONT_SETUP) == 0)
		doStandardSetup(vertical,buttonsAndOptions,getEditor(forWhat,display),getEditor(forWhat,editor),getEditor(forWhat,defaultEditor));
}
/**
 * Call this after using the constructor with the same parameters ONLY if 
 * OPTION_DONT_SETUP was included in buttonsAndOptions when the constructor
 * was called.
 * 
 * @param forWhat
 * @param display
 * @param editor
 * @param defaultEditor
 * @param vertical
 * @param buttonsAndOptions
 */
public void setup(LiveData forWhat,int display,int editor,int defaultEditor,boolean vertical,int buttonsAndOptions)
{
	if (forWhat != null)
		model = forWhat;
	else
		forWhat = (LiveData)model;
	doStandardSetup(vertical,buttonsAndOptions,getEditor(forWhat,display),getEditor(forWhat,editor),getEditor(forWhat,defaultEditor));	
}
//===================================================================
public ListEditor(Object forWhat,boolean vertical,int buttonsAndOptions)
//===================================================================
{
	model = forWhat;
	if ((buttonsAndOptions & OPTION_DONT_SETUP) == 0)
		setup(forWhat,vertical,buttonsAndOptions);
}
/**
 * Call this after using the constructor with the same parameters ONLY if 
 * OPTION_DONT_SETUP was included in buttonsAndOptions when the constructor
 * was called.
 * 
 * @param forWhat
 * @param vertical
 * @param buttonsAndOptions
 */
public void setup(Object forWhat, boolean vertical, int buttonsAndOptions)
{
	if (forWhat != null)
		model = forWhat;
	else
		forWhat = (LiveData)model;
	boolean pda = Gui.screenIs(Gui.PDA_SCREEN);
	int display = pda ? -1 : 0;
	int edit = pda ? 0 : -1;
	int def = model instanceof Copyable ? 0 : -1;
	vSplit = pda;
	if (!(model instanceof Copyable)) buttonsAndOptions &= ~ TOOL_BUTTON;
	doStandardSetup(vertical,buttonsAndOptions,getEditor(forWhat,display),getEditor(forWhat,edit),getEditor(forWhat,def));
}
/**
Set the list of items. The what parameter must be a Vector or an Object[]. If it is 
a Vector, then the "items" field is set to it - otherwise a new Vector is created
to hold all the elements of the array and that is assigned to "items".
**/
//===================================================================
public void setData(Object what)
//===================================================================
{
	if (what instanceof Object[])
		what = new mVector((Object[])what);
	if (what == null)
		what = new Vector();
	if (what instanceof Vector) {
		items = (Vector)what;		
		itemsChanged(false,-1);
	}
}
/**
Get the list of items. The what parameter must be a Vector. The Vector is first
cleared and then all the items in the "items" field is copied into it.
**/
//===================================================================
public void getData(Object obj)
//===================================================================
{
	if (obj == items) return;
	if (!(obj instanceof Vector)) return;
	((Vector)obj).clear();
	((Vector)obj).addAll(items);
}

/**
 * Get the number of items.
 */
//===================================================================
public int getItemCount()
//===================================================================
{
	return items.size();	
}
/**
 * Get the item at the specified index.
 * @param index The index of the item to get.
 * @return The item at the index.
 */
//===================================================================
public Object getItem(int index)
//===================================================================
{
	return items.get(index);
}
public Object model; 
/**
* Get the name for the item. By default if it is a LiveData, it will return getName(), otherwise
* it will return item.toString();
**/
//===================================================================
public String getNameFor(Object item)
//===================================================================
{
	if (item == null) return "(null)";
	if (item instanceof LiveData) return ((LiveData)item).getName();
	else {
		Wrapper w = Reflection.invokeMethod(item,"getName()Ljava/lang/String;",null);
		if (w != null && w.isInstance(String.class))
			return (String)w.getObject();
	}
	return item.toString();
}
//===================================================================
public Object getToPutInList(Object item)
//===================================================================
{
	return getNameFor(item);
}
/**
* This is the name of the item being edited.
**/
public String itemName = "Item";

/**
* This is the list of items.
**/
public List list;

public static final int UPDOWN_BUTTONS = 0x1;
public static final int TOPBOTTOM_BUTTONS = 0x2;
public static final int DELETE_BUTTON = 0x4;
public static final int NEW_BUTTON = 0x8;
public static final int EDIT_BUTTON = 0x10;
public static final int TOOL_BUTTON = 0x20;

public static final int ALL_BUTTONS = 0x3f;

public static final int OPTION_NO_LIST_MANIPULATION = 0x8000;
public static final int OPTION_DO_VSPLIT = 0x4000;
public static final int OPTION_DO_HSPLIT = 0x2000;
public static final int OPTION_EDITOR_BEFORE_LIST = 0x1000;
public static final int OPTION_EDITOR_AFTER_LIST = 0x0800;
/**
 * This option when used with a constructor tells the ListEditor not to
 * do the setup. You will have to explicitly call setup() afterwards.
 */
public static final int OPTION_DONT_SETUP = 0x0400;

public String listTitle = null;

public Menu toolsMenu;

//===================================================================
public void make(boolean reMake)
//===================================================================
{
	if (!made || reMake){
		itemsChanged(false,-1);
	}
	super.make(reMake);
}
//-------------------------------------------------------------------
protected Button addButton(CellPanel addTo,boolean vertical,String image,String title,String field)
//-------------------------------------------------------------------
{
	IImage mi = getIcon(image);
	return addButton(addTo,vertical,mi,title,field);
}
//-------------------------------------------------------------------
protected Button addButton(CellPanel addTo,boolean vertical,IImage mi,String title,String field)
//-------------------------------------------------------------------
{
	Button b;
	if (mi.getWidth() == 0) b = new Button(title);
	else b = new Button(new IconAndText(mi,"",getFontMetrics()));
	addTo.addNext(b);
	addField(b,field);
	if (vertical) addTo.endRow();
	return b;
}
/**
* Get the toolbar with the specified buttons.
**/

//===================================================================
private MenuItem iconizeMenu(Menu m,String text,IImage image)
//===================================================================
{
	MenuItem mi = new MenuItem().iconize(text,image,true);
	//mi.image = new IconAndText,getFontMetrics());
	mi.subMenu = m;
	return mi;
}
//===================================================================
public CellPanel getToolBar(boolean vertical,int buttons)
//===================================================================
{
	CellPanel cp = new ButtonBar();
	cp.modify(MouseSensitive,0);
	FontMetrics fm = getFontMetrics();
	
	IImage toolIcon = Fx.loadPicture("eve/SmallConfig.png");
	
	Menu m;
	Menu context = new Menu();
	Menu newMenu = m = new Menu();
	m.text = "New";
	m.addItem(new MenuItem("newItemOnTop",new IconAndText(topIcon,"New "+itemName+" On Top",fm)));
	m.addItem(new MenuItem("newItemAboveSelected",new IconAndText(upIcon,"New "+itemName+" Above Selected",fm)));
	m.addItem(new MenuItem("newItemBelowSelected",new IconAndText(downIcon,"New "+itemName+" Below Selected",fm)));
	m.addItem(new MenuItem("newItem",new IconAndText(bottomIcon,"New "+itemName+" At Bottom",fm)));
	Menu moveMenu = m = new Menu();
	m.text = "Move";
	m.addItem(new MenuItem("moveToTop",new IconAndText(topIcon,itemName+" To Top",fm)));
	m.addItem(new MenuItem("moveUp",new IconAndText(upIcon,itemName+" Up",fm)));
	m.addItem(new MenuItem("moveDown",new IconAndText(downIcon,itemName+" Down",fm)));
	m.addItem(new MenuItem("moveToBottom",new IconAndText(bottomIcon,itemName+" To Bottom",fm)));
	Menu deleteMenu = m = new Menu();
	m.addItem(new MenuItem("dontDelete",new IconAndText(stop,"No, Don't Delete "+itemName,fm)));
	m.addItem(new MenuItem("yesDelete",new IconAndText(tick,"Yes, Delete "+itemName,fm)));
	
	if (toolsMenu == null) toolsMenu = new Menu();
	toolsMenu.text = "Tools";
	if (defaultEditor != null){
		toolsMenu.addItem(new MenuItem("editDefault",new IconAndText(loadImage("eve/editsmall.png"),"Edit Default "+itemName,getFontMetrics())));
		toolsMenu.addItem(new MenuItem("setDefault",new IconAndText(loadImage("eve/savesmall.png"),"Set As Default "+itemName,getFontMetrics())));
	}
	if (editor != null){
		context.addItem(new MenuItem("editItem",new IconAndText(editIcon,"Edit "+itemName,fm)));
		context.addItem("-");
	}
	context.addItem(iconizeMenu(newMenu,"New",getIcon("new")));
	context.addItem(iconizeMenu(moveMenu,"Move",upIcon));
	if (confirmDelete){
		context.addItem(iconizeMenu(deleteMenu,"Delete",getIcon("delete")));		
	}else{
		context.addItem(new MenuItem("deleteItem",new IconAndText(getIcon("delete"),"Delete",fm)));
	}
	if (toolsMenu.items.size() > 0) {
		context.addItem("-");
		context.addItem(iconizeMenu(toolsMenu,"Tools",toolIcon));
	}
	addField(list,"contextMenuItem").setMenu(context);
	if ((buttons & TOOL_BUTTON) != 0){
		PullDownMenu pdm = new ButtonPullDownMenu("",toolsMenu);
		pdm.arrowDirection = 0;
		pdm.image = toolIcon;
		addField(pdm,"toolMenuItem");
		cp.addNext(pdm).setToolTip("Additional Tools");
		if (vertical) cp.endRow();
	}
	if ((buttons & NEW_BUTTON) != 0){
		addField(addButton(cp,vertical,"new","New","newItem"),"newMenu").setToolTip("Add a new "+itemName+"\n(Hold down for options)").setMenu(newMenu);
	}
	if (((buttons & EDIT_BUTTON) != 0) && (editor != null))
		addButton(cp,vertical,editIcon,"Edit","editItem").setToolTip("Edit "+itemName);
	if ((buttons & TOPBOTTOM_BUTTONS) != 0) addButton(cp,vertical,topIcon,"Top","moveToTop").setToolTip("Move "+itemName+" to Top");
	if ((buttons & UPDOWN_BUTTONS) != 0){
		addButton(cp,vertical,upIcon,"Up","moveUp").setToolTip("Move "+itemName+" Up");
		addButton(cp,vertical,downIcon,"Down","moveDown").setToolTip("Move "+itemName+" Down");
	}
	if ((buttons & TOPBOTTOM_BUTTONS) != 0) addButton(cp,vertical,bottomIcon,"Down","moveToBottom").setToolTip("Move "+itemName+" to Bottom");
	
	if ((buttons & DELETE_BUTTON) != 0){
		Control c = addButton(cp,vertical,"delete","Delete","deleteItem").setToolTip("Delete "+itemName+
			(confirmDelete ? "\n(Hold down to confirm)":""));
		if (confirmDelete){
			addField(c,"deleteMenu").setMenu(deleteMenu);
		}
	}
		
	return cp;
}
public MenuItem newMenu = new MenuItem();
public MenuItem deleteMenu = new MenuItem();
public MenuItem toolMenuItem = new MenuItem();
public MenuItem contextMenuItem = new MenuItem();
/**
* If a splittable panel is used, this says if the two sections are vertically layed out.
**/
//===================================================================
public boolean vSplit = true;
//===================================================================

public Editor display;
public Editor editor;
public Editor defaultEditor;

public boolean useSplittablePanel = true;
public boolean editorOnTop = true;
public int listRows = 5, listColumns = 30;

public Editor getDataEditor()
{
	if (editor == null) return display;
	return editor;
}
public void setListTitle(String listTitle)
{
	this.listTitle = listTitle;
	if (listArea != null)
		listArea.setText(listTitle);
}
private CellPanel listArea;

//===================================================================
public void doStandardSetup(boolean verticalToolBar,int buttonsAndOptions,Editor toDisplay,Editor toEdit,Editor defaultEditor)
//===================================================================
{
	int buttons = buttonsAndOptions & ALL_BUTTONS;
	int options = buttonsAndOptions & ~ALL_BUTTONS;
	if ((options & OPTION_DO_VSPLIT) != 0) vSplit = true;
	if ((options & OPTION_DO_HSPLIT) != 0) vSplit = false;
	if ((options & OPTION_EDITOR_BEFORE_LIST) != 0) editorOnTop = true;
	if ((options & OPTION_EDITOR_AFTER_LIST) != 0) editorOnTop = false;
	CellPanel all = new CellPanel();
	display = toDisplay;
	editor = toEdit;
	this.defaultEditor = defaultEditor;
	if (toEdit != null) {
		Gui.setOKCancel(toEdit);
		toEdit.defaultTitleTo("Edit "+itemName);
	}
	if (defaultEditor != null) {
		Gui.setOKCancel(defaultEditor);
		defaultEditor.defaultTitleTo("Edit Default "+itemName);
	}
	if (toEdit != null && toDisplay != null)
		toDisplay.modify(DisplayOnly,0);
	if (toEdit != null && defaultEditor == null && model instanceof Copyable) this.defaultEditor = toEdit;
 	list = new List(listRows,listColumns,false);
	list.notifyDataChangeOnSelect = false;
	if ((options & OPTION_NO_LIST_MANIPULATION) == 0) list.setToolTip("Right-click for Options");
	addField(list,"theList");
	listArea = new CellPanel();
	if (toEdit != null && buttons != 0) buttons |= EDIT_BUTTON;
	CellPanel cp = ((options & OPTION_NO_LIST_MANIPULATION) == 0) ? getToolBar(verticalToolBar,buttons) : new CellPanel();
	listArea.setText(listTitle);
	listArea.addNext(new ScrollBarPanel(list));
	if (verticalToolBar) listArea.addNext(cp).setCell(VSTRETCH).setControl(DONTFILL|CENTER);
	else listArea.endRow();
	if (!verticalToolBar) listArea.addLast(cp).setCell(HSTRETCH).setControl(DONTFILL|CENTER);
	CellPanel top, bottom = null;
	if (useSplittablePanel && toDisplay != null){
		int type = vSplit ? SplittablePanel.VERTICAL : SplittablePanel.HORIZONTAL;
		SplittablePanel sp = new SplittablePanel(type);
		top = sp.getNextPanel();
		bottom = sp.getNextPanel();
		all.addNext(sp);
	}else{
		all.addNext(top = new CellPanel());
		if (vSplit || toDisplay == null) all.endRow();
		if (toDisplay != null) all.addLast(bottom = new CellPanel());
	}
	CellPanel editorPanel = editorOnTop ? top : bottom;
	if (toDisplay != null) {
		editorPanel.setBorder(GuiStyle.inputEdge|BF_RECT,4);
	}
	if (toDisplay != null){
		int constraints = toDisplay.constraints;
		top.addLast(editorOnTop ? toDisplay : listArea);
		bottom.addLast(editorOnTop ? listArea : toDisplay);
		addField(toDisplay,"theEditor");
		(editorOnTop ? top : bottom).constraints = constraints;
		
	}else{
		top.addLast(listArea);
	}
	addLast(all);
}
//===================================================================
public void itemsChanged(boolean doNotify,int select)
//===================================================================
{
	int got = list.modify(Invisible,0);
	list.removeAll();
	for (int i = 0; i<items.size(); i++)
		list.addItem(getToPutInList(items.get(i)));
	list.select(select);
	if (select != -1) list.makeItemVisible(select);
	if (doNotify) notifyDataChange();
	list.updateItems();
	list.restore(got,Invisible);
	list.repaintNow();
	//if (doNotify) 
	newSelected(select);
}
Object editing = null;
//-------------------------------------------------------------------
protected void newSelected(int idx)
//-------------------------------------------------------------------
{
	if (display != null){
		if (idx == -1) display.setObject(editing = null);
		else display.setObject(editing = items.get(idx));
	}
	modifyFields("deleteItem,moveUp,moveDown,moveToTop,moveToBottom,editItem",list.selectedIndex == -1,Disabled,0,true);
}
//===================================================================
public void shown()
//===================================================================
{
	super.shown();
	newSelected(-1);
}
/**
 * Create and return a new default Object.
 */
public Object getNewObject()
//===================================================================
{
	if (model instanceof Copyable) return ((Copyable)model).getCopy();
	else {
		try{
			return model.getClass().newInstance();
		}catch(Exception e){
			return null;
		}
	}
}
/**
* This should be called if an item has its display name changed.
**/ 
//===================================================================
public void nameChanged(int idx)
//===================================================================
{
	list.changeItem(idx,getToPutInList(items.get(idx)));
	list.updateItems();
}
//===================================================================
public void doDelete(int idx,boolean doNotify)
//===================================================================
{
	if (idx<0 || idx>=items.size()) return;
	int was = list.selectedIndex;
	int toSel = was;
	if (idx == was)
		if (idx >= items.size()-1) 
			toSel--;
	items.removeElementAt(idx);
	itemsChanged(doNotify,toSel);
}

//-------------------------------------------------------------------
protected boolean execEditor(Object toEdit,Editor editor)
//-------------------------------------------------------------------
{
		Object dv = null;
		if (toEdit instanceof DataUnit) dv = ((DataUnit)toEdit).getCopy();
		editor.setObject(toEdit);
		if (editor.execute(null,Gui.FILL_FRAME) == IDOK) return true;
		if (dv != null) ((DataUnit)toEdit).copyFrom(dv);
		return false;
}
//===================================================================
public void doEdit(int idx)
//===================================================================
{
	if (editor != null && idx >= 0 && idx < items.size())
		if (execEditor(items.get(idx),editor))
			itemsChanged(true,idx);
}
/**
* This adds a new item as if the user had pressed the New Item button.
* @param toAdd The object to add.
* @param index The index to put the new object in. If it is -1 then it is added to the end.
* @param doNotify If this is true then a standard DataChanged event is generated.
*/
//===================================================================
public void doAddNew(Object toAdd,int index,boolean doNotify)
//===================================================================
{
	if (index == -1) items.add(toAdd);
	else items.insertElementAt(toAdd,index);
	itemsChanged(doNotify,index == -1 ? items.size()-1 : index);
}
/**
* This adds a new item as if the user had pressed the New Item button.
* It calls doAddNew(toAdd,-1,true);
* @param toAdd The object to add.
*/
//===================================================================
public final void doAddNew(Object toAdd)
//===================================================================
{
	doAddNew(toAdd,-1,true);
}

//-------------------------------------------------------------------
protected void shift(int idx,int dx)
//-------------------------------------------------------------------
{
	Object was = items.get(idx);
	items.removeElementAt(idx);
	items.insertElementAt(was,idx+dx);
	itemsChanged(true,idx+dx);
}

//-------------------------------------------------------------------
void addNewAndEdit(int location)
//-------------------------------------------------------------------
{
	if (location < 0) location = 0;
	if (location > items.size()) location = items.size();
	Object newOne = getNewObject();
	if (newOne == null) return;
	if (editor != null) {
		editor.setObject(newOne);
		if (editor.execute() != IDOK) return;
	}
	doAddNew(newOne,location,true);
}
//===================================================================
public void action(FieldTransfer ft,Editor ed)
//===================================================================
{
	action(ft.fieldName,ed);
}
//===================================================================
public void action(String name,Editor ed)
//===================================================================
{
	int idx = list.selectedIndex;
	if (name.equals("deleteItem")){
		if (confirmDelete)
			if (new MessageBox("Delete "+itemName,"Delete selected "+itemName+"?",MBYESNO).execute() != IDOK) return;
		doDelete(idx,true);
	}else if (name.equals("yesDelete")){
		doDelete(idx,true);
	}else if (name.equals("newItem")){
		addNewAndEdit(items.size());
	}else if (name.equals("newItemOnTop")){
		addNewAndEdit(0);
	}else if (name.equals("newItemAboveSelected")){
		addNewAndEdit(idx);
	}else if (name.equals("newItemBelowSelected")){
		addNewAndEdit(idx+1);
	}else if (name.equals("moveUp")){
		if (idx <= 0) return;
		shift(idx,-1);
	}else if (name.equals("moveDown")){
		int len = items.size();
		if (idx < 0 || idx >= len-1) return;
		shift(idx,1);
	}else if (name.equals("moveToTop")){
		if (idx <= 0) return;
		shift(idx,-idx);
	}else if (name.equals("moveToBottom")){
		int len = items.size();
		if (idx < 0 || idx >= len-1) return;
		shift(idx,len-idx-1);
	}else if (name.equals("editItem")){
		if (idx == -1) return;
		doEdit(idx);
	}else if (name.equals("editDefault")){
		if (model != null && defaultEditor != null)
			execEditor(model,defaultEditor);
	}else if (name.equals("setDefault")){
		if (idx == -1) return;
		if (new MessageBox("Set as Default "+itemName,"Set this "+itemName+" to be the default?",MBYESNO).execute() != IDOK) return;		
		Object got = getItem(idx);
		if (got instanceof Copyable) model = ((Copyable)got).getCopy();
	}else if (name.equals("theList") && idx != -1 && editor != null){//Double clicked
		doEdit(idx);
	}
}
//===================================================================
public void fieldEvent(FieldTransfer ft,Editor ed,Object event)
//===================================================================
{
	if (ft.fieldName.equals("theList") && (event instanceof ListEvent)){
		if 	(((Event)event).type == ListEvent.SELECTED){
			newSelected(list.selectedIndex);
		}
	}
}
//===================================================================
public void fieldChanged(FieldTransfer ft,Editor ed)
//===================================================================
{
	if (ft.fieldName.equals("theEditor")){
		if (editing == null || list.selectedIndex == -1) return;
		if (!getToPutInList(editing).equals(list.getItemAt(list.selectedIndex).label)){
			nameChanged(list.selectedIndex);
		}
	}else if (ft.fieldName.equals("newMenu")){
		action(newMenu.label,ed);
	}else if (ft.fieldName.equals("deleteMenu")){
		action(deleteMenu.label,ed);
	}else if (ft.fieldName.equals("toolMenuItem")){
		action(toolMenuItem.label,ed);
	}else if (ft.fieldName.equals("contextMenuItem")){
		action(contextMenuItem.label,ed);
	}else if (ft.fieldName.equals("theList")){
		//ewe.sys.Vm.debug("List changed!");
	}
}
/*
//===================================================================
public static void main(String args[]) throws Exception
//===================================================================
{
	Application.startApplication(args);
	String nm = args.length == 0 ? "eve.database.TestData" : args[0];
	Class c = Class.forName(nm);
	ListEditor ed = new ListEditor(c.newInstance(),true,0xffffffff);
	ed.title = c.getName();
	ed.execute();
	System.exit(0);
}
*/
//##################################################################
}
//##################################################################

