package eve.ui.data;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import eve.io.File;
import eve.io.Io;
import eve.ui.IFileSaver;
import eve.ui.MenuItem;
import eve.ui.MessageBox;
import eve.ui.ReportException;
//##################################################################
public class SaveableObject extends LiveObject{
//##################################################################

public MenuItem chosenMenuItem = new MenuItem();

public String textEncoderName = Io.getDefaultCodecName();
/**
* This is used to reset the Object to a "new" state. By default it will
* copy from a new instance of this class.
**/
//===================================================================
public void newObject()
//===================================================================
{
	copyFrom(getNew());
}
/**
* This is used to save the Object state. By default it will convert the
* object to a text string and then save it.
**/
//===================================================================
public void saveObject(File destination) throws IOException
//===================================================================
{
	String text = textEncode();
	OutputStreamWriter os = new OutputStreamWriter(destination.toWritableStream(false),textEncoderName);
	PrintWriter pw = new PrintWriter(os);
	pw.println(text);
	pw.close();
	savedTo(destination);
}
/**
 * This is used to retrieve the Object state from a File. By default it will
 * treat the file as a text string and decode it.
 */
//===================================================================
public void openObject(File source) throws IOException
//===================================================================
{
	BufferedReader sr = new BufferedReader(new InputStreamReader(source.toReadableStream(),textEncoderName));
	String got = sr.readLine();
	sr.close();
	try{
		if (got == null) throw new Exception();
		textDecode(got);
		openedFrom(source);
	}catch(Exception e){
		throw new IOException("The file: "+source+" did not contain correct data.");
	}
}

public static final int NewAction = 1;
public static final int OpenAction = 2;
public static final int SaveAction = 3;
public static final int ExitAction = 4;


/**
 * This will be called if an exception occured during processing of a File action. By
 * default it do nothing if the FileSaver is not null, otherwise it will display
 * a standard ReportException box.
 * @param whichFileAction one of NewAction, OpenAction, SaveAction.
 * @param e The exception that occured.
 */
//-------------------------------------------------------------------
protected void fileExceptionOccured(int whichFileAction,Exception e,FileSaver saver)
//-------------------------------------------------------------------
{
	if (saver == null || true)
		new ReportException(e,null,null,false).exec();
}
protected FileSaver getFileSaver(Editor ed)
{
	IFileSaver is = ed.getFileSaver();
	if (!(is instanceof FileSaver)) return null;
	return (FileSaver)is;
}
//===================================================================
public void New_action(Editor ed)
//===================================================================
{
	FileSaver saver = getFileSaver(ed);
	if (saver == null) return;
	try{
		if (!saver.newData(this,null,ed.getFrame())){
			if (saver.getException() != null) throw saver.getException();
			return;
		}
		newObject();
		ed.toControls();
	}catch(Exception e){
		//e.printStackTrace();
		fileExceptionOccured(NewAction,e,saver);
	}
}
//-------------------------------------------------------------------
private void doSave(boolean saveAs,Editor ed)
//-------------------------------------------------------------------
{
	FileSaver saver = getFileSaver(ed);
	if (saver == null) return;
	try{
		if (!saver.save(saveAs,this,ed.getFrame())){
			if (saver.getException() != null) throw saver.getException();
			return;
		}
	}catch(Exception e){
		//e.printStackTrace();
		fileExceptionOccured(SaveAction,e,saver);
	}
	
}
//===================================================================
public void Save_action(Editor ed)
//===================================================================
{
	doSave(false,ed);
}
//===================================================================
public void Save_As_action(Editor ed)
//===================================================================
{
	doSave(true,ed);
}
//===================================================================
public void Open_action(Editor ed)
//===================================================================
{
	FileSaver saver = getFileSaver(ed);
	if (saver == null) return;
	try{
		Object n = getNew();
		if (!saver.open(this,n,ed.getFrame())){
			if (saver.getException() != null) throw saver.getException();
			return;
		}
		copyFrom(n);
		ed.toControls();
	}catch(Exception e){
		//e.printStackTrace();
		fileExceptionOccured(OpenAction,e,saver);
	}
}
//===================================================================
public void Exit_action(Editor ed)
//===================================================================
{
	FileSaver saver = getFileSaver(ed);
	if (saver == null) return;
	try{
		Object n = getNew();
		if (!saver.checkSave(this,ed.getFrame())){
			if (saver.getException() != null) throw saver.getException();
			return;
		}
		saver.setHasChanged(false);
		ed.exit(0);
	}catch(Exception e){
		//e.printStackTrace();
		fileExceptionOccured(OpenAction,e,saver);
	}
}
/**
 * This is called after a successful save.
 * @param saved The File the object was saved to.
 */
//-------------------------------------------------------------------
protected void savedTo(File saved)
//-------------------------------------------------------------------
{
}
/**
 * This is called after a successful open.
 * @param saved The File the object was saved to.
 */
//-------------------------------------------------------------------
protected void openedFrom(File opened)
//-------------------------------------------------------------------
{
}
/**
 * This attempts to open a file to load this object.
 * @param source The File object or file name String.
 * @param ed The editor used to edit the SaveableObject.
 * @param showOpenIfFailed If this is true then the FileSaver associated with the editor (if any)
	will be displayed to attempt to open a new File instead.
 * @param showException If this is true then the full file open error will be displayed if an
	error occured opening the file.
 * @return true if a file was opened, false otherwise.
 */
//===================================================================
public boolean tryOpen(Object source,Editor ed,boolean showOpenIfFailed,boolean showException)
//===================================================================
{
	File f = (source instanceof File) ? (File)source : File.getNewFile(source.toString());
	try{
		openObject(f);
		if (ed != null){
			FileSaver fs = getFileSaver(ed);
			if (fs != null) {
				fs.setLastSaved(f.getFullPath());
				fs.setHasChanged(false);
			}
		}
		return true;
	}catch(IOException e){
		if (showException)
			new ReportException(e,"Could not open file.","Check the file name.",false).execute();
		else
			new MessageBox("File Error","Could not open file:\n"+source,MessageBox.MBOK).execute();
		if (ed == null) return false;
		FileSaver fs = getFileSaver(ed);
		if (fs == null) return false;
		try{
			return fs.open(null,this,ed.getFrame());
		}catch(IOException ex){
			if (showException)
				new ReportException(ex,"Could not open file.","Check the file name.",false).execute();
			else
				new MessageBox("File Error","Could not open file:\n"+source,MessageBox.MBOK).execute();
			return false;
		}
	}
}
/**
 * This attempts to open a file to load this object.
 * @param source The File object or file name String.
 * @param ed The editor used to edit the SaveableObject.
 * @return true if a file was opened, false otherwise.
 */
//===================================================================
public boolean tryOpen(Object source,Editor ed)
//===================================================================
{
	return tryOpen(source,ed,true,false);
}
//##################################################################
}
//##################################################################

