package eve.util;
import eve.sys.Convert;
import eve.sys.Vm;
/**
* A CharArray is a utility class that represents a number of text characters. 
* It is generally used as a buffer for text storage.<p>
* Associated with the CharArray is a char [] value called "data" which holds
* the data characters (starting from index 0) and the "length" variable specifies how
* many of those characters are considered valid (which may be less than the actual
* length of the data variable).
**/
//##################################################################
public class CharArray implements Encodable, Copyable{
//##################################################################
/**
* These are the data characters.
**/
public char [] data = new char[0];//Do not move this variable.
/**
* This denotes the number of valid characters in the data array.
**/
public int length = 0;//Do not move this variable.
/**
 * A Convenience method. You can access the length field directly.
 * @return the length field.
 */
public int length()
{
	return length;
}
//===================================================================
public Object getCopy()
//===================================================================
{
	CharArray ba = new CharArray();
	ba.length = length;
	ba.data = new char[length];
	if (length != 0) System.arraycopy(data,0,ba.data,0,length);
	return ba;
}
/**
* This adds space to the CharArray at the specified index. After this is
* done you can safely write into the data starting at index "where" for
* a length of numcharacters.
**/
//===================================================================
public boolean makeSpace(int where,int numcharacters)
//===================================================================
{
	if (numcharacters <= 0) return true;
	if (length+numcharacters > data.length) {
		int expansion = data.length;
		if (expansion > 100000) expansion = 100000;
		if ((length+numcharacters) > (expansion+data.length))
			expansion = 1+length+numcharacters-data.length;
		int newLen = data.length+expansion;
		//if (true) newLen = length+numcharacters;
		char [] nd = new char[newLen];
		System.arraycopy(data,0,nd,0,length);
		data = nd;
	}
	if (where < length)
		System.arraycopy(data,where,data,where+numcharacters,length-where);
	length += numcharacters;
	return true;
}
/**
* This writes into the data at the specific location. Make sure you have called
* makeSpace() before if necessary.
**/
//===================================================================
public void write(int where,char [] src,int start,int length)
//===================================================================
{
	System.arraycopy(src,start,data,where,length);
}
/**
* Insert characters into the CharArray.
 * @param where The index in this CharArray where the data will go.
 * @param src The source data characters.
 * @param start The start index in the source data.
 * @param length The number of characters to copy.
 */
//===================================================================
public CharArray insert(int where, char[] src,int start,int length)
//===================================================================
{
	if (makeSpace(where,length)) write(where,src,start,length);
	return this;
}
public CharArray remove(int where, int count)
{
	if (count < 0) throw new IllegalArgumentException();
	if (where >= length) return this;
	if (where+count > length) count = length-where;
	int left = length-(where+count);
	if (left != 0) System.arraycopy(data,where+count,data,where,left);
	length -= count;
	return this;
}
/**
 * 
 * @param oldChar
 * @param newChar
 * @param start
 * @param length
 * @return itself
 */
public CharArray replace(char oldChar, char newChar, int start, int length)
{
	int old = indexOf(oldChar,start);
	if (old == -1 || old >= start+length) return this;
	for (int i = start; i<length; i++)
		if (data[i] == oldChar) data[i] = newChar;
	return this;
}
/**
 * 
 * @param oldChar
 * @param newChar
 * @return itself
 */
public CharArray replace(char oldChar, char newChar)
{
	return replace(oldChar,newChar,0,length);
}
/**
* Add characters to the end of the CharArray.
 * @param src The source data characters.
 * @param start The start index in the source data.
 * @param length The number of characters to copy.
 */
//===================================================================
public CharArray append(char [] src,int start,int length)
//===================================================================
{
	return insert(this.length,src,start,length);
}
public CharArray insert(char c,int where)
{
	if (makeSpace(where,1)) data[where] = c;
	return this;
}
public CharArray append(char c)
{
	if (makeSpace(length,1)) data[length-1] = c;
	return this;
}
public CharArray(String s)
{
	if (s == null) return;
	data = s.toCharArray();
	length = s.length();
}
//===================================================================
public CharArray append(String src)
//===================================================================
{
	if (src == null) src = "null";
	return append(Vm.getStringChars(src),0,src.length());
}
/**
* This will ensure that the size of the "data" character array is at least
* newCapacity. If the size is already greater than or equal to newCapacity, then
* this will have no effect.
* @param newCapacity The minimum size of the data array.
*/
//===================================================================
public void ensureCapacity(int newCapacity)
//===================================================================
{
	if (data != null && data.length >= newCapacity) return;
	int sz = data == null ? 0 : data.length;
	if (newCapacity < (sz*2+2)) newCapacity = sz*2+2;
	toNewLength(newCapacity);
}
//===================================================================
public String toString()
//===================================================================
{
	return new String(data,0,length);
}
//===================================================================
public void setData(char[] data)
//===================================================================
{
	this.data = data;
	this.length = data.length;
}
public CharArray append(long value)
{
	int need = Convert.formatLong(value,null,0);
	makeSpace(length,need);
	Convert.formatLong(value,data,length-need);
	return this;
}
public CharArray append(int value)
{
	return append((long)value);
}
public CharArray append(double value)
{
	int need = Convert.formatDouble(value,null,0);
	makeSpace(length,need);
	Convert.formatDouble(value,data,length-need);
	return this;
}
//===================================================================
public CharArray() 
//===================================================================
{
}
/**
* This will actually use the "data" char array as the data for the CharArray. It will not copy the data.
**/
//===================================================================
public CharArray(char[] data)
//===================================================================
{
	setData(data);
}
/**
 * Return an array of chars the exact length of this CharArray containing
 * a copy of the chars in this CharArray.
 */
//===================================================================
public char[] toChars()
//===================================================================
{
	char [] got = new char[length];
	if (length != 0) System.arraycopy(data,0,got,0,length);
	return got;
}
/**
* Moves the data into a new data array of the specified length - which
* must be greater than or equal to the number of elements in the array.
**/
//-------------------------------------------------------------------
protected void toNewLength(int newLength)
//-------------------------------------------------------------------
{	
	char[] nd = new char[newLength];
	if (length != 0) System.arraycopy(data,0,nd,0,length);
	data = nd;
}
public char charAt(int idx)
{
	return data[idx];
}
/**
 * Remove excess storage so that the length of the data array matches the
 * number of elements in it.
 */
//===================================================================
public void trim()
//===================================================================
{
	if (data.length > length) toNewLength(length);
}
/**
 * Make enough space to receive incoming data up to a maximum of toReceive characters.
 * This does not affect the length value, it just ensures that after it is called you
 * can safely write starting at data[length] for a length of toReceive characters.
 *
 * After calling this and receiving the data, call received(int received) to indicate
 * how many characters were actually received and placed into the buffer.
 *
 * @param toReceive the maximum number of characters to receive.
 * @param append if this is false the length is set to zero.
 * @return The current length.
 */
//===================================================================
public int prepareReceive(int toReceive,boolean append)
//===================================================================
{
	if (!append) setLength(0);
	ensureCapacity(length+toReceive);
	return length;
}
/**
 * Call this after calling prepareReceive() and after having received some or zero data.
 * It updates the number of characters in the array.
 * @param received the number of characters actually received.
 */
//===================================================================
public void received(int received)
//===================================================================
{
	if (received > 0) length += received;
}
/**
* This sets the length of the valid data within the array and will either truncate or extend
* the array as necessary.
**/
//===================================================================
public void setLength(int len) 
//===================================================================
{
	if (len < 0) throw new IllegalArgumentException();
	if (len <= length) length = len;
	else if (len > length){
		ensureCapacity(len);
		Utils.zeroArrayRegion(data,length,len-length);
		length = len;
	}
}
public void append(StringBuffer source,int srcStart,int srcEnd)
{
	insert(source,length,srcStart,srcEnd);
}
public void insert(StringBuffer source,int where,int srcStart,int srcEnd)
{
	int len = srcEnd-srcStart;
	makeSpace(where,len);
	source.getChars(srcStart,srcEnd,data,where);
}
public void append(StringBuffer source)
{
	insert(source,length,0,source.length());
}
public void insert(StringBuffer source,int where)
{
	insert(source,where,0,source.length());
}
/**
 * Set the "length" field to zero and return this CharArray object.
 */
public CharArray clear()
{
	length = 0;
	return this;
}

//===================================================================
public int indexOf(int ch) {return indexOf(ch,0);}
public int indexOf(int ch,int start)
//===================================================================
{
	return SubString.indexOf((char)ch,data,0,length,start,0);
}
//===================================================================
public int lastIndexOf(int ch) {return lastIndexOf(ch,length-1);}
public int lastIndexOf(int ch,int start)
//===================================================================
{
	return SubString.indexOf((char)ch,data,0,length,start,SubString.BACKWARDS);
}
//===================================================================
public boolean endsWith(String other)
//===================================================================
{
	return endsWith(Vm.getStringChars(other),0,other.length());
}
//===================================================================
public boolean endsWith(char[] otherChars,int otherStart,int otherLength)
//===================================================================
{
	if (otherLength > length) return false;
	return SubString.equals(data,length-otherLength,otherLength,otherChars,otherStart,otherLength,0);
}
public boolean startsWith(String other)
{
	return startsWith(Vm.getStringChars(other),0,other.length());
}
//===================================================================
public boolean startsWith(char[] otherChars,int otherStart,int otherLength)
//===================================================================
{
	if (otherLength > length) return false;
	return SubString.equals(data,0,otherLength,otherChars,otherStart,otherLength,0);
}
//===================================================================
public int indexOf(String other)
{
	return indexOf(Vm.getStringChars(other),0,other.length(),0);
}
public int indexOf(String other,int start)
{
	return indexOf(Vm.getStringChars(other),0,other.length(),start);
}
public int indexOf(char[] otherChars, int otherStart, int otherLength, int start)
//===================================================================
{
	if (otherChars == null) throw new NullPointerException();
	char [] oc = otherChars;
	if (otherLength <= 0) return start;//-1;
	if (otherLength > length) return -1;
	char first = oc[otherStart];
	for (int s = start; s<length; s++){
		int idx = indexOf(first,s); 
		if (idx == -1) return -1;
		if (length-idx < otherLength) return -1;
		if (SubString.equals(data,idx,otherLength,oc,otherStart,otherLength,SubString.STARTS_WITH)) return idx;
		s = idx;
	}
	return -1;
}
//===================================================================
public int lastIndexOf(String other)
{
	return lastIndexOf(Vm.getStringChars(other),0,other.length(),length-1);
}
public int lastIndexOf(String other,int start)
{
	return lastIndexOf(Vm.getStringChars(other),0,other.length(),start);
}

public int lastIndexOf(char[] otherChars, int otherStart, int otherLength, int start)
//===================================================================
{
	if (otherChars == null) throw new NullPointerException();
	char [] oc = otherChars;
	if (otherLength <= 0) return start;//-1;
	if (otherLength > length) return -1;
	char first = oc[otherStart];
	for (int s = start; s>=0; s--){
		int idx = lastIndexOf(first,s);
		if (idx == -1) return -1;
		if (length-idx >= otherLength)
			if (SubString.equals(data,idx,otherLength,oc,otherStart,otherLength,SubString.STARTS_WITH)) return idx;
		s = idx;
	}
	return -1;
}
/**
 * This is called when the CharArray is cached.
 */
public void cached()
{
	length = 0;
}
public static CharArray unNull(CharArray destination)
{
	if (destination == null) destination = new CharArray();
	return destination;
}
//##################################################################
}
//##################################################################

