package eve.util;

/**
* A RangeList manages a set of Range object - but you should never add
* Range objects to it directly. The RangeList always ensures that its
* collection of Ranges are fully optimized. That is if two ranges
* are contiguous, they are merged to form a single range. Similarly
* overlapping ranges are also merged. The ranges
* are also always sorted in increasing order.
**/
//##################################################################
public class RangeList extends mVector{
//##################################################################

protected static Range buff = new Range(0,-1);

//===================================================================
public boolean inRange(int start) {return inRange(start,start);}
//===================================================================

//===================================================================
/**
 * Return if the specified range is already completely within the ranges
 * covered by this RangeList.
 * @param start 
 * @param end
 * @return true if the specified range is already completely within the ranges
 * covered by this RangeList.
 */
public synchronized boolean inRange(int start,int end)
//===================================================================
{
	if (start > end){
		int t = end; end = start; start = t;
	}
	int idx = findRangeWith(start);
	if (idx == -1) return false;
	Range p = (Range)get(idx);
	return p.last >= end;
}
//===================================================================

/**
 * Find the index of the Range that contains the specified value.
 * @param value the value to look for.
 * @return the index of the Range that contains the value or -1 if it
 * is not contained in the RangeList.
 */
public synchronized int findRangeWith(int value)
//===================================================================
{
	int count = size();
	for (int i = 0; i<count; i++){
		Range p = (Range)get(i);
		if (p.last < value) continue;
		if (p.first > value) return -1;
		return i;
	}	
	return -1;
}
//===================================================================
/**
 * Add this range to the RangeList, consolidating ranges as necessary to
 * minimize the number of Range objects within the list.
 * @param start the start of the range.
 * @param end the end of the range.
 */
public synchronized void addRange(int start,int end)
//===================================================================
{
	if (start > end){
		int t = end; end = start; start = t;
	}
	Range p = new Range(start,end);
	int count = size();
	for (int i = 0; i<count; i++){
		Range p2 = (Range)get(i);
		if (p2.first <= start) continue;
		add(i,p);
		p = null;
		break;
	}	
	if (p != null) add(p);
	//System.out.println(this);
	fixup();
	//ewe.sys.Vm.debug(toString());
	//System.out.println("------------------------------------------------");
}

//===================================================================

/**
 * Remove a range from the list
 * @param start the start of the range.
 * @param end the end of the range.
 */
public synchronized void removeRange(int start, int end)
//===================================================================
{
	if (start > end){
		int t = end; end = start; start = t;
	}
	int count = size();
	for (int i = 0; i<count; i++){
		Range me = (Range)get(i);
		if (me.first <= start){
			if (me.last < start) continue;
			if (me.last >= end) { //Here the found range covers the selected range.
				Range p2 = new Range(end+1,me.last);
				add(i+1,p2);
			}
			me.last = start-1;
		}else if (me.first <= end){
			me.first = end+1;
		}else
			break;
	}
	//System.out.println(this);
	fixup();
	//ewe.sys.Vm.debug(toString());
	//System.out.println("------------------------------------------------");
}
//-------------------------------------------------------------------
protected void fixup()
//-------------------------------------------------------------------
{
	int count = size();
	for (int i = 0; i<count-1; i++){
		Range me = (Range)get(i);
		if (me.last < me.first) continue;
		for (int j = i+1; j<count; j++){
			Range other = (Range)get(j);
			if (other.first <= me.last+1){
				me.last = Math.max(me.last,other.last);
				other.last = other.first-1;
			}else
				break;
		}
	}
	for (int i = 0; i<count; i++){
		Range me = (Range)get(i);
		if (me.last < me.first) {
			del(i--);
			count--;
		}
	}
}
/*
public static void main(String args[])
{
	RangeList r = new RangeList();
	r.addRange(5,8);
	r.addRange(1,3);
	r.addRange(12,20);
	r.addRange(4,10);
	r.addRange(11,11);
	r.removeRange(10,10);
	r.removeRange(8,12);
	r.addRange(24,27);
	r.removeRange(4,26);
}
*/
//===================================================================
/**
 * Count the number of items (not ranges) contained within all the Ranges
 * in the RangeList.
 * @return the number of items (not ranges) contained within all the Ranges
 * in the RangeList.
 */
public synchronized int countItems()
//===================================================================
{
	int c = 0;
	int count = size();
	for (int i = 0; i<count; i++){
		Range me = (Range)get(i);
		if (me.last < me.first) continue;
		c += me.last-me.first+1;
	}
	return c;
}
//===================================================================
public synchronized int [] toIntArray()
//===================================================================
{
	int total = countItems();
	int [] ret = new int[total];
	int c = 0;
	int count = size();
	for (int i = 0; i<count; i++){
		Range me = (Range)get(i);
		if (me.last < me.first) continue;
		int len = me.last-me.first+1;
		Utils.getIntSequence(ret,c,me.first,1,len);
		c += len;
	}
	return ret;
}
//===================================================================
public synchronized int getItemAtIndex(int idx)
//===================================================================
{
	int c = 0;
	int count = size();
	for (int i = 0; i<count; i++){
		Range me = (Range)get(i);
		if (me.last < me.first) continue;
		int len = me.last-me.first+1;
		if (c+len > idx) return me.first+(idx-c);
		c += len;
	}
	return -1;
}
/**
* Use this to shift up or down, but be careful shifting down.
**/
//===================================================================
public synchronized void shift(int fromWhere,int places)
//===================================================================
{
	int c = 0;
	int count = size();
	for (int i = 0; i<count; i++){
		Range me = (Range)get(i);
		if (me.last < me.first) continue;
		if (me.last < fromWhere) continue;
		if (me.first < fromWhere){
			Range p2 = new Range(fromWhere,me.last);
			me.last = fromWhere-1;
			add(i+1,p2);
			continue;
		}else{
			me.first += places;
			me.last += places;
		}
	}
}
//===================================================================
public synchronized int countRanges()
//===================================================================
{
	return size();
}
//===================================================================
public synchronized Range rangeAt(int rangeIndex,Range dest)
//===================================================================
{
	if (rangeIndex < 0 || rangeIndex >= size()) return null;
	if (dest == null) dest = new Range(0,-1);
	dest.set((Range)get(rangeIndex));
	return dest;
}
//===================================================================
public RangeList(){}
//===================================================================

//===================================================================
public synchronized void removeAndShiftUp(int start,int end)
//===================================================================
{
	removeRange(start,end);
	shift(end,-((end-start)+1));
}
//##################################################################
}
//##################################################################

