package eve.util;

/**
* A TagList stores a list of Tags (a positive integer tag value associated with an Object value)
* using arrays of integers and object references instead of arrays of Tag objects.<p>
* A TagList only allows the storage of one Object value for each distinct Tag value. That is,
* if you call set(int tag, Object value) and a value already exists for that integer tag value, 
* then this new value will overwrite the old one.
* <p>
* A TagList <b>is</b> synchronized.
**/
//##################################################################
public class TagList{
//##################################################################

protected int num = 0;
int tags [] = null;
Object values [] = null;
public int expandSize = 5;

public static final int EmptyTag = -1;
//===================================================================
public int size() {return num;}
//===================================================================
/**
 * Reset to zero without freeing resources.
 */
public synchronized void zero()
{
	if (num != 0)
		Utils.zeroArrayRegion(values,0,num);
	num = 0;
}
/**
 * Zero the TagList and free resources.
 */
public synchronized void clear()
{
	zero();
	tags = null;
	values = null;
}
/**
 * Returns the int array used for storing the tags.
 * @return The actual int[] used for storing the tags, not a copy.
 */
public synchronized int[] getTags()
{
	if (tags == null) tags = new int[0];
	return tags;
}
/**
 * Returns the Object array used for storing the tag values.
 * @return The actual Object[] used for storing the tag values, not a copy.
 */
public synchronized Object[] getValues()
{
	if (values == null) values = new Object[0];
	return values;
}
//-------------------------------------------------------------------
protected int getTagIndex(int tag)
//-------------------------------------------------------------------
{
	int startIndex = 0;
	for (int i = startIndex; i<num; i++)
		if (tags[i] == tag) return i;
	return -1;
}
//-------------------------------------------------------------------
protected int getEmptyIndex() {return getTagIndex(EmptyTag);}
//-------------------------------------------------------------------

/**
* Get the tag at the specified index. A tag value of EmptyTag (-1) indicates
* not tag value stored at that point.
**/
//===================================================================
public synchronized Tag getAtIndex(int index,Tag dest)
//===================================================================
{
	if (index >= num || index < 0) return null;
	if (dest == null) dest = new Tag();
	dest.tag = tags[index]; dest.value = values[index];
	return dest;
}
/**
* Get the tag Object for the specified tag integer value.
* @param tag The integer tag value to look for.
* @param dest An optional destination tag object.
* @return The destination or new Tag object containing the tag data, or null if
* the tag value is not stored.
*/
//==================================================================
public synchronized Tag get(int tag,Tag dest) {return getAtIndex(getTagIndex(tag),dest);}
//===================================================================

/**
* @param tag The integer tag value to look for.
 * @param defaultValue a default Object value to return if the tag is not found
 * @return The object value of that tag, or the default value if it is not found.
 */
//===================================================================
public synchronized Object getValue(int tag,Object defaultValue)
//===================================================================
{
	int idx = getTagIndex(tag);
	if (idx == -1) return defaultValue;
	return values[idx];
}
//===================================================================
void goingToAdd(int toAdd)
//===================================================================
{
	int len = values == null ? 0 : values.length;
	int more = num+toAdd-len;
	if (more <= 0) return;
	if (more < expandSize) more = expandSize;
	int [] nt = new int[len+more];
	if (len != 0) System.arraycopy(tags,0,nt,0,len);
	tags = nt;
	Object [] nv = new Object[len+more];
	if (len != 0) System.arraycopy(values,0,nv,0,len);
	values = nv;
}
public synchronized void ensureCapacity(int capacity)
{
	int curSize = tags == null ? 0 : tags.length;
	int need = capacity-curSize;
	if (need > 0) goingToAdd(need);
}
/**
* Set the Object value for a tag.
* @param tag The integer tag value to set.
* @param value The Object value to set.
*/
//===================================================================
public synchronized void set(int tag,Object value)
//===================================================================
{
	int idx = getTagIndex(tag);
	if (idx == -1) idx = getEmptyIndex();
	if (idx != -1){
		tags[idx] = tag;
		values[idx] = value;
		return;
	}
	goingToAdd(1);
	tags[num] = tag;
	values[num] = value;
	num++;
}
/**
* Set all the tag values from the other TagList.
* @param tl Another TagList to copy values from.
*/
//===================================================================
public synchronized void set(TagList tl)
//===================================================================
{
	if (tl == null) return;
	Tag tg = null;
	for (int i = 0; i<tl.size(); i++) {
		tg = (Tag)tl.getAtIndex(i,tg);
		set(tg.tag,tg.value);
	}
}
/**
* Remove a tag value.
* @param tag The integer tag value to remove.
*/
//===================================================================
public synchronized void clear(int tag)
//===================================================================
{
	int idx = getTagIndex(tag);
	if (idx == -1) return;
	tags[idx] = EmptyTag;
	values[idx] = null;
}
/**
* Set a value for a tag if it has not already been set.
* @param tag The integer tag value.
* @param value The object tag value.
*/
//===================================================================
public  synchronized void defaultTo(int tag,Object value)
//===================================================================
{
	int idx = getTagIndex(tag);
	if (idx != -1) return;
	set(tag,value);
}
/**
* Check if the specified tag has a value set for it.
* @param tag The integer tag value to check for.
*/
//===================================================================
public  synchronized boolean hasTag(int tag) {return getTagIndex(tag) != -1;}
//===================================================================

//##################################################################
}
//##################################################################

